/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

/*
 * Language support definition file for the network compiler. In this
 * file the names of the preprocessors, parsers and code generators
 * are defined for each individual system configuration.
 */


/*
 * Define the language classes and the languages to be supported.
 */
#include <sysconfig.h> /*C.S. 21 11 2000 */

#define NC_MAXCLASSES   2                       /* Number of language classes */

typedef struct languages {                      /* Language structure */
char    *nc_preprocessor;                       /* Preprocessor */
char    *nc_parser;                             /* Parser */
char    *nc_coder[NC_MAXCLASSES];               /* Coder(s) */
char    *nc_ppflags[NC_MAXCLASSES];             /* Flags to preprocessor */
} nc_tlan;

#define NC_COMPILED     0                       /* Compiled language class */
#define NC_INTERPRETED  1                       /* Interpreted language class */

#define NC_C            0                       /* C */
#define NC_MODULA2      1                       /* Modula II */
#define NC_FORTRAN      2                       /* Fortran */
#define NC_ADA          3                       /* Ada */

#define NC_DEFCLASS     NC_COMPILED             /* Default language class */
#define NC_DEFLAN       NC_C                    /* Default language */

#define NC_TMPFILE      "/tmp/nc"               /* Temporary file path */


/*
 * C language definitions.
 */

#define NCC_PP          CPP_DEFAULT             /* Preprocessor */
#define NCC_P           "np"                    /* Parser */
#define NCC_CC          "ncc"                   /* Coder for compiled language class */
#define NCC_CN          "ncn"                   /* Coder for Nodal language class */
#define NCC_CCFLAGS     "-DCC"                  /* Preprocessor flags for compiled language class */
#define NCC_CNFLAGS     "-DCC -DNODAL"          /* Preprocessor flags for Nodal language class */


/*
 * Modula-II language definitions.
 */

#define NCM_PP          CPP_DEFAULT     /* Preprocessor */
#define NCM_P           "np"                    /* Parser */
#define NCM_CC          "nmc"                   /* Coder for compiled language class */
#define NCM_CN          "nmn"                   /* Coder for Nodal language class */
#define NCM_CCFLAGS     "-DMODULA2"             /* Preprocessor flags for compiled language class */
#define NCM_CNFLAGS     "-DMODULA2 -DNODAL"     /* Preprocessor flags for Nodal language class */


/*
 * Fortran-77 language definitions.
 */

#define NCF_PP          CPP_DEFAULT     /* Preprocessor */
#define NCF_P           "np"                    /* Parser */
#define NCF_CC          "nfc"                   /* Coder for compiled language class */
#define NCF_CN          "nfn"                   /* Coder for Nodal language class */
#define NCF_CCFLAGS     "-DFORTRAN"             /* Preprocessor flags for compiled language class */
#define NCF_CNFLAGS     "-DFORTRAN -DNODAL"     /* Preprocessor flags for Nodal language class */


/*
 * Ada language definitions.
 */

#define NCA_PP          CPP_DEFAULT     /* Preprocessor */
#define NCA_P           "np"                    /* Parser */
#define NCA_CC          "nac"                   /* Coder for compiled language class */
#define NCA_CN          "nan"                   /* Coder for Nodal language class */
#define NCA_CCFLAGS     "-DADA"                 /* Preprocessor flags for compiled language class */
#define NCA_CNFLAGS     "-DADA -DNODAL"         /* Preprocessor flags for Nodal language class */
