/*
****************************************************************************
* MODULE   : (Screen) Editor for NODAL.
* FILENAME : nodaledi.c
* VERSION  : 1.4
* STATUS   : Development
* SYSTEM   : NODAL EDITOR.
* HISTORY  : April/May 84 created.                                     HPChr
*            07/08/88 C version created                                HPChr
*            05/10/88 full screen editor as separate process           HPChr
*
****************************************************************************
*/

#include <ctype.h>
#include "systemke.h"
#include "nodaledi.h"
#include "stringha.h"
#include "hardchar.h"
#include "screenha.h"
#include "virtscre.h"

#define   skipmark  '%'
#define   expbegin  '<'
#define   expend    '>'
#define   killword  '\\'
#define   killine   '<'

static bool EditOk, edflg;
static int retcurs = 1;
static int copyfrompos;

/*
** local procedure declarations
*/
#ifdef ANSII
void Posright(Descriptor *);
void Posleft(Descriptor *);
void ForwTab();
void BackTab();
void insertchar(char, Descriptor *, Edcmnd *);
void DelChars(Descriptor *, int);
void DeleteInLine(Descriptor *);
void Forward(Descriptor *);
void Backward(Descriptor *);
void Copychar(Descriptor *, Descriptor *, Edcmnd *);
void linedit(Descriptor *, Edcmnd *, char *);
void SaveHistory(Descriptor *);
bool GetHistory(Edcmnd, Descriptor *);
#else
void Posright();
void Posleft();
void ForwTab();
void BackTab();
void insertchar();
void DelChars();
void DeleteInLine();
void Forward();
void Backward();
void Copychar();
void linedit();
void SaveHistory();
bool GetHistory();
#endif


static void Posright(dp)
Descriptor *dp;
{
   if (textpos.pos < dp->bytecount - 1) {
      textpos.pos++;
   }
   if (! Expand) {
      ++copyfrompos;
   }
}


static void Posleft(dp)
Descriptor *dp;
{
   if (textpos.pos > 0) {
      --textpos.pos;
   }
   if (copyfrompos > 0) {
      --copyfrompos;
   }
}


static void ForwTab()
{
   int oldp;

   oldp = textpos.pos++;
   while (textpos.pos < MAXLENG-1 && Tabulators[textpos.pos] != TAB) {
     textpos.pos++;
   }
   if (textpos.pos > MAXLENG-1) textpos.pos = MAXLENG-1;
   copyfrompos += textpos.pos - oldp;
}


static void BackTab()
{
   int oldp;

   oldp = textpos.pos;
   if (textpos.pos > 0) {
     textpos.pos--;
     while ((textpos.pos > 0) && (Tabulators[textpos.pos] != TAB)) {
	textpos.pos--;
     }
   }
   if (copyfrompos >= oldp - textpos.pos) {
      copyfrompos -= (oldp - textpos.pos);
   } else {
      copyfrompos = 0;
   }
}


static void insertchar(ch, dp, cmnd)
char  ch;
Descriptor *dp;
Edcmnd  *cmnd;
{
   int   ii;
   bool  display, ok = TRUE;
   char  cc, oldch;

   display = (*cmnd != copyprev);
   if (*cmnd == control) { /* accept control character */
      ch = Getinp(cmnd);
   }
   oldch = dp->stringpointer[textpos.pos];
   while (textpos.pos > dp->writepointer)
      wci(dp, ' ');

   if (textpos.pos == dp->writepointer && dp->writepointer < dp->bytecount)
      dp->writepointer++;
   else if (Expand && textpos.pos < dp->writepointer) {
      ii = dp->writepointer;
      if (dp->writepointer > dp->bytecount - 1) ii = dp->bytecount-1;
      while (ii > textpos.pos) {
	 dp->stringpointer[ii] = dp->stringpointer[ii-1];
	 ii--;
      }
      if (dp->writepointer < dp->bytecount)
	 dp->writepointer++;
   }
   dp->stringpointer[textpos.pos] = ch;
   if (display)
      Typechar(dp, oldch);

   Posright(dp);
   if (Expand && (*cmnd == copyprev || *cmnd == copychar))
      copyfrompos++;

   edflg = TRUE;
}


static void DelChars(dp, nrchar)
Descriptor *dp;
int         nrchar;
{
   int i2, i;

   if (dp->writepointer > 0 && textpos.pos < dp->writepointer) {
      EraseChars(dp, nrchar);
      i2 = textpos.pos;
      for (i = i2 + nrchar; i < dp->writepointer; i++)
	 dp->stringpointer[i2++] = dp->stringpointer[i];

      for (i = i2; i < dp->bytecount; i++)
	dp->stringpointer[i] = ' ';

      dp->writepointer = i2;
      edflg = TRUE;
   }
}


static void DeleteInLine(dp)
Descriptor *dp;
{
   int      nrchar;
   bool     ok;
   Edcmnd   cmnd;
   char     ch;

   ch = Getinp(&cmnd);
   if (cmnd == delline) {
      if (len(dp) > 0) {
	 textpos.pos = 0;
	 SetCurs(dp);
	 EraseChars(dp, dp->writepointer);
	 reset(dp);
	 edflg = TRUE;
	 copyfrompos = 0;
      }
   } else {
      ok = TRUE;
      if (cmnd == newline) {
	 if (dp->writepointer > textpos.pos) { /* to end of line */
	    nrchar = dp->writepointer - textpos.pos;
	 } else {
	    ok = FALSE;
	    textpos.pos = dp->writepointer;
	 }
      } else if (cmnd == backw) {
	 nrchar = textpos.pos + 1; /* from beginning of line */
	 copyfrompos = textpos.pos = 0;
	 SetCurs(dp);
      } else if (cmnd == empty) {  /* search line for character == ch */
	 nrchar = 1;
	 while (textpos.pos + nrchar <= dp->writepointer &&
	       dp->stringpointer[textpos.pos + nrchar - 1] != ch) {
	    nrchar++;
	 }
	 ok = textpos.pos + nrchar <= dp->writepointer;
      } else {
	 ok = FALSE;
      }
      if (ok)
	 DelChars(dp, nrchar);
   }
}


static void Forward(dp)
Descriptor *dp;
{
   int      i, oldp;
   Edcmnd   cmnd;
   char     ch2;

   oldp = (i = textpos.pos);
   ch2 = Getinp(&cmnd);
   if (cmnd == forw || cmnd == newline) {
      i = dp->writepointer;
   } else if (cmnd == empty) {
      while (i < dp->writepointer && dp->stringpointer[i] != ch2) {
	 i++;
      }
      if (i >= dp->writepointer) {
	 i = textpos.pos;
      }
   }
   if (i == dp->bytecount) i--;
   textpos.pos = i;

   copyfrompos += textpos.pos - oldp;
}


static void Backward(dp)
Descriptor *dp;
{
   int      i, oldp;
   Edcmnd   cmnd;
   char     ch2;

   oldp = i = textpos.pos;
   ch2 = Getinp(&cmnd);
   if (cmnd == backw || cmnd == newline) {
      i = 0;
   } else if (cmnd == empty) {
      while (i > 0 && dp->stringpointer[i] != ch2)
	 i--;

      if (i == 0 && dp->stringpointer[i] != ch2) {
	 i = textpos.pos;
      }
   }

   textpos.pos = i;

   if (copyfrompos >= oldp - textpos.pos)
      copyfrompos -= (oldp - textpos.pos);
   else
      copyfrompos = 0;

}


static void Copychar(dp, cpdp, cmnd)
Descriptor *dp, *cpdp;
Edcmnd     *cmnd;
{
   cardinal   i, ii, prevpos, minpos;
   bool       ok, Newl;
   char       ch;

   ok = TRUE; Newl = FALSE;
   minpos = 0;

   ok = (ok && cpdp->writepointer > minpos);
   prevpos = textpos.pos;
   ii = (i = copyfrompos);
   if (*cmnd == copyprev || *cmnd == copynext) {
      ch = Getinp(cmnd);
      if (ok) {
	 if (*cmnd == copyprev) {
	    ii = cpdp->writepointer-1;
	 } else if (*cmnd == newline) {
	    ii = cpdp->writepointer-1;
	    Newl = TRUE;
	 } else if (*cmnd == empty) {
	    while (ii < cpdp->writepointer && cpdp->stringpointer[ii] != ch) {
	       ++ii;
	    }
	 } else {
	    return;
	 }
      }
      *cmnd = copyprev;
   } else {
      *cmnd = copychar;
   }
   if (ok) {
      if (i < cpdp->writepointer) {
	 while (i <= ii) {
	    ch = cpdp->stringpointer[i++];
	    insertchar(ch, dp, cmnd);
	 }
	 if (*cmnd == copyprev) {
	    WriteDes(dp, prevpos); /* type line */
	 }
	 if (Newl) {
	   *cmnd = newline;
	 }
      }
   }
}


static void linedit(dp, cmnd, ch)
Descriptor *dp;
Edcmnd     *cmnd;
char       *ch;
/*             */
{
   for (;;) {
      SetCurs(dp);
      *ch = Getinp(cmnd);
      switch (*cmnd) {
	case expand  : SetExpand(); break;
	case right   : Posright(dp); break;
	case left    : Posleft(dp); break;
	case rubout  : if (textpos.pos > 0) {
			  Posleft(dp); SetCurs(dp);
		       } else
			 break;
	case delchar : DelChars(dp, 1); break;
	case forw    : Forward(dp); break;
	case backw   : Backward(dp); break;
	case ftab    : ForwTab(); break;
	case btab    : BackTab(); break;
	case empty:
	case control : insertchar(*ch, dp, cmnd); break;
	default:
	   return;
      } /* switch */
   }
}


#define    HIST 10

static NodalString  History[HIST];
static cardinal     nrhist=0, currhist=0, savehist=0;
static bool         histflag = FALSE;

static void SaveHistory(buflin)
Descriptor *buflin;
/*  Save last line in history */
{
    char *ptr;
    cardinal i;

    buflin->readpointer =  0;
    ptr = History[savehist];
    if (len(buflin) > 0) {
       for (i = 0; i < buflin->writepointer; i++) {
	  ptr[i] = buflin->stringpointer[i];
       }
       if (buflin->bytecount > buflin->writepointer) {
	  ptr[buflin->writepointer] = nul;
       }
       currhist = savehist++;
       if (savehist >= HIST) {
	  savehist = 0;
       }
       if (nrhist < HIST) {
	  ++nrhist;
       }
    }
    histflag = FALSE;
}


static bool GetHistory(cmnd, buflin)
Edcmnd cmnd;
Descriptor *buflin;
/*  Get history line from round robin buffer */
{
   char *ptr;
   cardinal i;

   if (nrhist == 0)
      return FALSE;
   reset(buflin);
   if (cmnd == up) {
      if (histflag) { /* allready done one */
	 if (((nrhist == HIST) && (currhist == savehist)) ||
	    ((nrhist < HIST) && (currhist == 0))) {
	    histflag = FALSE;
	    if (currhist > 0) {
	       --currhist;
	    } else {
	       currhist = nrhist-1;
	    }
	    /* return */
	    return (TRUE);
	 }
	 if (currhist > 0) {
	    --currhist;
	 } else {
	    currhist = nrhist-1;
	 }
      } /* else use the last currhist from previous save */
   } else { /* cmnd == down */
      ++currhist;
      if (currhist >= nrhist) {
	 currhist = 0;
      }
      if (histflag && (((nrhist == HIST) && (currhist == savehist)) ||
		     ((nrhist < HIST) && (currhist == 0)))) {
	 histflag = FALSE;
	 if (currhist > 0) {
	    --currhist;
	 } else {
	    currhist = nrhist-1;
	 }
	 return (TRUE);
      }
   }
   histflag = TRUE;
   ptr = History[currhist];
   i = 0;
   while (*ptr != nul && i < MAXLENG) {
      buflin->stringpointer[i++] = *ptr++;
   }
   buflin->writepointer = i;
   return (TRUE);
}


void InitTerminal()
{
   EditOk = ScreenTyp();
}


void ExitTerminal()
{
   ScreenCall(CCEND);
}


GetlReturn  getl(edlin, buflin)
Descriptor  **edlin, **buflin;
/*
** read line from input device with editing
*/
{
   Descriptor *dp;
   cardinal   bytc;
   bool   expandmode = FALSE, edflg = TRUE, sign;
   int    cc, ch, cl, full = 0;

   (*edlin)->readpointer = 0;
   (*buflin)->readpointer = ((*buflin)->writepointer = 0);
   do {
      sign = FALSE;
      if ((cc = Read()) == del )
	 cc = soh;
      if ( cc == cr || cc == lf ) { /* ctm - terminate edit */
				   /*  cr - terminate edit */
	 edflg = FALSE;
      } else if (cc < fs && NODALmode == interactive) {
	 if ( cc == soh ) { /* cta - backspace one character */
	    if ((ch = gcd(*buflin)) != EOST) {
	       vs_Write(bs); vs_Write(' '); vs_Write(bs);
	    } else {
	       sign = TRUE;
	    }
	 } else if ( cc == stx ) { /* ctb - end edition */
	    return (ERRLIN);
	 } else if ( cc == etx ) { /* ctc - copy one character from old line */
	    if ((ch = gci(*edlin)) != EOST) {
	       vs_Write(ch);
	       full = wci(*buflin, ch);
	    } else {
	       sign = TRUE;
	    }
	 } else if ( cc == eot ) { /* ctd - copy remainder of old line */
	    while ( (ch=gci(*edlin)) != EOST ) {
	       vs_Write(ch);
	       full = wci(*buflin, ch);
	    }
	    edflg = FALSE;
	 } else if ( cc == enq ) { /* cte - complement insert mode */
	    if ( !expandmode ) {
	       vs_Write(expbegin);
	       expandmode = TRUE;
	    } else {
	       vs_Write(expend);
	       expandmode = FALSE;
	    }
	 } else if ( cc == ack ) { /* ctf - copy the remainder of old line
				   without listing */
	     while ( (ch= gci(*edlin)) != EOST ) {
		full = wci(*buflin, ch);
	     }
	     edflg = FALSE;
	 } else if ( cc == bs ) {    /* cth - copy to end of old line */
	    while ( (ch = gci(*edlin)) != EOST ) {
	       vs_Write(ch);
	       full = wci(*buflin, ch);
	    }
	 } else if ( cc == si ) { /* cto - copy up to specified character */
	    cl = Read();
	    while ((ch = gci(*edlin)) != EOST && ch != cl) {
	       vs_Write(ch);
	       full = wci(*buflin, ch);
	    }
	    if ( ch == cl ) {
	       incr(*edlin, -1);
	    } else {
	       sign = TRUE;
	    }
	 } else if ( cc == dle ) { /* ctp - skip up to specified character */
	    cl = Read();
	    while ( (ch = gci(*edlin)) != EOST && ch != cl) {
	       vs_Write(skipmark);
	    }
	    if ( ch == cl ) {
	       incr(*edlin, -1);
	    } else {
	       sign = TRUE;
	    }
	 } else if ( cc == dc1 ) { /* ctq - delete line */
	    (*edlin)->readpointer = 0;
	    (*buflin)->readpointer = ((*buflin)->writepointer = 0);
	    vs_Write(killine); vs_WriteLn();
	 } else if ( cc == dc2 ) {     /* ctr - retype line */
	    bytc = (*edlin)->readpointer;
	    vs_WriteLn(); sout(*edlin);
	    (*edlin)->readpointer = bytc;
	    vs_WriteLn(); sout(*buflin);
	    (*buflin)->readpointer = 0;
	 } else if ( cc == dc3 ) { /* cts - skip one character */
	    if (gci(*edlin) != EOST) {
	       vs_Write(skipmark);
	    } else {
	       sign = TRUE;
	    }
	 } else if ( cc == syn ) { /* ctv - take next characeter literally */
	    ch = Read();
	    vs_Write(ch);
	    full = wci(*edlin, ch);
	 } else if ( cc == etb ) { /* ctw - delete last word */
	    while ( (ch = gcd(*buflin)) == ' ') ;
	    while ( (ch != EOST) && (ch != ' ') ) {
		 ch = gcd(*buflin);
	    }
	    vs_Write(killword);
	 } else if ( cc == can ) { /* ctx - skip through specified character */
	    cl = Read();
	    vs_Write(skipmark);
	    while ( (ch= gci(*edlin)) != EOST && ch != cl) {
		vs_Write(skipmark);
	    }
	    if ( ch != cl ) {
	       sign = TRUE;
	    }
	 } else if ( cc == em ) { /* cty - reedit */
	    while ( (ch = gci(*edlin)) != EOST ) {
	       vs_Write(ch);
	       full = wci(*buflin, ch);
	    }
	    dp = *edlin; *edlin = *buflin; *buflin = dp;
	 } else if ( cc == sub ) { /* ctz - copy through specified character */
	    cl = Read();
	    while ( (ch = gci(*edlin)) != EOST && ch != cl) {
	       vs_Write(ch);
	       full = wci(*buflin, ch);
	    }
	    if ( ch != EOST) {
	       vs_Write(ch);
	       full = wci(*buflin, ch);
	    } else {
	       sign = TRUE;
	    }
	 } else if ( cc == esc ) {
	     return (ESCAPED);
	 } else {
	    sign = TRUE;
	 }
	 if ( sign ) {
	    vs_Write(bel);
	 }
      } else if ( cc >= ' ' && cc <= '~') {
	 if ( NODALmode == interactive ) vs_Write(cc);
	 full = wci(*buflin, cc);
	 if (!expandmode) {
	    incr(*edlin, 1);
	 }
      } else {
	 if ( NODALmode != interactive ) {
	    NodalException(IllegalEdit);
	 }
	 vs_Write(bel);
      }
   } while (edflg && full != EOST);
   if ( NODALmode == interactive ) vs_WriteLn();
   return (RETOK);
}


GetlReturn EditLine(pstr, edlin, buflin)
char *pstr;
Descriptor **edlin, **buflin;
/*  read line from input device with editing */
{
   char        ch;
   Descriptor *dp;
   Edcmnd      cmnd;
   GetlReturn  retval;

   if ( NODALmode != interactive || !EditOk)
      return getl(edlin, buflin);

   InitScreenHandler(strlen(pstr));
   textpos.pos = 0;
   if ((*buflin)->writepointer > 0) {
      /* return with ERRLIN */
      Curs.X = retcurs;
      SetCurs(*buflin);
      WriteDes(*buflin, 0);
      textpos.pos = (*buflin)->readpointer;
      (*buflin)->readpointer = 0;
   } else {
      reset(*buflin);
   }
   edflg = FALSE;
   copyfrompos = textpos.pos;
   for (;;) {
      linedit(*buflin, &cmnd, &ch);
      switch (cmnd) {
	case debug   : retcurs = Curs.X; retval = ERRLIN; goto lexit;
	case copychar:
	case copyprev: Copychar(*buflin, *edlin, &cmnd);/* copy char(s) */
		       SetCurs(*buflin); break;
	case delline: DeleteInLine(*buflin); break;
	case cesc:
	case home : retval = ESCAPED; goto lexit;
	case newline   : break;
	case up:
	case down  : if ((strlen(pstr) == 1) && GetHistory(cmnd, *buflin)) {
			   textpos.pos = 0;
			   SetCurs(*buflin);
			   WriteDes(*buflin, 0);
			   textpos.pos = (*buflin)->writepointer;
			   copyfrompos = textpos.pos;
			   edflg = len(*buflin) > 0;
			 }
		      break;
	default :
	  vs_Write(bel);
      }/* switch  */
      if (cmnd == newline) {
	 retval = RETOK;
	 if (strlen(pstr) == 1 && edflg)
	    SaveHistory(*buflin);
	 break;
      }
   } /* for */
lexit:
   if (Expand)
      SetExpand();
   LeaveEd();
   if (retval != ERRLIN)
      vs_WriteLn();
   return (retval);
}


bool Screenedit(linenr)
cardinal linenr;
/*
** edits a nodal text in the buffer; Linenr is the line to be shown in
** the first page; PED compatible
*/
{
   NodalString lnstr, cntxtstr;
   Descriptor  des, cntxtdes;
   char  *envval, *getenv();
   char  fname[20], *mktemp(), *head1, *strcpy(), *strrchr();
   int pid, stat;
   char *args[5];
   FILE  *fd;
   static  char *envnoded = "NODEDI";
   static  char *defnoded = "/usr/bin/nodedit";
   static  char *enved = "EDITOR";
   static  char *defed = "/bin/vi";
   bool   options_ok;

   nsset(&des, lnstr);
   nsset(&cntxtdes, cntxtstr);
   if (EditOk) {
      mscop(&cntxtdes, "-f");
      GetProgram(CurrContext->program, &cntxtdes);
      if ((envval = getenv(envnoded)) == NIL || strlen(envval) == 0) {
	 if ((fd = xfopen(defnoded, "r")) == NIL){
	    if ((envval = getenv(enved)) == NIL || strlen(envval) == 0) {
	       if ((fd = xfopen(defed, "r")) == NIL){
		  NodalException(IllegalEdit);
	       } else {
		  xfclose(fd);
		  envval = defed;
	       }
	    }
	 } else {
	    xfclose(fd);
	    envval = defnoded;
	 }
      }
      if ((args[0] = strrchr(envval,'/')) == NIL)
	 args[0] = envval;
      else
	 args[0]++;
      options_ok = strcmp(args[0], "nodedit") == 0;
      mscop(&des, "-l");
      onus(&des, linenr);

      /* open temp file */
      head1 = mktemp(strcpy(fname, "/tmp/XXXXXX"));
      OpenOutput(head1);
      ListProgram(0, 0x6400, FALSE);
      CloseOutput();
      if ((pid = xfork()) == 0) {  /* child */
	 /* invoke function */
	 if (options_ok) {
	    args[1] = lnstr;
	    args[2] = cntxtstr;
	    args[3] = head1;
	    args[4] = NIL;
	 } else {
	    args[1] = head1;
	    args[2] = NIL;
	 }
	 execv(envval, args);
	 exit(2); /* never executed */
      } else if (pid > 0) {
	 while (xwait(&stat) != pid)
	    ;
	 if ((stat & 0xFF) == 0) { /*exit*/
	    stat = (stat >> 8) & 0xFF; /* exit code */
	    if (stat == 0) { /* ok, read file back */
	       EraseLines(0, 0x6363);
	       if ((stat = loadtfile(head1)) < 1) {
		  unlink(head1);
		  NodalException(FileError);
	       }
	    } else if (stat == 2) { /* not ok, error */
	       unlink(head1);
	       NodalException(NoFile);
	    }
	 }
	 unlink(head1);
      } else {
	 unlink(head1);
	 NodalException(Unallocated);
      }
   } /* if */
   return (EditOk);
}

int loadtfile(filename)
char *filename;
{

     Descriptor  dp;
     NodalString   linestr;
     FILE *ln;
     longreal LineNumber;
     cardinal LineCode;
     int i;

     nsset(&dp, linestr);
     if ((ln = xfopen(filename, "r")) == NIL ) {
	return (0);
     } /* if /*/
     for (;;){
	if (xfgets(linestr, MAXLENG, ln) == NIL) {
	   if (feof(ln))
	      break;
	   else {
	      xfclose(ln);
	      return (-1);
	   }
	}
	i = 0;
	while (i < MAXLENG && linestr[i] != '\n' && linestr[i] != '\0')
	   i++;

	dp.readpointer = 0;
	dp.writepointer = i;
	if (fin(&dp, &LineNumber) ) {
	   LineCode = lnum(LineNumber);
	   if ( (LineCode % 0x100) != 0 )
	       LineStore(LineCode, &dp);
	} /* if */
     }
     xfclose(ln);
     return (1);
}
