/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 *
 * 03.07.90 kk  removed interrupt interface
 */

#include "sysconfig.h"
#include "np.h"


char *module_name       = NP_DEF_MODULENAME;
char *prefix_name       = NP_DEF_PREFIXNAME;
int  np_message_binding = 0;


/*
 * Generate comments, #includes, and client binding.
 * Also check if the module name has been defined by the user.
 * This action is called before the module body has been reduced.
 */

module_header(symbol)
	struct object *symbol;
{
	if (recursive_flag)
		return;

	if( np_module ) {
		module_name = symbol->o_name = np_module;
	} else {
		module_name = symbol->o_name;
	}

	if( np_gen_prefix ) {
		prefix_name = make_prefix(module_name);
	}

	defs_header(module_name);
	support_header(module_name);
	if( np_gen_client ) {
		client_header(module_name);
	}
	if( np_gen_server ) {
		server_header(module_name);
	}
}

module_tail()
{
	end_defs_header(module_name);
}


/*
 * Generate the support header.
 */

support_header(module_name)
	char *module_name;
{
	fprintf(support1,
"/*\n\
 * Support routines for %s service.\n\
 * Generated by NC compiler version %d.%d.\n\
 */\n\
#include \"%s_defs.h\"\n",
		module_name, NC_VERSION, NC_REVISION, module_name);
}


/*
 * Recursively parse a module to get types and constants.
 */

ref_module(symbol)
	struct object *symbol;
{
	int save_offset;
	char *save_input_file;
	char buf[MAXSTR];
	int p;

	save_offset = ftell(stdin);
	save_input_file = input_file;
	sprintf(buf, "%s.her", symbol->o_name);
	input_file = buf;
	if (freopen(input_file, "r", stdin) != NULL) {
		p = save_parser_state();
		(void) yyparse();
		restore_parser_state(p);
	} else {
		error(SERIOUS, "file %s not found", input_file);
	}
	input_file = save_input_file;
	freopen(input_file, "r", stdin);
	fseek(stdin, save_offset, 0);
}


/*
 * Check that a module name was listed in the DEPENDS UPON list.
 */

check_dependency(name)
	struct object *name;
{
	if (!member(name, dependencies)) {
		error(SERIOUS, "service %s not in dependency list", name->o_name);
		return (0);
	}
	return (1);
}


/*
 * Generate functions for client and server calls to a procedure.
 */

proc_functions(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	if( np_gen_server ) {
		server_procedure(proc_name, type, proc_number, random_id);
	}
	if( np_gen_client ) {
		client_procedure(proc_name, type, proc_number, random_id);
	}
}


/*
 * Generate functions for client and server calls to an execution.
 */

exec_functions(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	if( np_gen_server ) {
		server_procedure(proc_name, type, proc_number, random_id);
	}
	if( np_gen_client ) {
		client_execution(proc_name, type, proc_number, random_id);
	}
}


/*
 * Generate functions for client and server message calls.
 */

message_functions(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	if( np_gen_server ) {
		server_message(proc_name, type, proc_number, random_id);
		np_message_binding = 1;
	}
	if( np_gen_client ) {
		client_message(proc_name, type, proc_number, random_id);
	}
}


/*
 * Generate functions for client and server file access calls.
 */

file_functions(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	if( np_gen_server ) {
		server_file_access(proc_name, type, proc_number, random_id);
	}
	if( np_gen_client ) {
		client_file_access(proc_name, type, proc_number, random_id);
	}
}


/*
 * Generate functions for client and server memory access calls.
 */

memory_functions(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	if( np_gen_server ) {
		server_memory(proc_name, type, proc_number, random_id);
	}
	if( np_gen_client ) {
		client_memory(proc_name, type, proc_number, random_id);
	}
}


/*
 * Generate code relating to binding.
 * This action is called after the entire module has been reduced.
 */

wrapup_module(mod)
	struct object *mod;
{
	if (recursive_flag)
		return;
	/*
	 * module_name should have been set by now,
	 * but a little paranoia never hurt anyone.
	 */
	if (!streq(mod->o_name, module_name)) {
		error(FATAL, "internal error (module): conflicting service names %s and %s",
			mod->o_name, module_name);
	}
	if( np_gen_client ) {
		client_binding();
	}
	if( np_gen_server ) {
		if( np_message_binding ) {
			message_binding();
		}
		server_binding();
	}
}


/*
 * Generate the addressing mode for the parameter passing.
 */

genamode( class_modifier, address_mode )
	enum cmod class_modifier;
	enum addr address_mode;
{
	if( class_modifier != O_ARG )
		error(FATAL, "internal error (genamode): illegal argument");
	return( (int)address_mode );
}


/*
 * Generate the direction mode for the parameter passing.
 */

gendmode( class_modifier, direction_mode )
	enum cmod class_modifier;
	enum dir direction_mode;
{
	if( class_modifier != O_ARG )
		error(FATAL, "internal error (gendmode): illegal argument");
	return( (int)direction_mode );
}
