/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

#include "sysconfig.h"
#include <ctype.h>
#include "np.h"


/*
 * Generate the array translation and size functions.
 */

struct object *
array_type(size, element_type)
	struct object *size, *element_type;
{
	struct object *type, *bt;
	int true_size;
	char *ref, *format;

	type = make_type(C_ARRAY);
	type->t_basetype = element_type;
	type->t_size = size;
	type->t_name = gensym();
	/*
	 * Print a C definition for the array.
	 */
	fprintf(header, "\ntypedef %s %s[%s];\n",
		rep(element_type), type->t_name, rep(size));
	/*
	 * Generate a sizeof function for the array.
	 * The size is equal to the sum of the sizes of each element.
	 */
	bt = basetype(element_type);
	ref = refstr(bt);
	true_size = true_value(size);
	if (bt->t_xsize != -1) {
		/*
		 * The element type, and hence the array, is fixed-size,
		 * so just define a macro.
		 */
		type->t_xsize = true_size * bt->t_xsize;
		fprintf(header,
"\n\
#define sizeof_%s(p) %d\n",
			rep(type), type->t_xsize);
	} else {
		/*
		 * The element type is variable-size, so define a function.
		 */
		type->t_xsize = -1;
		fprintf(support1,
"\n\
static int\n\
sizeof_%s(p)\n\
\tregister %s *p;\n\
{\n\
\tregister int size = 0;\n\
\tregister int i;\n\
\n\
\tfor (i = 0; i < %d; i += 1)\n\
\t\tsize += sizeof_%s(%sp[i]);\n\
\treturn (size);\n\
}\n",
			rep(type), rep(element_type), true_size,
			rep(element_type), ref);
	}
	/*
	 * Define translation functions.
	 */
	xfn_header(EXTERNALIZE, type, element_type);
	xfn_header(INTERNALIZE, type, element_type);
	format =
"{\n\
\tregister _btype *bp;\n\
\tregister int i;\n\
\n\
\tbp = buf;\n\
\tfor (i = 0; i < %d; i += 1)\n\
\t\tbp += %s(%sp[i], bp);\n\
\treturn (bp - buf);\n\
}\n";
	fprintf(support1, format,
		true_size, xfn(EXTERNALIZE, element_type), ref);
	fprintf(support2, format,
		true_size, xfn(INTERNALIZE, element_type), ref);
	return (type);
}
