/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

#include "sysconfig.h"
#include "np.h"


/*
 * Generate the client stub for a memory access call.
 */

client_memory(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	list p, q;
	int nresults, fixed_size, variable_size;
#ifdef CHK_INOUT
	int first;
#endif
	struct object *t, *bt, *result_type;
	char *result_name, *ref, *rtname, *result_ref, *bufname;
	int a_mode, d_mode;

	/*
	 * Make sure there is at most one result returned.
	 */
	nresults = length(type->t_results);
	if (nresults == 1)
		/*
		 * Could be multiple names with one type.
		 */
		nresults = length(car(car(type->t_results)));
	if (nresults > 1) {
		error(SERIOUS, "procedures that return multiple results are not supported");
		return;
	}
	if (nresults) {
		result_name = ocar(car(car(type->t_results)))->o_name;
		result_type = ocdr(car(type->t_results));
		rtname = rep(result_type);
	} else {
		rtname = "void";
	}

	/*
	 * Stub routine for client.
	 */
	if (nresults) {
		p = type->t_results;
		q = car(car(p));
		t = ocdr(car(p));
		result_ref = indamode(basetype(t), ocar(q)->o_addr);
	} else {
		result_ref = "";
	}

	extern_general( rtname, result_ref, prefix_name, proc_name, "write", "read");

	/*
	 * Generate the stub for writeing.
	 */
	fprintf(client,
"\n\
int\n\
%s%s_write(",
		prefix_name, proc_name);
#ifdef CHK_INOUT
	for (p = type->t_args, first = 1; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_OUT) {
			if ( !first )
				fprintf(client, ", ");
			for (q = car(car(p)); q != NIL; q = cdr(q)) {
				if (q != car(car(p)))
					fprintf(client, ", ");
				fprintf(client, "%s", ocar(q)->o_name);
			}
			first = 0;
		}
	}
#else
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			fprintf(client, "%s", ocar(q)->o_name);
			if (cdr(q) != NIL)
				fprintf(client, ", ");
		}
		if (cdr(p) != NIL)
			fprintf(client, ", ");
	}
#endif
	fprintf(client, ")\n");
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
#ifdef CHK_INOUT
		if (ocar(q)->o_dir != D_OUT) {
#else
		{
#endif
			t = ocdr(car(p));
			ref = indamode(basetype(t), ocar(q)->o_addr);
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(client, "\t%s %s%s;\n", rep(t), ref, ocar(q)->o_name);
		}
	}
	fprintf(client, "{\n");
	fprintf(client,
"\tregister _btype *_buf;\n\
\tregister _btype *_bp;\n"
		);

	/*
	 * Generate code to externalize the arguments.
	 */
	p = type->t_args;
	q = car(car(p));
	bt = basetype(ocdr(car(p)));
	bufname = ocar(q)->o_name;
	/* ref = refstr(bt); */
	ref = indamode(bt, ocar(q)->o_addr);
	fprintf(client,
"\n\
\t_buf = _bp = %s%s;\n",
		ref, bufname);

#ifdef CHK_INOUT
	gen_memexternalize( type, P_SPECIFIED );
#else
	gen_memexternalize( type, P_NP_ALL );
#endif

	/*
	 * Return the number of bytes.
	 */
	fprintf(client,
"\treturn(_bp - _buf);\n\
}\n");


	/*
	 * Generate the stub for reading.
	 */
	fprintf(client,
"\n\
int\n\
%s%s_read(",
		prefix_name, proc_name);
#ifdef CHK_INOUT
	for (p = type->t_args, first = 1; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_IN) {
			if ( !first )
				fprintf(client, ", ");
			for (q = car(car(p)); q != NIL; q = cdr(q)) {
				if (q != car(car(p)))
					fprintf(client, ", ");
				fprintf(client, "%s", ocar(q)->o_name);
			}
			first = 0;
		}
	}
#else
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			fprintf(client, "%s", ocar(q)->o_name);
			if (cdr(q) != NIL)
				fprintf(client, ", ");
		}
		if (cdr(p) != NIL)
			fprintf(client, ", ");
	}
#endif
	fprintf(client, ")\n");
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
#ifdef CHK_INOUT
		if (ocar(q)->o_dir != D_IN) {
#else
		{
#endif
			t = ocdr(car(p));
			ref = indamode(basetype(t), ocar(q)->o_addr);
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(client, "\t%s %s%s;\n", rep(t), ref, ocar(q)->o_name);
		}
	}
	fprintf(client, "{\n");
	if (nresults) {
		q = car(car(type->t_results));
		if (ocar(q)->o_addr == A_VALUE)
			fprintf(client, "\t%s %s;\n", rtname, result_name);
		else
			fprintf(client, "\tstatic %s %s;\n", rtname, result_name);
	}
	fprintf(client,
"\tregister _btype *_buf;\n\
\tregister _btype *_rp;\n"
		);

	/*
	 * Generate code to internalize the arguments.
	 */
	p = type->t_args;
	q = car(car(p));
	bt = basetype(ocdr(car(p)));
	/* ref = refstr(bt); */
	ref = indamode(bt, ocar(q)->o_addr);
	fprintf(client,
"\n\
\t_buf = _rp = %s%s;\n",
		ref, bufname);

	/*
	 * Generate code to internalize the results.
	 */
	gen_memresults( type, nresults, result_type, result_name );

	/*
	 * Generate code to internalize the OUT arguments.
	 */
#ifdef CHK_INOUT
 	gen_meminternalize( type, P_SPECIFIED );
#else
	gen_meminternalize( type, P_NP_ALL );
#endif

	/*
	 * Return the number of bytes.
	 */
	fprintf(client,
"\treturn(_rp - _buf);\n\
}\n");

}
