/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

#include "sysconfig.h"
#include "np.h"


/*
 * Generate the client stub for a message call.
 */

client_message(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	list p, q;
	int nresults, nerrors, fixed_size, variable_size;
	int first;
	struct object *t, *bt, *result_type;
	char *result_name, *ref, *rtname, *result_ref;
	int a_mode, d_mode;

	/*
	 * Make sure there is at most one result returned.
	 */
	nresults = length(type->t_results);
	if (nresults == 1)
		/*
		 * Could be multiple names with one type.
		 */
		nresults = length(car(car(type->t_results)));
	if (nresults > 0) {
		error(SERIOUS, "message cannot have return parameter");
		return;
	}
	if (nresults) {
		result_name = ocar(car(car(type->t_results)))->o_name;
		result_type = ocdr(car(type->t_results));
		rtname = rep(result_type);
	} else {
		rtname = "void";
	}

	/*
	 * Stub routine for client.
	 */
	if (nresults) {
		p = type->t_results;
		q = car(car(p));
		t = ocdr(car(p));
		result_ref = indamode(basetype(t), ocar(q)->o_addr);
	} else {
		result_ref = "";
	}

	extern_general(rtname, result_ref, prefix_name, proc_name, "cl_send", "cl_recv");

	/*
	 * Generate the stub for sending.
	 */
	fprintf(client,
"\n\
int\n\
%s%s_cl_send(",
		prefix_name, proc_name);
	for (p = type->t_args, first = 1; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_OUT) {
			if ( !first )
				fprintf(client, ", ");
			for (q = car(car(p)); q != NIL; q = cdr(q)) {
				if (q != car(car(p)))
					fprintf(client, ", ");
				fprintf(client, "%s", ocar(q)->o_name);
			}
			first = 0;
		}
	}
	fprintf(client, ")\n");
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_OUT) {
			t = ocdr(car(p));
			ref = indamode(basetype(t), ocar(q)->o_addr);
			for (q = car(car(p)); q != NIL; q = cdr(q))
#if defined(APOLLO10K)
			    /* floats and doubles are put into registers if not static */
			    if( ((t->t_id == T_FLOATING) || (t->t_id == T_LONG_FLOATING)  )
				    && ( ocar(car(car(type->t_args)))->o_addr == A_VALUE ) )
				fprintf(server, "\tstatic %s %s%s;\n", rep(t), ref, ocar(q)->o_name);
			    else
#endif
				fprintf(client, "\t%s %s%s;\n", rep(t), ref, ocar(q)->o_name);
		}
	}
	fprintf(client, "{\n");
	fprintf(client,
"\t_btype *_buf;\n\
\tregister _btype *_bp;\n"
		);
	/*
	 * Determine the size of the arguments.
	 * This is like the code in record_type().
	 */
	fixed_size = 0;
	variable_size = 0;
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_OUT) {
			bt = basetype(ocdr(car(p)));
			if (bt->t_xsize == -1) {
				variable_size = 1;
			} else {
				/* count multiple arguments of the same type */
				fixed_size += length(car(car(p))) * bt->t_xsize;
			}
		}
	}

	if (!variable_size) {
		/*
		 * The argument list is fixed-size.
		 */
		fprintf(client,
"\n\
\t_buf = nc_allocate(CALL_HEADER_BYTES + %d);\n",
			fixed_size);
	} else {
		/*
		 * There are some variable-size arguments.
		 */
		fprintf(client,
"\tregister int _n = %d;\n\
\n",
			fixed_size);
		for (p = type->t_args; p != NIL; p = cdr(p)) {
			q = car(car(p));
			if (ocar(q)->o_dir != D_OUT) {
				t = ocdr(car(p));
				bt = basetype(t);
				if (bt->t_xsize != -1)
					continue;
				/* ref = refstr(bt); */
				ref = refamode(bt, ocar(q)->o_addr);
				for (q = car(car(p)); q != NIL; q = cdr(q)) {
					fprintf(client,
"\t_n += sizeof_%s(%s%s);\n",
						rep(t), ref, ocar(q)->o_name);
				}
			}
		}
		fprintf(client,
"\t_buf = nc_allocate(CALL_HEADER_BYTES + _n);\n"
			);
	}
	fprintf(client,
"\t_bp = &_buf[CALL_HEADER_BYTES];\n"
		);

	/*
	 * Begin critical section.
	 */
	client_begin_critical();

	/*
	 * Generate code to externalize the arguments.
	 */
	gen_externalize( type, P_SPECIFIED );

	if (!variable_size) {
		fprintf(client,
"\tif((nc_send_call(NCC_MESSAGE, _%s_import, %d, %ldL, _buf, CALL_HEADER_BYTES + %d)) == -1)\n",
			module_name, proc_number, random_id, fixed_size);
	} else {
		fprintf(client,
"\tif((nc_send_call(NCC_MESSAGE, _%s_import, %d, %ldL, _buf, CALL_HEADER_BYTES + _n)) == -1)\n",
			module_name, proc_number, random_id);
	}
	gen_error(client, type, "int", "", 1);

	/*
	 * End critical section.
	 */
	client_end_critical();

	fprintf(client,
"\treturn(0);\n}\n");


	/*
	 * Generate the stub for receiveing.
	 */
	fprintf(client,
"\n\
int\n\
%s%s_cl_recv(",
		prefix_name, proc_name);
	for (p = type->t_args, first = 1; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_IN) {
			if ( !first )
				fprintf(client, ", ");
			for (q = car(car(p)); q != NIL; q = cdr(q)) {
				if (q != car(car(p)))
					fprintf(client, ", ");
				fprintf(client, "%s", ocar(q)->o_name);
			}
			first = 0;
		}
	}
	fprintf(client, ")\n");
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_IN) {
			t = ocdr(car(p));
			ref = indamode(basetype(t), ocar(q)->o_addr);
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(client, "\t%s %s%s;\n", rep(t), ref, ocar(q)->o_name);
		}
	}
	fprintf(client, "{\n");
	if (nresults) {
		q = car(car(type->t_results));
		if (ocar(q)->o_addr == A_VALUE)
			fprintf(client, "\t%s %s;\n", rtname, result_name);
		else
			fprintf(client, "\tstatic %s %s;\n", rtname, result_name);
	}
	fprintf(client,
"\t_btype *_bp;\n\
\tregister _btype *_rp;\n\n"
		);

	/*
	 * Begin critical section.
	 */
	client_begin_critical();

	fprintf(client,
"\tif(nc_getclmsg(_%s_import, %d, &_bp) != 0)\n", module_name, proc_number);
	gen_error(client, type, "int", "", 1);

	/*
	 * Generate code to internalize the results.
	 */
	gen_mesresults( type, nresults, result_type, result_name );

	/*
	 * Generate code to internalize the OUT arguments.
	 */
	gen_internalize( type, P_SPECIFIED );

	/*
	 * End critical section.
	 */
	client_end_critical();

	/*
	 * Deallocate the buffer and return the return value if any.
	 */
	if (nresults) {
		p = type->t_results;
		q = car(car(p));
		t = ocdr(car(p));
		ref = refamode(basetype(t), chamode(ocar(q)->o_addr));
		fprintf(client,
"\tnc_deallocate(_bp);\n\
\treturn (%s%s);\n\
}\n",
			ref, result_name);
	} else {
		fprintf(client,
"\tnc_deallocate(_bp);\n\
\treturn(0);\n}\n");
	}
}
