/*@(#)cproc.c	1.2.1.2          5/8/89*/
/*
 * Creation 1986 by Inge B. Kristiansen.
 *
 * 08/05/89 (phd+vfr) GLOBALs
 *
 * xx.05.90 kk  Version 2 of NC/RPC
 * 21.05.90 kk  Modified call to the nc library routine. Added returning error.
 * 25.06.90 tg  Returning zero as error and raising exception added
 * 10.09.90 kk  NCC_NOREPLY message in case no return values
 * 06.12.90 kk  added static on local variables of type float and double
 *              in case of APOLLO10K. Otherwise they are allocated in
 *              registers and cannot be dereferenced (result only).
 * 06.12.90 kk  corrected returning exception.
 */

#include "sysconfig.h"
#include "np.h"

extern struct object *globallist;

/*
 * Generate the client header and the client binding.
 */

client_header(module_name)
	char *module_name;
{
	fprintf(client,
"/*\n\
 * Client stubs for %s service.\n\
 * Generated by NC compiler version %d.%d.\n\
 */\n\
#include \"%s_support.c\"\n",
		module_name, NC_VERSION, NC_REVISION, module_name);

	fprintf(client,
"\n\
static int _%s_import;\n",
		module_name);
}


/*
 * Generate the client stub for a procedure call.
 */

client_procedure(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	list p, q;
	int nresults, fixed_size, variable_size, returned_values;
	struct object *t, *bt, *result_type;
	char *result_name, *ref, *procref, *rtname;
	int a_mode, d_mode;

	/*
	 * Make sure there is at most one result returned.
	 */
	nresults = length(type->t_results);
	if (nresults == 1)
		/*
		 * Could be multiple names with one type.
		 */
		nresults = length(car(car(type->t_results)));
	if (nresults > 1) {
		error(SERIOUS, "procedures that return multiple results are not supported");
		return;
	}
	if (nresults) {
		result_name = ocar(car(car(type->t_results)))->o_name;
		result_type = ocdr(car(type->t_results));
		rtname = rep(result_type);
	} else {
		rtname = "void";
	}

	/*
	 * Stub routine for client.
	 */
	if (nresults) {
		p = type->t_results;
		q = car(car(p));
		t = ocdr(car(p));
		procref = indamode(basetype(t), ocar(q)->o_addr);
	} else {
		procref = "";
	}

	extern_procedure(rtname, procref, prefix_name, proc_name);




	fprintf(client,
"\n\
%s %s\n\
%s%s(",
		rtname, procref, prefix_name, proc_name);
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			fprintf(client, "%s", ocar(q)->o_name);
			if (cdr(q) != NIL)
				fprintf(client, ", ");
		}
		if (cdr(p) != NIL)
			fprintf(client, ", ");
	}
	fprintf(client, ")\n");
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		t = ocdr(car(p));
		ref = indamode(basetype(t), ocar(q)->o_addr);
		for (q = car(car(p)); q != NIL; q = cdr(q))
			fprintf(client, "\t%s %s%s;\n", rep(t), ref, ocar(q)->o_name);
	}
	fprintf(client, "{\n");
	if (nresults) {
		q = car(car(type->t_results));
		if (ocar(q)->o_addr == A_VALUE) {
#if defined(APOLLO10K)
		    /* floats and doubles are put into registers if not static */
		    if( (result_type->t_id == T_FLOATING) || (result_type->t_id == T_LONG_FLOATING)  )
			    fprintf(client, "\tstatic %s %s;\n", rtname, result_name);
		    else
#endif
			fprintf(client, "\t%s %s;\n", rtname, result_name);
		} else
			fprintf(client, "\tstatic %s %s;\n", rtname, result_name);
	}
	fprintf(client,
"\t_btype *_buf;\n\
\tregister _btype *_bp;\n\
\tregister _btype *_rp;\n"
		);
	/*
	 * Determine the size of the arguments.
	 * This is like the code in record_type().
	 */
	fixed_size = 0;
	variable_size = returned_values = 0;
	if (globallist != (struct object *)NULL) {
	    for (p = globallist->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_OUT) {
		    bt = basetype(ocdr(car(p)));
		    if (bt->t_xsize == -1) {
			variable_size = 1;
		    } else {
			/* count multiple arguments of the same type */
			fixed_size += length(car(car(p))) * bt->t_xsize;
		    }
		}
	    }
	}
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir != D_OUT) {
			bt = basetype(ocdr(car(p)));
			if (bt->t_xsize == -1) {
				variable_size = 1;
			} else {
				/* count multiple arguments of the same type */
				fixed_size += length(car(car(p))) * bt->t_xsize;
			}
		}
		if (ocar(q)->o_dir != D_IN) {
			returned_values = 1;
		}
	}
	if (!variable_size) {
		/*
		 * The argument list is fixed-size.
		 */
		fprintf(client,
"\n\
\t_buf = nc_allocate(CALL_HEADER_BYTES + %d);\n",
			fixed_size);
	} else {
		/*
		 * There are some variable-size arguments.
		 */
		fprintf(client,
"\tregister int _n = %d;\n\
\n",
			fixed_size);
		if (globallist != (struct object *)NULL) {
		    for (p = globallist->t_args; p != NIL; p = cdr(p)) {
			q = car(car(p));
			if (ocar(q)->o_dir != D_OUT) {
			    t = ocdr(car(p));
			    bt = basetype(t);
			    if (bt->t_xsize != -1)
				continue;
			    /* ref = refstr(bt); */
			    ref = refamode(bt, ocar(q)->o_addr);
			    for (q = car(car(p)); q != NIL; q = cdr(q)) {
				fprintf(client,
					"\t_n += sizeof_%s(%s%s);\n",
					rep(t), ref, ocar(q)->o_name);
			    }
			}
		    }
		}
		for (p = type->t_args; p != NIL; p = cdr(p)) {
			q = car(car(p));
			if (ocar(q)->o_dir != D_OUT) {
				t = ocdr(car(p));
				bt = basetype(t);
				if (bt->t_xsize != -1)
					continue;
				/* ref = refstr(bt); */
				ref = refamode(bt, ocar(q)->o_addr);
				for (q = car(car(p)); q != NIL; q = cdr(q)) {
					fprintf(client,
"\t_n += sizeof_%s(%s%s);\n",
						rep(t), ref, ocar(q)->o_name);
				}
			}
		}
		fprintf(client,
"\t_buf = nc_allocate(CALL_HEADER_BYTES + _n);  /* nc_call() will free */\n"
			);
	}
	fprintf(client,
"\t_bp = &_buf[CALL_HEADER_BYTES];\n"
		);

	/*
	 * Generate code to externalize the arguments.
	 */
	if (globallist != (struct object *)NULL) {
	    gen_externalize( globallist, P_SPECIFIED );
	}
	gen_externalize( type, P_SPECIFIED );

	/* if we do not have results we call nc_send_call() */
	if (!(nresults || returned_values || (globallist != (struct object *)NULL))) {
	    if (!variable_size) {
		fprintf(client,
"\tif((nc_send_call(NCC_NOREPLY, _%s_import, %d, %ldL, _buf, CALL_HEADER_BYTES + %d)) == -1)\n",
			module_name, proc_number, random_id, fixed_size);
	    } else {
		fprintf(client,
"\tif((nc_send_call(NCC_NOREPLY, _%s_import, %d, %ldL, _buf, CALL_HEADER_BYTES + _n)) == -1)\n",
			module_name, proc_number, random_id);
	    }
	    gen_error(client, type, rtname, procref, nresults);
	    fprintf(client, "}\n");
	    return;
	}

	if (!variable_size) {
		fprintf(client,
"\tif((_rp = _bp = nc_call(NCC_CALL, _%s_import, %d, %ldL, _buf, CALL_HEADER_BYTES + %d, (char *)0, 0)) == 0)\n",
			module_name, proc_number, random_id, fixed_size);
	} else {
		fprintf(client,
"\tif((_rp = _bp = nc_call(NCC_CALL, _%s_import, %d, %ldL, _buf, CALL_HEADER_BYTES + _n, (char *)0, 0)) == 0)\n",
			module_name, proc_number, random_id);
	}
	/* check the results and return error or set exception */
	gen_error(client, type, rtname, procref, nresults);

	/*
	 * Generate code to internalize the results.
	 */
	gen_results( type, nresults, result_type, result_name );

	/*
	 * Generate code to internalize the OUT arguments.
	 */
	if (globallist != (struct object *)NULL) {
	    gen_internalize( globallist, P_SPECIFIED );
	}
	gen_internalize( type, P_SPECIFIED );

	/*
	 * Deallocate the buffer and return the return value if any.
	 */
	 fprintf(client, "\tnc_deallocate(_bp); /* allocated by nc_call */\n");
	if (nresults) {
		p = type->t_results;
		q = car(car(p));
		t = ocdr(car(p));
		ref = refamode(basetype(t), chamode(ocar(q)->o_addr));
		fprintf(client, "\treturn (%s%s);\n}\n",
			ref, result_name);
	} else {
		fprintf(client, "}\n");
	}
}


/*
 * Generate function for importing server module.
 */

client_binding()
{
	/* Generate the import function.                                */
	client_import( "import_service", module_name, "" );

	/* Generate the free function.                                  */
	client_free( "free_service", module_name );

	/* Generate the bind function.                                  */
	client_bindto( "bind_service", module_name );
}


/*
 * Generate error handling code after the remote proc call in the stub
 */
/*
 * There are  three cases for exception (error) handler generation:
 *	- the exception is indicated by returning a null buffer pointer by the
 *	  call_remote_proc function:
 *		- (-1) should be returned in case of an error. (Minus )
 *		- (0) should be returned in case of an error. (Zero  )
 *              - an exception should be raised, to be compatible
 *                with the old version of the network compiler (Xception
 *                or Exception)
 */

int 
gen_error(where, type, rtname, procref, nresult)
	FILE *where;
	struct object *type;
	char *rtname, *procref;
	int nresult;
{
	char exchg;

	if ( length(type->t_errors) == 0)
		exchg = 'M';
	else
		exchg= (ocar((type->t_errors))->o_name[0]);

	switch (exchg) {
	    case 'M':
	    case 'm':
		if (nresult == 0 )      /* void procedure */
			fprintf(where, "\t\treturn ;\n");
		else
			fprintf(where, "\t\treturn ((%s %s)-1);\n", rtname, procref);
		return(0);
	    case 'Z':
	    case 'z':
		if (nresult == 0 )      /* void procedure */
			fprintf(where, "\t\treturn;\n");
		else
			fprintf(where, "\t\treturn ((%s %s) 0);\n", rtname, procref);
		return(0);
	    case 'X':
	    case 'x':
	    case 'E':
	    case 'e':
		fprintf(where, "\t\tnc_raise( errtype, errno, errstring );\n" );
		return(0);
	    default:
		error(WARNING, "valid exception params: M, Z, X or E: default (-1) assumed");
		fprintf(where, "\t\treturn ((%s %s)-1);\n", rtname, procref);
		return(-1);
	}
}
