#include "sysconfig.h"
#include <ctype.h>
#include "np.h"


/*
 * Make a sequence type.
 */

struct object *
sequence_type(size, element_type, allocation)
	struct object *size, *element_type;
	int allocation;
{
	struct object *type, *bt;
	int true_size;
	char *ref, *format;

	type = make_type(C_SEQUENCE);
	type->t_basetype = element_type;
	type->t_size = size;
	type->t_name = gensym();
	/*
	 * Print a C definition for the sequence.
	 */
	fprintf(header,
"\n\
typedef struct {\n\
\tidl_ushort_int length;\n\
\t%s *sequence;\n\
} %s;\n",
		rep(element_type), type->t_name);
	/*
	 * Generate a sizeof function for the sequence.
	 * The size is equal to 1 (for the length word)
	 * plus the sum of the sizes of each element.
	 */
	type->t_xsize = -1;
	bt = basetype(element_type);
	ref = refstr(bt);
	true_size = true_value(size);
	if (bt->t_xsize != -1) {
		/*
		 * The element type is fixed-size, so just define a macro.
		 */
		if( true_size == 0 ) {
			fprintf(header,
"\n\
#define sizeof_%s(p) (sizeof(idl_ushort_int) + (p)->length * %d)\n",
				rep(type), bt->t_xsize);
		} else {
			fprintf(header,
"\n\
#define sizeof_%s(p) (((p)->length < %d) ? (sizeof(idl_ushort_int) + (p)->length * %d) : (sizeof(idl_ushort_int) + %d))\n",
				rep(type), true_size, bt->t_xsize, true_size * bt->t_xsize);
		}
	} else {
		/*
		 * The element type is variable-size, so define a function.
		 */
		fprintf(support1,
"\n\
static int\n\
sizeof_%s(p)\n\
\tregister %s *p;\n\
{\n\
\tregister int size = 2;\n\
\tidl_ushort_int length;\n\
\tregister int i;\n\
\n",
			rep(type), rep(type));

		if( true_size == 0 ) {
			fprintf(support1,
"\tlength = p->length;\n"
			);
		} else {
			fprintf(support1,
"\tlength = (p->length < %d) ? p->length : %d;\n",
			true_size, true_size);
		}

		fprintf(support1,
"\tfor (i = 0; i < length; i += 1)\n\
\t\tsize += sizeof_%s(%sp->sequence[i]);\n\
\treturn (size);\n\
}\n",
			rep(element_type), ref);
	}
	/*
	 * Define translation functions.
	 */
	xfn_header(EXTERNALIZE, type, type);
	xfn_header(INTERNALIZE, type, type);
	format =
"{\n\
\t_btype *bp;\n\
\tidl_ushort_int length;\n\
\tregister int i;\n\
\n";
	fprintf(support1, format, xfn(EXTERNALIZE, cardinal_type));
	fprintf(support2, format, xfn(INTERNALIZE, cardinal_type));
	/*
	 * Get the length of the sequence for externalize.
	 */
	if( true_size == 0 ) {
		fprintf(support1,
"\tlength = p->length;\n"
		);
	} else {
		fprintf(support1,
"\tlength = (p->length < %d) ? p->length : %d;\n",
			true_size, true_size);
	}
	/*
	 * Externalize/internalize the length itself.
	 */
	format =
"\tbp = buf + %s(&length, buf);\n";
	fprintf(support1, format, xfn(EXTERNALIZE, cardinal_type));
	fprintf(support2, format, xfn(INTERNALIZE, cardinal_type));
	/*
	 * Get the length of the sequence for internalize.
	 * In this case ignore true_size. We have to read what is
	 * on the net anyway, and the destination is allocated
	 * dynamically.
	 */
	if( true_size == 0 ) {
		fprintf(support2,
"\tp->length = length;\n"
		);
	} else {
		fprintf(support2,
"\tif ((p->length = length) > %d)\n\
\t\tnc_raise(ETNCSTB, ENCCTOOLSEQ, \"sequence too long\");\n",
			true_size);
	}
	/*
	 * The internalize function needs to allocate space
	 * for the sequence elements dynamically.
	 */
	/* only if noalloc not specified */
	if (allocation) {
	    fprintf(support2,
"\tp->sequence = (%s *)\n\
\t\tnc_allocate(length * sizeof(%s));\n",
		rep(type->t_basetype), rep(type->t_basetype));
	}
	format =
"\tfor (i = 0; i < length; i += 1)\n\
\t\tbp += %s(%sp->sequence[i], bp);\n\
\treturn (bp - buf);\n\
}\n";
	fprintf(support1, format, xfn(EXTERNALIZE, element_type), ref);
	fprintf(support2, format, xfn(INTERNALIZE, element_type), ref);
	return (type);
}

/*
 * Similar to sequence, nc_sequence allows bigger variable length arrays
 */

struct object *
nc_sequence_type(size, element_type, allocation)
	struct object *size, *element_type;
	int allocation;
{
	struct object *type, *bt;
	int true_size;
	char *ref, *format;

	type = make_type(C_SEQUENCE);
	type->t_basetype = element_type;
	type->t_size = size;
	type->t_name = gensym();
	/*
	 * Print a C definition for the sequence.
	 */
	fprintf(header,
"\n\
typedef struct {\n\
\tunsigned long length;\n\
\t%s *sequence;\n\
} %s;\n",
		rep(element_type), type->t_name);
	/*
	 * Generate a sizeof function for the sequence.
	 * The size is equal to 1 (for the length word)
	 * plus the sum of the sizes of each element.
	 */
	type->t_xsize = -1;
	bt = basetype(element_type);
	ref = refstr(bt);
	true_size = true_value(size);
	if (bt->t_xsize != -1) {
		/*
		 * The element type is fixed-size, so just define a macro.
		 */
		if( true_size == 0 ) {
			fprintf(header,
"\n\
#define sizeof_%s(p) (sizeof(idl_ulong_int) + (p)->length * %d)\n",
				rep(type), bt->t_xsize);
		} else {
			fprintf(header,
"\n\
#define sizeof_%s(p) (((p)->length < %d) ? (sizeof(idl_ulong_int) + (p)->length * %d) : (sizeof(idl_ulong_int) + %d))\n",
				rep(type), true_size, bt->t_xsize, true_size * bt->t_xsize);
		}
	} else {
		/*
		 * The element type is variable-size, so define a function.
		 */
		fprintf(support1,
"\n\
static int\n\
sizeof_%s(p)\n\
\tregister %s *p;\n\
{\n\
\tregister int size = 4;\n\
\tidl_ulong_int length;\n\
\tregister int i;\n\
\n",
			rep(type), rep(type));

		if( true_size == 0 ) {
			fprintf(support1,
"\tlength = p->length;\n"
			);
		} else {
			fprintf(support1,
"\tlength = (p->length < %d) ? p->length : %d;\n",
			true_size, true_size);
		}

		fprintf(support1,
"\tfor (i = 0; i < length; i += 1)\n\
\t\tsize += sizeof_%s(%sp->sequence[i]);\n\
\treturn (size);\n\
}\n",
			rep(element_type), ref);
	}
	/*
	 * Define translation functions.
	 */
	xfn_header(EXTERNALIZE, type, type);
	xfn_header(INTERNALIZE, type, type);
	format =
"{\n\
\t_btype *bp;\n\
\tidl_ulong_int length;\n\
\tregister int i;\n\
\n";
	fprintf(support1, format, xfn(EXTERNALIZE, long_cardinal_type));
	fprintf(support2, format, xfn(INTERNALIZE, long_cardinal_type));
	/*
	 * Get the length of the sequence for externalize.
	 */
	if( true_size == 0 ) {
		fprintf(support1,
"\tlength = p->length;\n"
		);
	} else {
		fprintf(support1,
"\tlength = (p->length < %d) ? p->length : %d;\n",
			true_size, true_size);
	}
	/*
	 * Externalize/internalize the length itself.
	 */
	format =
"\tbp = buf + %s(&length, buf);\n";
	fprintf(support1, format, xfn(EXTERNALIZE, long_cardinal_type));
	fprintf(support2, format, xfn(INTERNALIZE, long_cardinal_type));
	/*
	 * Get the length of the sequence for internalize.
	 * In this case ignore true_size. We have to read what is
	 * on the net anyway, and the destination is allocated
	 * dynamically.
	 */
	if( true_size == 0 ) {
		fprintf(support2,
"\tp->length = length;\n"
		);
	} else {
		fprintf(support2,
"\tif ((p->length = length) > %d)\n\
\t\tnc_raise(ETNCSTB, ENCCTOOLSEQ, \"sequence too long\");\n",
			true_size);
	}
	/*
	 * The internalize function needs to allocate space
	 * for the sequence elements dynamically.
	 */
	/* only if noalloc not specified */
	if (allocation) {
	    fprintf(support2,
"\tp->sequence = (%s *)\n\
\t\tnc_allocate(length * sizeof(%s));\n",
		rep(type->t_basetype), rep(type->t_basetype));
	}
	format =
"\tfor (i = 0; i < length; i += 1)\n\
\t\tbp += %s(%sp->sequence[i], bp);\n\
\treturn (bp - buf);\n\
}\n";
	fprintf(support1, format, xfn(EXTERNALIZE, element_type), ref);
	fprintf(support2, format, xfn(INTERNALIZE, element_type), ref);
	return (type);
}


