/*@(#)sproc.c	1.2.1.2          5/8/89*/
/*
 * Creation 1986 by Inge B. Kristiansen.
 */
/*
 * 08/05/89 (phd+vfr) GLOBALs
 * xx.05.90 kk  Version 2 of NC/RPC
 * 06.11.90 kk  avoid generation of "n += 0;"
 *              corrected calculation of fixed size with global's
 * 06.12.90 kk  added static on local variables of type float and double
 *              in case of APOLLO10K. Otherwise they are allocated in
 *              registers and cannot be dereferenced.
 */

#include "sysconfig.h"
#include "np.h"

extern struct object *globallist;

/*
 * Generate the server header. The server binding is left for the
 * wrapup_module() procedure.
 */

server_header(module_name)
	char *module_name;
{
	fprintf(server,
"/*\n\
 * Server stubs for %s service.\n\
 * Generated by NC compiler version %d.%d.\n\
 */\n\
#include \"%s_support.c\"\n",
		module_name, NC_VERSION, NC_REVISION, module_name);

	fprintf(server,
"\n\
static int _%s_export;\n",
		module_name);
}


/*
 * Generate the server stub for a procedure call.
 */

server_procedure(proc_name, type, proc_number, random_id)
	char *proc_name;
	struct object *type;
	int proc_number;
	long random_id;
{
	list p, q;
	int nresults, fixed_size, variable_size;
	struct object *t, *bt, *result_type;
	char *result_name, *ref, *rtname;
	int a_mode, d_mode;

	/*
	 * Make sure there is at most one result returned.
	 */
	nresults = length(type->t_results);
	if (nresults == 1)
		/*
		 * Could be multiple names with one type.
		 */
		nresults = length(car(car(type->t_results)));
	if (nresults > 1) {
		error(SERIOUS, "procedures that return multiple results are not supported");
		return;
	}
	if (nresults) {
		result_name = ocar(car(car(type->t_results)))->o_name;
		result_type = ocdr(car(type->t_results));
		rtname = rep(result_type);
	} else {
		rtname = "void";
	}

	/*
	 * Server routine.
	 */
	if (nresults) {
		p = type->t_results;
		q = car(car(p));
		t = ocdr(car(p));
		ref = indamode(basetype(t), ocar(q)->o_addr);
	} else {
		ref = "";
	}

	fprintf(server,
"\nserver_%s(_bp, _result_base, _result_len)\n\
\t_btype *_bp;\n\
\t_btype **_result_base;\n\
\tint *_result_len;\n\
{\n\
\tregister int _n;\n",
		proc_name);
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		t = ocdr(car(p));
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
#if defined(APOLLO10K)
		    /* floats and doubles are put into registers if not static */
		    if( (t->t_id == T_FLOATING) || (t->t_id == T_LONG_FLOATING)  )
			fprintf(server, "\tstatic %s %s;\n", rep(t), ocar(q)->o_name);
		    else
#endif
			fprintf(server, "\t%s %s;\n", rep(t), ocar(q)->o_name);
		}
	}
	if (nresults) {
		p = type->t_results;
		q = car(car(p));
		t = ocdr(car(p));
		ref = indamode(basetype(t), ocar(q)->o_addr);
#if defined(APOLLO10K)
		/* floats and doubles are put into registers if not static */
		if( ((result_type->t_id == T_FLOATING) || (result_type->t_id == T_LONG_FLOATING)  )
			    && ( ocar(car(car(type->t_results)))->o_addr == A_VALUE ) )
			fprintf(server, "\tstatic %s %s%s;\n", rtname, ref, result_name);
		else
#endif
			fprintf(server, "\t%s %s%s;\n", rtname, ref, result_name);
	}
	fprintf(server, "\n");

	/*
	 * Begin critical section.
	 */
	server_begin_critical();

	/*
	 * Generate code to internalize the arguments.
	 */
	if (globallist != (struct object *)NULL) {
	    for (p = globallist->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		/* a_mode = genamode(ocar(q)->o_cmod, ocar(q)->o_addr); */
		/* d_mode = gendmode(ocar(q)->o_cmod, ocar(q)->o_dir); */
		t = ocdr(car(p));
		ref = refstr(basetype(t));
		if (ocar(q)->o_dir != D_OUT) {
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(server, "\t_bp += %s(%s%s, _bp);\n",
					xfn(INTERNALIZE, t), ref, ocar(q)->o_name);
		}
	    }
	}
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		/* a_mode = genamode(ocar(q)->o_cmod, ocar(q)->o_addr); */
		/* d_mode = gendmode(ocar(q)->o_cmod, ocar(q)->o_dir); */
		t = ocdr(car(p));
		ref = refstr(basetype(t));
		if (ocar(q)->o_dir != D_OUT) {
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(server, "\t_bp += %s(%s%s, _bp);\n",
					xfn(INTERNALIZE, t), ref, ocar(q)->o_name);
		}
	}
	/*
	 * Generate code to call the procedure.
	 */
	if (nresults)
		fprintf(server, "\t%s = %s(", result_name, proc_name);
	else
		fprintf(server, "\t%s(", proc_name);

	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		t = ocdr(car(p));
		ref = refamode(basetype(t), chamode(ocar(q)->o_addr));
		for (q = car(car(p)); q != NIL; q = cdr(q)) {
			fprintf(server, "%s%s", ref, ocar(q)->o_name);
			if (cdr(q) != NIL)
				fprintf(server, ", ");
		}
		if (cdr(p) != NIL)
			fprintf(server, ", ");
	}
	fprintf(server, ");\n");

	/*
	 * Determine the size of the result.
	 */
	if (nresults) {
		bt = basetype(result_type);
		if (bt->t_xsize == -1) {

		/* ref = refstr(basetype(result_type)); */
		q = car(car(type->t_results));
		ref = refamode(basetype(result_type), ocar(q)->o_addr);
		fprintf(server,
"\t_n = sizeof_%s(%s%s);\n",
			rtname, ref, result_name);
		} else {
			fprintf(server,
"\t_n = sizeof(%s);\n", rtname);
		}
	} else {
		fprintf(server,
"\t_n = 0;\n"
			);
	}

	/*
	 * Determine the size of the arguments.
	 * This is like the code in record_type().
	 */
	fixed_size = 0;
	variable_size = 0;
	if (globallist != (struct object *)NULL) {
	    for (p = globallist->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir == D_OUT || ocar(q)->o_dir == D_INOUT) {
		    bt = basetype(ocdr(car(p)));
		    if (bt->t_xsize == -1) {
			variable_size = 1;
		    } else {
			/* count multiple arguments of the same type */
			fixed_size += length(car(car(p))) * bt->t_xsize;
		    }
		}
	    }
	}
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		if (ocar(q)->o_dir == D_OUT || ocar(q)->o_dir == D_INOUT) {
			bt = basetype(ocdr(car(p)));
			if (bt->t_xsize == -1) {
				variable_size = 1;
			} else {
				/* count multiple arguments of the same type */
				fixed_size += length(car(car(p))) * bt->t_xsize;
			}
		}
	}
	if (!variable_size) {
		/*
		 * The argument list is fixed-size.
		 */
		if (fixed_size != 0) {
			fprintf(server, "\t_n += %d;\n", fixed_size);
		}
	} else {
		/*
		 * There are some variable-size arguments.
		 */
		fprintf(server,
"\t_n += %d;\n",
			fixed_size);
		if (globallist != (struct object *)NULL) {
		    for (p = globallist->t_args; p != NIL; p = cdr(p)) {
			q = car(car(p));
			if (ocar(q)->o_dir == D_OUT || ocar(q)->o_dir == D_INOUT) {
				t = ocdr(car(p));
				bt = basetype(t);
				if (bt->t_xsize != -1)
					continue;
				ref = refstr(bt);
				for (q = car(car(p)); q != NIL; q = cdr(q)) {
					fprintf(server,
"\t_n += sizeof_%s(%s%s);\n",
						rep(t), ref, ocar(q)->o_name);
				}
			}
		    }
		}
		for (p = type->t_args; p != NIL; p = cdr(p)) {
			q = car(car(p));
			if (ocar(q)->o_dir == D_OUT || ocar(q)->o_dir == D_INOUT) {
				t = ocdr(car(p));
				bt = basetype(t);
				if (bt->t_xsize != -1)
					continue;
				ref = refstr(bt);
				for (q = car(car(p)); q != NIL; q = cdr(q)) {
					fprintf(server,
"\t_n += sizeof_%s(%s%s);\n",
						rep(t), ref, ocar(q)->o_name);
				}
			}
		}
	}

	/*
	 * Generate code to externalize the result.
	 */
	if (nresults) {
		q = car(car(type->t_results));
		/* a_mode = genamode(ocar(q)->o_cmod, ocar(q)->o_addr); */
		/* d_mode = gendmode(ocar(q)->o_cmod, ocar(q)->o_dir); */
		/* ref = refstr(basetype(result_type)); */
		ref = refamode(basetype(result_type), ocar(q)->o_addr);
		fprintf(server,
"\t*_result_len = RETURN_HEADER_BYTES + _n;\n\
\t_bp = nc_allocate(RETURN_HEADER_BYTES + _n);\n\
\t*_result_base = _bp;\n\
\t_bp = &_bp[RETURN_HEADER_BYTES];\n\
\t_bp += %s(%s%s, _bp);\n",
			xfn(EXTERNALIZE, result_type), ref, result_name);
	} else {
		fprintf(server,
"\t*_result_len = RETURN_HEADER_BYTES + _n;\n\
\t_bp = nc_allocate(RETURN_HEADER_BYTES + _n);\n\
\t*_result_base = _bp;\n\
\t_bp = &_bp[RETURN_HEADER_BYTES];\n"
			);
	}

	/*
	 * Generate code to externalize the arguments.
	 */
	if (globallist != (struct object *)NULL) {
	    for (p = globallist->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		t = ocdr(car(p));
		ref = refstr(basetype(t));
		if (ocar(q)->o_dir == D_OUT || ocar(q)->o_dir == D_INOUT) {
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(server, "\t_bp += %s(%s%s, _bp);\n",
					xfn(EXTERNALIZE, t),
					ref, ocar(q)->o_name);
		}
	    }
	}
	for (p = type->t_args; p != NIL; p = cdr(p)) {
		q = car(car(p));
		/* a_mode = genamode(ocar(q)->o_cmod, ocar(q)->o_addr); */
		/* d_mode = gendmode(ocar(q)->o_cmod, ocar(q)->o_dir); */
		t = ocdr(car(p));
		ref = refstr(basetype(t));
		if (ocar(q)->o_dir == D_OUT || ocar(q)->o_dir == D_INOUT) {
			for (q = car(car(p)); q != NIL; q = cdr(q))
				fprintf(server, "\t_bp += %s(%s%s, _bp);\n",
					xfn(EXTERNALIZE, t),
					ref, ocar(q)->o_name);
		}
	}

	/*
	 * End critical section.
	 */
	server_end_critical();

	fprintf(server,
"}\n"
		);
}


/*
 * Generate export function for server.
 */

server_binding()
{
	/* Generate the procedure call tables.                          */
	server_table( module_name );

	/* Generate the export function.                                */
	server_export( "export_service", module_name, "" );
}
