/*
 * Copyright (C) 1986 by Inge B. Kristiansen
 * All rights reserved.
 */

#include "sysconfig.h"
#include "np.h"


/*
 * Global variables maintained by this module. Array of parameter names
 * and types to be used for export/import. The table is terminated by a
 * name which is zero.
 */

struct ie_object ie_table[] = {
	{ IE_SERVICE,           "_service",             T_STRING,       0,              A_VALUE },
	{ IE_HOSTNAME,          "_host",                T_STRING,       0,              A_VALUE },
	{ IE_PORTNUMBER,        "_port",                T_CARDINAL,     0,              A_VALUE },
	{ IE_PROTECTION,        "_protection",          T_CARDINAL,     0,              A_VALUE },
	{ IE_EXECUTION,         "_execution",           T_CARDINAL,     0,              A_VALUE },
	{ IE_LOAD,              "_load",                T_CARDINAL,     0,              A_VALUE },
	{ IE_FILENAME,          "_file",                T_STRING,       0,              A_VALUE },
	{ IE_DIRECTORY,         "_directory",           T_CARDINAL,     0,              A_VALUE },
	{ IE_DIRHOST,           "_directory_host",      T_STRING,       0,              A_VALUE },
	{ IE_DIRPORT,           "_directory_port",      T_CARDINAL,     0,              A_VALUE },
	{ IE_PROTOCOL,          "_protocol",            T_COMPOSITE,    "protocol",     A_REF   },
	{ IE_LOADPROTOCOL,      "_loadprotocol",        T_COMPOSITE,    "protocol",     A_REF   },
	{ IE_DIRPROTOCOL,       "_dirprotocol",         T_COMPOSITE,    "protocol",     A_REF   },
	{ IE_TOABSOLUTE,        "_toabsolute",          T_CARDINAL,     0,              A_VALUE },
	{ IE_TORELATIVE,        "_torelative",          T_CARDINAL,     0,              A_VALUE },
	{ IE_TORELATIVE,        0,                      T_CARDINAL,     0,              A_VALUE }
};

/*
 * Get the connection information corresponiding to a given type.
 * The code is given as a parameter.
 */

struct object *
get_predef( type )
	enum typeid type;
{
	struct object **t;
	int i;

	/* Find the type in the table.                                  */
	for( i = 0, t = &boolean_type; i < sizeof(struct pretypes)/sizeof(struct object *); ++i, ++t ) {
		if( (*t)->t_id == type ) {
			return( *t );
		}
	}

	/* Did not find the type in the table.                          */
	return( 0 );
}


/*
 * Get the connection information corresponiding to a given type.
 * The code is given as a parameter.
 */

struct ie_object *
get_connection( parameter )
	enum ie_conn parameter;
{
	struct ie_object *ie;

	/* Find the parameter in the table.                             */
	for( ie = ie_table; ie->ie_name != 0; ++ie ) {
		if( ie->ie_type == parameter ) {
			return( ie );
		}
	}

	/* Did not find the parameter.                                  */
	return( 0 );
}


/*
 * Make the connection tree to be used to generate the the export/import
 * parameter lists. This procedure will return an argument tree
 * corresponding to the parameter specified as input.
 */

list
make_connection( parameter )
	enum ie_conn parameter;
{
	struct object *o;
	struct object *t;
	struct ie_object *ie;

	/* Find the parameter in the table.                             */
	if( (ie = get_connection( parameter )) == (struct ie_object *)0 ) {
		error(FATAL, "illegal parameter class in make_connection");
		/* NOTREACHED */
	}

	/* Construct the type to be used.                               */
	if( ie->ie_typeid != T_COMPOSITE ) {
		if( (t = get_predef( ie->ie_typeid )) == (struct object *)0 ) {
			error(FATAL, "cannot find predefined type in make_connection");
			/* NOTREACHED */
		}
	} else {
		t = make_symbol( ie->ie_tname );
		if( lookup( values, t ) == 0 ) {
			declare( &values, t, unspecified_type );
		}
	}

	/* Construct the symbol to be used.                             */
	o = make_symbol( ie->ie_name );

	/* Declare the parameter in the connection symbol table.        */
	declare(&connections, o, t);

	/* Return the complete argument on correct form.                */
	return( cons( (list)make_arg( cons( (list)o, NIL ),
	    make_transmission(ie->ie_amode, D_IN, F_USER ) ), (list)type_conv( t ) ) );
}
