/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

#include "sysconfig.h"
#include <signal.h>
#include "np.h"
#include <string.h>
#include <strings.h>

char yyfilename[MAXSTR];
char infilename[MAXSTR];
char *input_file;
char header_file[MAXSTR];
FILE *header;
char support_file1[MAXSTR];
FILE *support1;
char support_file2[MAXSTR];
FILE *support2;
char client_file[MAXSTR];
FILE *client;
char server_file[MAXSTR];
FILE *server;
list values, types, procedures, dependencies, connections;
int recursive_flag, errs;

/*
 * Main module.
 */
void nterrupt(x)
	int x;
{
	errs = 1;
	goodbye();
}

main(argc, argv)
	int argc;
	char **argv;
{
	argc = np_getargs( argc, argv );
	if (argc != 2) {
		fprintf(stderr, "Usage: %s filename\n", argv[0]);
		exit(1);
	}
	strcpy(yyfilename, argv[1]);
	input_file = yyfilename;
	if (freopen(input_file, "r", stdin) == NULL) {
		perror(input_file);
		exit(1);
	}
	tempname(header_file);
	tempname(client_file);
	tempname(server_file);
	tempname(support_file1);
	tempname(support_file2);
	signal(SIGINT, nterrupt);
	if ((header = fopen(header_file, "w")) == NULL) {
		perror(header_file);
		goodbye();
	}
	if ((client = fopen(client_file, "w")) == NULL) {
		perror(client_file);
		(void) fclose(header);
		goodbye();
	}
	if ((server = fopen(server_file, "w")) == NULL) {
		perror(server_file);
		(void) fclose(header);
		(void) fclose(client);
		goodbye();
	}
	if ((support1 = fopen(support_file1, "w")) == NULL) {
		perror(support_file1);
		(void) fclose(header);
		(void) fclose(client);
		(void) fclose(server);
		goodbye();
	}
	if ((support2 = fopen(support_file2, "w")) == NULL) {
		perror(support_file2);
		(void) fclose(header);
		(void) fclose(client);
		(void) fclose(server);
		(void) fclose(support1);
		goodbye();
	}
	setup_predefs();
	client_initconnection();
	server_initconnection();
	(void) yyparse();
	(void) fclose(header);
	(void) fclose(client);
	(void) fclose(server);
	if (errs == 0) {
		register int c;

		if (procedures != NIL) {
			(void) freopen(support_file2, "r", support2);
			while ((c = getc(support2)) != EOF)
				(void) putc(c, support1);
		}
		(void) fclose(support1);
		(void) fclose(support2);
		(void) unlink(support_file2);
		if (procedures != NIL)
			rename_file(support_file1, module_name, "_support.c");
		else
			(void) unlink(support_file1);
		if (procedures != NIL) {
			if( np_gen_client ) {
				rename_file(client_file, module_name, "_client.c");
			}
			if( np_gen_server ) {
				rename_file(server_file, module_name, "_server.c");
			}
		} else {
			(void) unlink(client_file);
			(void) unlink(server_file);
		}
		rename_file(header_file, module_name, "_defs.h");
	}
	goodbye();
}

/*
 * Rename the source file to be dest.suffix
 * Update the buffer containing source with dest.suffix
 */
rename_file(source, dest, suffix)
	char *source, *dest, *suffix;
{
	char newname[MAXSTR];

	(void) sprintf(newname, "%s%s", dest, suffix);
	if (rename(source, newname) == -1) {
		perror(newname);
	} else {
		strcpy(source, newname);
	}
}

/* VARARGS2 */
error(how_bad, s, args)
	enum severity how_bad;
	char *s;
{
	extern int yylineno;

	fprintf(stderr, "%s: %d: ", input_file, yylineno);
	if (how_bad == WARNING)
		fprintf(stderr, "Warning: ");
	/*_doprnt(s, &args, stderr);*/
	(void) putc('\n', stderr);
	if (how_bad == SERIOUS)
		errs = 1;
	if (how_bad == FATAL) {
		errs = 1;
		goodbye();
	}
}

yyerror(s)
	char *s;
{
	error(FATAL, s);
}

np_error( prog_name, s, x, y ) char *prog_name; char *s;
{
	fprintf( stderr, prog_name );
	fprintf( stderr, ": 0: " );
	fprintf( stderr, s, x, y );
	fprintf( stderr, "\n" );
}

/*
 * Check if a flag is set and display an error message if it is.
 */
check_flag( flag, message )
	short flag;
	char *message;
{
	if( flag ) {
		error(SERIOUS, "%s attribute is already updated", message);
	}
	return( flag );
}

goodbye()
{
	if (errs) {
		(void) unlink(header_file);
		(void) unlink(client_file);
		(void) unlink(server_file);
		(void) unlink(support_file1);
		(void) unlink(support_file2);
	} else {
		if( !np_gen_server ) {
			(void) unlink(server_file);
		}
		if( !np_gen_client ) {
			(void) unlink(client_file);
		}
	}
	exit(errs);
}

tempname(buf)
	char *buf;
{
	static int n = 0;

	(void) sprintf(buf, "tmp%d.%d", n, getpid());
	n += 1;
}
