/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

#include "sysconfig.h"
#include "np.h"


/*
 * Array type optimization procedure. The procedure will try to locate
 * an array of the same type and size. If successful, these conversion
 * procedures will be used.
 */

struct object *
optim_array( symlist, type, size, element_type )
	list symlist;
	struct object *type, *size, *element_type;
{
	int true_size;
	struct object *bt, *t;
	list p, q;

	/* Compute the true size of the array from size object.         */
	true_size = true_value( size );

	/* Go through the symbol table and try to find an element which */
	/* can be reused.                                               */
	for (p = symlist; p != NIL; p = cdr(p)) {
		q = car(p);
		t = ocdr(q);

		/* If the base type and the size is the same, we can    */
		/* reuse the code.                                      */
		if( (t->o_class == O_TYPE) &&
          	    (element_type->t_id == ((bt = basetype(t))->t_id)) &&
		    (true_size == true_value( bt->t_size )) ) {
			return( t );
		}
	}

	/* If it is an array of prefefined types, it has to be appended */
	/* to the symbol table.                                         */
	if( element_type->t_constr == C_PREDEF ) {
		declare( &values, make_symbol( gensym() ), type );
	}

	return (0);
}
