/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

#include "sysconfig.h"
#include "np.h"


/*
 * Get the server type bits.
 */

unsigned short
make_stype( sp )
	char **sp;
{
	unsigned short permission = 0;

	/* Find the server type from the string.                        */
	switch( **sp ) {
		case 'd':
			permission = NP_F_SDIR;
			break;

		case '-':
			permission = NP_F_SNORM;
			break;
	}

	/* Return the found type to the caller.                         */
	(*sp)++;
	return( permission );
}

/*
 * Get the exclusion mode bits.
 */

unsigned short
make_exmode( sp )
	char **sp;
{
	unsigned short permission = 0;

	/* Find the exclusion mode from the string.                     */
	switch( **sp ) {
		case 'm':
			permission = NP_F_EMUTUAL;
			break;

		case 'n':
			permission = NP_F_EMAXCONN;
			break;

		case 'x':
			permission = NP_F_EEXTERN;
			break;

		case '-':
			permission = NP_F_ENORM;
			break;
	}

	/* Return the found mode to the caller.                         */
	(*sp)++;
	return( permission );
}

/*
 * Get the permission flags.
 */

unsigned short
make_pflags( sp )
	char **sp;
{
	unsigned short permission = 0;
	int i, j;

	/* Loop through user, group and other permission.               */
	for( i = 0; i < 3; ++i ) {
		/* Loop through read, write and unused permission.      */
		for( j = 0; j < 3; ++j, ++(*sp) ) {
			switch( **sp ) {
				case 'r':
					permission |= (NP_F_PUREAD >> (3*i) );
					break;

				case 'w':
					permission |= (NP_F_PUWRITE >> (3*i) );
					break;
			}
		}
	}

	/* Return the found type to the caller.                         */
	return( permission );
}

/*
 * Parse the permission string.
 */

int
make_permission( s )
	char *s;
{
	unsigned short permission = 0;

	/* Get the server type.                                         */
	permission |= make_stype( &s );

	/* Get the exclusion mode.                                      */
	permission |= make_exmode( &s );

	/* Get the permission flag bits.                                */
	permission |= make_pflags( &s );

	/* Return the complete permission bits.                         */
	return( (int)permission );
}

