/* errtype.h */
/*
 * Error type definitions
 *
 * Error types are set by runtime libraries in a global variable errtype
 * together with the error number set in errno and error string pointed to
 * by errstring.
 * To facilitate support of runtime libraries, errfile and errline are set
 * to identify the place where error occured.
 */
#ifndef ERRTYPE
#define ERRTYPE

#include "sysconfig.h"
#if defined(OLDVMS)
#include errno
#else
#include <errno.h>
#endif

extern int errtype;
extern char *errstring;
extern int errline;
extern char *errfile;
#ifdef _NO_PROTO
extern void nc_perror();
#else /* _NO_PROTO */
extern void nc_perror(char *s);
#endif /* _NO_PROTO */

/* operating systems (1-9) */
#define ETSYS           0
#define ETUNIX          (ETSYS+1)       /* Unix system call errors */
#define ETOS9           (ETSYS+2)       /* OS9 system errors */
#define ETVMS           (ETSYS+3)       /* VMS system errors */
#define ETSYNTR         (ETSYS+4)       /* SYNTRON (sic !!!) monitor errors */

#define ETOFSREM        1000            /* error type offset for remote errors */

/* communication protocols and tools (10-50) */
#define ETCOM           10
#define ETNCRT          (ETCOM+0)       /* NC/RPC runtime */
#define ETNCSTB         (ETCOM+1)       /* NC/RPC stub */
#define ETTCPIP         (ETCOM+2)       /* socket/tcpip interface in user code */
#define ETPROT          (ETCOM+3)       /* protocol interface (for version 1 of nc/RPC ) */
#define ETRX            (ETCOM+4)       /* errors from RX protocol */
#define ETMIL           (ETCOM+5)       /* errors from MIL1553 protocol */
#define ETPCIP          (ETCOM+6)       /* errors from PCIP library */
#define ETNODAL         (ETCOM+7)       /* NODAL errors */
#define ETNCNSTB        (ETCOM+8)       /* NC/RPC NODAL stub */

#define ETAPPL          50              /* application level */

#ifndef M_KERNEL
/* set error variables and return -1 */
#define RETURNERR(x,y,z) { errfile = __FILE__; errline = __LINE__; \
	errno = x; \
	errtype = y; \
	errstring = z; \
	return (-1); }
/* set error variables and return NULL */
#define RETURNERRP(ret,x,y,z) { errfile = __FILE__; errline = __LINE__; \
	errno = x; \
	errtype = y; \
	errstring = z; \
	return (ret); }
/* set error variables without return */
#define SETERR(x,y,z) { errfile = __FILE__; errline = __LINE__; \
	errno = x; \
	errtype = y; \
	errstring = z; \
	}
#define CLEARERR { \
	errno = 0; \
	errtype = 0; \
	errstring = (char *)0; \
	}
#define PRINTERR(x) { \
	fprintf(stderr, "%s: error %d of type %d received in (file %s,line %d) \n", \
		x, errno, errtype, errfile, errline); \
	fflush(stderr); \
	}
#endif

#endif /* ERRTYPE */
