/* %W% %E% */
/*
 * Creation 1986 by Inge B. Kristiansen.
 * Modifications:
 * xx.05.90 kk  Version 2.0
 *              Complete review and rewrite. Separated into user (ncrpc.h)
 *              and system include files.
 * CS - 07.09.2000 : cleand up include section
 *                   removed some unsed defintions
 */


#ifndef NCC_H
#define NCC_H
#include "sysconfig.h"
#include <errno.h>
/*
 * DCE (Distributed Computing Environment) RPC is the anticipated remote
 * procedure call standard. IDL is the Interface Description Language
 * for this standard.
 * These types are DCE types for forward compatibility. Use these types
 * whenever possible !!!!
 */
#if !defined(idl_char)
#define idl_boolean        unsigned char
#define idl_false          false
#define idl_true           true
#define idl_byte           unsigned char
#define idl_char           unsigned char
#define idl_small_int      char
#define idl_usmall_int     unsigned char
#define idl_short_int      short int
#define idl_ushort_int     unsigned short int
#define idl_long_int       nc_32int
#define idl_ulong_int      nc_32uint

#if defined(NC_NATIVE64)
#define idl_hyper_int      nc_64int
#define idl_uhyper_int     nc_64uint
#else /* NC_NATIVE64 */
#if defined(LOW_ENDER)
struct idl_hyper_int_rep   {idl_ulong_int low; idl_long_int high;};
#define idl_hyper_int      struct idl_hyper_int_rep
struct idl_uhyper_int_rep  {idl_ulong_int low; idl_ulong_int high;};
#define idl_uhyper_int     struct idl_uhyper_int_rep
#endif
#if defined(HIGH_ENDER)
struct idl_hyper_int_rep   {idl_long_int high; idl_ulong_int low;};
#define idl_hyper_int      struct idl_hyper_int_rep
struct idl_uhyper_int_rep  {idl_ulong_int high; idl_ulong_int low;};
#define idl_uhyper_int     struct idl_uhyper_int_rep
#endif
#endif /* NC_NATIVE64 */

#define idl_short_float    float
#define idl_long_float     double
#endif /* idl_char */

#define externalize_idl_boolean(p, buf) move_byte(p, buf)
#define internalize_idl_boolean(p, buf) move_byte(buf, p)
#define externalize_idl_byte(p, buf) move_byte(p, buf)
#define internalize_idl_byte(p, buf) move_byte(buf, p)
#define externalize_idl_char(p, buf) move_byte(p, buf)
#define internalize_idl_char(p, buf) move_byte(buf, p)

#define externalize_idl_small_int(p, buf) move_byte(p, buf)
#define internalize_idl_small_int(p, buf) move_byte(buf, p)
#define externalize_idl_short_int(p, buf) move_short(p, buf)
#define internalize_idl_short_int(p, buf) move_short(buf, p)
#define externalize_idl_long_int(p, buf) move_long(p, buf)
#define internalize_idl_long_int(p, buf) move_long(buf, p)
#define externalize_idl_hyper_int(p, buf) move_hyper(p, buf)
#define internalize_idl_hyper_int(p, buf) move_hyper(buf, p)

#define externalize_idl_usmall_int(p, buf) move_byte(p, buf)
#define internalize_idl_usmall_int(p, buf) move_byte(buf, p)
#define externalize_idl_ushort_int(p, buf) move_short(p, buf)
#define internalize_idl_ushort_int(p, buf) move_short(buf, p)
#define externalize_idl_ulong_int(p, buf) move_long(p, buf)
#define internalize_idl_ulong_int(p, buf) move_long(buf, p)
#define externalize_idl_uhyper_int(p, buf) move_hyper(p, buf)
#define internalize_idl_uhyper_int(p, buf) move_hyper(buf, p)
#if defined(VAX)
#define externalize_idl_short_float(p, buf) externalize_floating(p, buf)
#define internalize_idl_short_float(p, buf) internalize_floating(p, buf)
#define externalize_idl_long_float(p, buf) externalize_long_floating(p, buf)
#define internalize_idl_long_float(p, buf) internalize_long_floating(p, buf)
#else
#define externalize_idl_short_float(p, buf) move_long(p, buf)
#define internalize_idl_short_float(p, buf) move_long(buf, p)
#define externalize_idl_long_float(p, buf) move_hyper(p, buf)
#define internalize_idl_long_float(p, buf) move_hyper(buf, p)
#endif


/*
 * Standard protocol routine structure. A new protocol must conform
 * to this structure to be incorporated in the scheme.
 */

typedef struct protocol {
	short p_flags;
	char  *p_name;
	short p_unused[5];
	int   (*p_socket)();
	int   (*p_select)();
	int   (*p_open)();
	int   (*p_iocntl)();
	int   (*p_write)();
	int   (*p_read)();
	int   (*p_flush)();
	int   (*p_close)();
	int   (*p_kill)();
	int   (*p_abort)();
	int   (*p_cleanup)();
	int   (*p_getaddr)();
} protocol;

#define NC_P_FSERVER    01
#define NC_P_FCLIENT    02
#define NC_P_FINIT	04

/*
 * Protocol definitions.
 */

extern  protocol p_null;
extern  protocol p_rx;
extern  protocol p_rtcp;

#define NC_P_NULL       (&p_null)
#define NC_P_FIRST      (&p_null)
#define NC_P_RX         (&p_rx)
#define NC_P_RTCP       (&p_rtcp)

/*
 * Server protection modes.
 */
#define NC_PR_PUBLIC    0
#define NC_PR_PRIVATE   1
#define NC_PR_RUSEROK   2

#define NC_PRX_PUBLIC    0
#define NC_PRX_TPUBLIC   1
#define NC_PRX_PRIVATE   2
#define NC_PRX_TPRIVATE  3

/*
 * Server execution modes.
 */
#define NC_EX_ANY       0
#define NC_EX_MOTHER    1
#define NC_EX_DETACH    2
#define NC_EX_CHILD     NC_EX_DETACH
#define NC_EX_ORPHAN    NC_EX_DETACH
#define NC_EX_MASTER    3

/*
 * Server scheduling modes.
 */
#define NC_SC_RROBIN    0
#define NC_SC_AGE       1

/*
 * Message types.
 */

#define NCC_CALL        1               /* Call type message */
#define NCC_RETURN      2               /* Return type message */
#define NCC_MESSAGE     3               /* Message type message */
#define NCC_NOREPLY     4               /* call message with no reply expected */
#define NCC_UNRELIABLE  5               /* unreliable call message (with no reply expected) */
					/* can be used for broadcasts                       */
#define NCMS_PORT       800             /* master server port */


/*
 * Nodal message types used in stubs (must match LEP Nodal message types)
 */
#define NCN_EXEC        5
#define NCN_READ        1
#define NCN_WRITE       2

/*
 * Definitions for the remote nc library (ncr).
 */
#define NCR_ARE_YOU_OK  1
#define NCR_EXIT        2
#define NCR_EXIT_CODE   11


/*
 * request type for open
 */
#define NCC_SEND	9
#define NCC_RECV        10


/*
 * Timeout structure.
 * This timeout structure is kept only for compatibility reasons.
 * It is used only to pass the timeout absolute ( in tm_c_l) and 
 * relative (in tm_nr_l) in the rtcp_protocol to the rtcp_socket.
 * See also nc/common/timeout.c nc_tmformat(cd).
 */

/* definitions taken from rudp.h */
typedef struct timeouts {
	unsigned short tm_c_t;          /* connection timer */
	unsigned short tm_c_l;          /* connection lifetime */
	unsigned short tm_r_t;          /* retransmit timer */
	unsigned short tm_rrq_t;        /* retransmit request timer */
	unsigned short tm_r_n;          /* number of retransmissions */
	unsigned short tm_nr_t;         /* no response timer */
	unsigned short tm_n_t;          /* network timer */
} timeouts;

#define IO_READ         0x1             /* read operation is legal */
#define IO_WRITE        0x2             /* write operation is legal */
#define _IOCCD(x,y)     ((x<<8)|y)      /* macro to build a command */
#define _NC_IOCSETTMOUT       9       /* set timeout structure */
#define NC_IOCSETTMOUT        _IOCCD('r', _NC_IOCSETTMOUT)

extern char nc_version;
extern char nc_revision;

/* End of the common part for protocols */

/*
 * Used to raise exception in the stub
 */
extern int errtype;
extern char * errstring;

/*
 * Address specification.
 */
typedef struct nc_address {
	char *nc_host;
	int  nc_port;
} nc_address;
typedef int sr_init;

/*
 * Mode control flag area definition.
 */
#define NC_F_STYPE      0170000         /* Type of server */
#define NC_F_SNORM      0010000         /* - Normal */
#define NC_F_SDIR       0020000         /* - Directory */

#define NC_F_ETYPE      0007000         /* Exclusion type */
#define NC_F_ENORM      0001000         /* - Normal */
#define NC_F_EMUTUAL    0002000         /* - Mutual exclusion */
#define NC_F_EEXTERN    0003000         /* - External */
#define NC_F_EMAXCONN   0004000         /* - Max number of connections */

#define NC_F_PUREAD     0000100         /* User read permission */
#define NC_F_PUWRITE    0000200         /* User write permission */

#define NC_F_PGREAD     0000010         /* Group read permission */
#define NC_F_PGWRITE    0000020         /* Group write permission */

#define NC_F_POREAD     0000001         /* Other read permission */
#define NC_F_POWRITE    0000002         /* Other write permission */

/*
 * Protocol headers. Should be multiply of 4 to optimise for 32 bit architectures
 */
#define CALL_HEADER_BYTES 28
#define RETURN_HEADER_BYTES 16

/* Nodal headers and size definitions */
#define CALL_NODAL_BYTES 24
#define RETURN_NODAL_BYTES 24
#define STRING_ELEMENT_HEADER 4
#define WORD_SIZE 2


/*
 * return/exit definitions for nc_serverloop() after processing the call
 */
#define NC_EX_NULL      0
#define NC_EX_PARENT    0xff
#define NC_EX_SPAWNED   0xff00

/* return conditions for mother-mode servers */
#define NC_EX_NEVER     (0)             /* never exits */
#define NC_EX_RETNOW    (1)             /* return from serverloop */
#define NC_EX_RETNXT    (1<<1)          /* set to NC_EX_RETNOW */
#define NC_EX_RETTMO    (1<<2)          /* return after timeout */

/* return conditions for spawned servers */
#define NC_EX_ONFORK    (1<<8)          /* exit after fork (little use) */
#define NC_EX_NOW       (1<<8)          /* exit now (after processing call) */
#define NC_EX_EXTNXT    (1<<9)          /* process one call after fork and exit */
#define NC_EX_TIMEOUT   (1<<10)         /* exit after timeout (normally use this) */
#define NC_EX_RWAIT     (1<<11)         /* bacward comp. sets NC_EX_TIMEOUT */
#define NC_EX_CUSTOM    (1<<12)         /* return to user code after timeout */




/*
 * Directory server registration modes.
 */
#define NC_D_NODIRECTORY 0
#define NC_D_DIRECTORY   1
#define NC_D_REGION      2

/*
 * Loader server file load modes.
 */
#define NC_L_NOLOAD     0
#define NC_L_PERMANENT  1
#define NC_L_TEMPORARY  2

/*
 * Empty objects of various types.
 */
#define NC_E_STRING     (char *)0
#define NC_E_INT        0
#define NC_E_LONG       0L

/*
 * Exception handling.
 */

/*
 * Define the layout of an exception handler. This is inserted in the
 * source at the appropriate place.
 */

extern  int nc_setjmp;

#define on(condition) \
	if( !( (nc_setjmp = setjmp( nc_allocexception() )) && (condition) ) ) { \
		if( nc_setjmp != 0 ) rnext( 0, 0, (char *)0 ); \
	} else

/*
 * Define the appropriate predefined condition codes.
 */

#define noexceptions    0
#define allexceptions   1

/*
 * Define the data structure holding the exception information.
 */

typedef struct nct_exception {
#ifdef SILLYERRNO      /* on some systems errno is #defined as *_errno() */
int     errno;
#else
int     errorno;
#endif
int     errtype;
char    *errmsg;
} nct_exception;

/*
 * Server statistics
 */

#define TYPES 6 /* Location 0 : Unused */
                /* Location 1 : NCC_CALL */
                /* Location 2 : NCC_RETURN */
                /* Location 3 : NCC_MESSAGE */
                /* Location 4 : NCC_NOREPLY */
                /* Location 5 : NCC_UNRELIABLE */

typedef idl_long_int mtype_array[TYPES]; /* One entry for each type of call */

typedef struct nc_stat { idl_short_int import;
                         idl_short_int export;
                         idl_short_int forks;
                         idl_short_int dummy;
                         idl_long_int cl_calls;
                         idl_long_int cl_sendlen;
                         idl_long_int cl_receivelen;
                         mtype_array call_type;
                         idl_long_int sr_receivelen;
                         idl_long_int sr_sendlen;} nc_statistics_t;

extern nc_statistics_t nc_statistics;


/*
 * Error numbers.
 */
#define ENCERRORS       0       /* begin of NC errors */
#define ENCMAXSERVICES  (ENCERRORS+1)     /* Maximum number of service per process */
#define ENCNOSERVICE    (ENCERRORS+2)     /* No service of this type */
#define ENCCBADSR       (ENCERRORS+3)     /* Bad service */
#define ENCCBADPR       (ENCERRORS+4)     /* Bad procedure number */
#define ENCCBADID       (ENCERRORS+5)     /* Bad procedure (random) id */
#define ENCCNOPROT      (ENCERRORS+6)     /* No such protocol available */
#define ENCCALLOC       (ENCERRORS+7)     /* Memory allocation failed */
#define ENCCINCSEND     (ENCERRORS+8)     /* Buffer incompletely sent */
#define ENCCINCRECV     (ENCERRORS+9)     /* Buffer incompletely received */
#define ENCCTOOLSEQ     (ENCERRORS+10)     /* Too long sequence */
#define ENCCEXPORT      (ENCERRORS+11)     /* Service is already exported */
#define ENCCTYPE        (ENCERRORS+12)     /* Illegal NC message type */
#define ENCCACCES       (ENCERRORS+13)     /* Permission denied */
#define ENCCNOMASTER    (ENCERRORS+14)     /* no reply from the master server */
#define ENCCNOREPL      (ENCERRORS+15)     /* timeout on reply for asynch call */
#define ENCCEXECUTION   (ENCERRORS+16)     /* Execution type violation */
#define ENCCPROTECTION  (ENCERRORS+17)     /* Protection violation */
#define ENCCUNSUPP      (ENCCERRORS+18)         /* Unsupported feature */

#define ENCNUNEXDTYPE   (ENCERRORS+19)     /* Unexpected Nodal data type */
#define ENCNILLDELIM    (ENCERRORS+20)     /* Illegal Nodal delimiter (should be -1) */


#define ENCLAST         (ENCERRORS+20)  /* last NC error */

/*
 * Low-level byte moving for one byte only
 */
#define move_byte(a, b) (*(char *)(b) = *(char *)(a), 1)

/*
 * Low-level byte moving, with byte-swapping.
 * Use these definitions for the VAX, Intel8086 and other low-enders.
 */
#if defined(LOW_ENDER)
#define move_short(a, b) (*(char *)(b) = *((char *)(a)+1), *((char *)(b)+1) = *(char *)(a), 2)
#define move_long(a, b) (*(char *)(b) = *((char *)(a)+3), \
 *((char *)(b)+1) = *((char *)(a)+2), \
 *((char *)(b)+2) = *((char *)(a)+1), \
 *((char *)(b)+3) = *(char *)(a), 4)



#define move_int(a, b) move_long(a, b)


#define move_hyper(a, b) (*(char*)(b) = *((char*)(a)+7), \
 *((char*)(b)+1) = *((char*)(a)+6), \
 *((char*)(b)+2) = *((char*)(a)+5), \
 *((char*)(b)+3) = *((char*)(a)+4), \
 *((char*)(b)+4) = *((char*)(a)+3), \
 *((char*)(b)+5) = *((char*)(a)+2), \
 *((char*)(b)+6) = *((char*)(a)+1), \
 *((char*)(b)+7) = *(char*)(a), 8)
#endif /* LOW_ENDER */


#if defined(HIGH_ENDER)
/*
 * Low-level byte moving, without byte-swapping for aligning machines.
 * Use these definitions for the 68000, IBMRT and other high-enders.
 */
#define move_short(a, b) (*(char *)(b) = *(char *)(a), *((char *)(b)+1) = *((char *)(a)+1), 2)
#define move_long(a, b) (*(char *)(b) = *(char *)(a),\
*((char *)(b)+1) = *((char *)(a)+1),\
*((char *)(b)+2) = *((char *)(a)+2),\
*((char *)(b)+3) = *((char *)(a)+3), 4)
#define move_int(a, b) move_long(a, b)

#define move_hyper(a, b) (*(char *)(b) = *(char *)(a),\
*((char *)(b)+1) = *((char *)(a)+1),\
*((char *)(b)+2) = *((char *)(a)+2),\
*((char *)(b)+3) = *((char *)(a)+3),\
*((char *)(b)+4) = *((char *)(a)+4),\
*((char *)(b)+5) = *((char *)(a)+5),\
*((char *)(b)+6) = *((char *)(a)+6),\
*((char *)(b)+7) = *((char *)(a)+7), 8)
#endif /* HIGH_ENDER */

/*
 * Low-level word moving, without byte-swapping for aligned words.
 * Use these definitions for new aligned moving in high-enders.
 */
#define move_idl_short(a, b) (*(short *)(b) = *(short *)(a), 2)
#define move_idl_long(a, b) (*(long *)(b) = *(long *)(a), 4)
#define move_idl_hyper(a, b) (*(long *)(b) = *(long *)(a), (*((long *)(b)+4) = *((long *)(a)+4), 8)

/*
 * Buffer type.
 */
typedef char _btype;

/*
 * Predefined types.
 */
typedef char boolean;
#define externalize_boolean(p, buf) move_byte(p, buf)
#define internalize_boolean(p, buf) move_byte(buf, p)

typedef unsigned short cardinal;
typedef unsigned short short_cardinal;
#define externalize_cardinal(p, buf) move_short(p, buf)
#define internalize_cardinal(p, buf) move_short(buf, p)

typedef unsigned long long_cardinal;
#define externalize_long_cardinal(p, buf) move_long(p, buf)
#define internalize_long_cardinal(p, buf) move_long(buf, p)

typedef char character;
#define externalize_character(p, buf) move_byte(p, buf)
#define internalize_character(p, buf) move_byte(buf, p)

typedef float floating;

#if defined(VAX)
extern int externalize_floating();
extern int internalize_floating();
#else
#define externalize_floating(p, buf) move_long(p, buf)
#define internalize_floating(p, buf) move_long(buf, p)
#endif

typedef double long_floating;
#if defined(VAX)
extern int externalize_long_floating();
extern int internalize_long_floating();
#else
#define externalize_long_floating(p, buf) move_hyper(p, buf)
#define internalize_long_floating(p, buf) move_hyper(buf, p)
#endif /* VMS */

typedef short integer;
typedef short short_integer;
#define externalize_integer(p, buf) move_short(p, buf)
#define internalize_integer(p, buf) move_short(buf, p)

typedef long long_integer;
#define externalize_long_integer(p, buf) move_long(p, buf)
#define internalize_long_integer(p, buf) move_long(buf, p)

typedef char *string;
typedef char *string_ref;
#define sizeof_ref_string sizeof_string
#define externalize_ref_string externalize_string
/*
 * sizeof_string(), externalize_string(),  internalize_string() and
 * internalize_ref_string()
 * are defined in library routines.
 */

typedef unsigned short unspecified;
typedef unsigned short short_unspecified;
#define externalize_unspecified(p, buf) move_short(p, buf)
#define internalize_unspecified(p, buf) move_short(buf, p)

typedef unsigned long long_unspecified;
#define externalize_long_unspecified(p, buf) move_long(p, buf)
#define internalize_long_unspecified(p, buf) move_long(buf, p)

/* Obsolete sizeof definitions - remove in version 3 */
#define sizeof_boolean(p) sizeof(char)
#define sizeof_cardinal(p) sizeof(unsigned short)
#define sizeof_long_cardinal(p) sizeof(unsigned long)
#define sizeof_character(p) sizeof(char)
#define sizeof_floating(p) sizeof(float)
#define sizeof_long_floating(p) sizeof(double)
#define sizeof_integer(p) sizeof(short)
#define sizeof_long_integer(p) sizeof(long)
#define sizeof_unspecified(p) sizeof(unsigned short)
#define sizeof_long_unspecified(p) sizeof(unsigned long)
/* end of obsolete defs */

#define sizeof_idl_boolean(p)   sizeof(idl_boolean)
#define sizeof_idl_false(p)     sizeof(idl_false)
#define sizeof_idl_true(p)      sizeof(idl_true)
#define sizeof_idl_byte(p)      sizeof(idl_byte)
#define sizeof_idl_char(p)      sizeof(idl_char)
#define sizeof_idl_small_int(p) sizeof(idl_small_int)
#define sizeof_idl_usmall_int(p) sizeof(idl_usmall_int)
#define sizeof_idl_short_int(p) sizeof(idl_short_int)
#define sizeof_idl_ushort_int(p) sizeof(idl_ushort_int)
#define sizeof_idl_long_int(p)  sizeof(idl_long_int)
#define sizeof_idl_ulong_int(p) sizeof(idl_ulong_int)
#define sizeof_idl_hyper_int(p) sizeof(idl_hyper_int)
#define sizeof_idl_uhyper_int(p) sizeof(idl_uhyper_int)
#define sizeof_idl_short_float(p) sizeof(idl_short_float)
#define sizeof_idl_long_float(p) sizeof(idl_long_float)


/*
 * Translation for enumerated types.
 * We are assuming size of the C enum as 32 bit, unless _ENUM_16 is defined
 */
#define sizeof_enumeration(p) 4
#if defined(_ENUM_16)
#if defined(LOW_ENDER)
#define internalize_enumeration(p, buf) (*(char *)(p) = *((char *)(buf)+1),\
*((char *)(p)+1) = *(char *)(buf), 4)
#define externalize_enumeration(p, buf) (*(char *)(buf) = *((char *)(p)+1),\
*((char *)(buf)+1) = *(char *)(p),\
*((char *)(buf)+2) = 0,\
*((char *)(buf)+3) = 0, 4)
#endif
#else /* _ENUM_16 */
#define internalize_enumeration(p, buf) move_long(buf, p) /* !!!!! */
#define externalize_enumeration(p, buf) move_long(p, buf)
#endif /* _ENUM_16 */
/*
 * Header definition for a sequence type.
 */
typedef struct {
cardinal length;
char *sequence;
} _sequence;

typedef struct {
idl_ulong_int length;
idl_long_int *sequence;
} nc_sequence;

#ifdef _NO_PROTO

/*
 * Common library calls (for clients and servers)
 */
extern void nc_perror();
extern char * nc_perrorstr();
extern void nc_trace();
extern void nc_tracefile();
extern int settrace();
extern char *nc_allocate();
extern int nc_deallocate();
extern unsigned long nc_calltmo();
extern void nc_closeup();
extern void nc_freeall();

/*
 * User library calls on client side
 */
extern int nc_newserv();
extern int nc_wait_reply();
extern unsigned short nc_getrxport();
extern unsigned nc_callflags();

/*
 * Remote library calls (for clients)
 */
extern idl_long_int ncr_util();
extern idl_ushort_int ncr_newserv();
extern long int ncr_link();


/*
 * User library calls on server side
 */
extern  int nc_serverloop();
extern  int nc_wait_msg();
extern  int nc_getcalladdr();
extern  int nc_getuidgid();
extern  int nc_getcallorder();


/*
 * Calls used by client stubs
 */
extern int import_service();
extern _btype *nc_call();
extern long int nc_send_call();
extern _btype *nc_get_reply();
extern  int free_service();
extern _btype *nc_nodal_call();
extern long int nc_send_nodal_call();
extern _btype *nc_get_nodal_reply();

/*
 * Calls used by server stubs
 */
extern  int export_service();
extern  int nc_getcall();
extern  int nc_getmsg();
extern  int server_message();
extern  int server_memory();

/*
 * Calls used by client and server stubs
 */
extern  _btype *nc_callocate();
extern  int nc_cdeallocate();
extern  int *nc_allocexception();
extern  nct_exception *rexcept();
extern  int nc_raise();
extern  int rnext();
extern  float *_fptr();
extern  double *_dfptr();

#else /* _NO_PROTO */

/*
 * Common library calls (for clients and servers)
 */
extern void nc_perror(char *);
extern char * nc_perrorstr();
extern void nc_trace(char *);
extern void nc_tracefile(char *);
extern int settrace(char *, void (*proc) (/* ??? */));
extern char *nc_allocate(int);
extern int nc_deallocate(char *);
extern unsigned long nc_calltmo(int , long unsigned int );
extern void nc_closeup();
extern void nc_freeall();

/*
 * User library calls on client side
 */
extern int nc_newserv(int );
extern int nc_wait_reply(long int , protocol *, long int );
extern unsigned short nc_getrxport(void);
extern unsigned nc_callflags(int service_entry_id, unsigned int flags);

/*
 * Remote library calls (for clients)
 */
extern idl_long_int ncr_util(idl_long_int);
extern idl_ushort_int ncr_newserv(idl_long_int);
extern long int ncr_link();

/*
 * User library calls on server side
 */
extern  int nc_serverloop(long unsigned int timeout, protocol *p, int whenexit, int verbose);
extern  int nc_wait_msg(long int timeout, protocol *p, int *service_entry_id, int *message_id);
extern  int nc_getcalladdr(long unsigned int *host, short unsigned int *port);
extern  int nc_getuidgid(short unsigned int *uid, short unsigned int *gid);
extern  int nc_getcallorder(unsigned char *byteorder, unsigned char *floattype);


/*
 * Calls used by client stubs
 */
extern int import_service(char *name, char *host_name, unsigned int port,
		int protection, int execution, int load, char *file_name,
		int directory, char *dirhostname, unsigned int dirport,
		protocol *p, protocol *loadp, protocol *dirp, long int connlife,
		long int noresponse, unsigned int flags, char *idffile,
		char *parameters, int policy, int maxconnections,
		unsigned int service, int *service_entry_ptr);
extern _btype *nc_call(int mtype, int service_entry_id,
		unsigned int proc_number, long unsigned int random_id,
		_btype *outbuf, int outlen, _btype *inbuf, int inlen);
extern long int nc_send_call(int mtype, int service_entry_id,
		unsigned int proc_number, long unsigned int random_id,
		_btype *buf, int buffer_size);
extern _btype *nc_get_reply(int mtype, int service_entry_id, _btype *inbuf, int inlen);
extern  int free_service(int service_entry_id);
extern _btype *nc_nodal_call(int ntype, int service_entry_id, char *proc_name,
		_btype *outbuf, int outlen, _btype *inbuf, int inlen);
extern long int nc_send_nodal_call(int service_entry_id, _btype *outbuf, int outlen);
extern _btype *nc_get_nodal_reply(int service_entry_id, _btype *inbuf, int inlen);

/*
 * Calls used by server stubs
 */
extern int export_service(char *name, char *host_name, unsigned int port,
			  int protection, int execution, int load, 
			  char *file_name, int directory, char *dirhostname,
			  unsigned int dirport, protocol *p,
			  protocol *loadp, protocol *dirp, long int connlife,
			  long int noresponse, unsigned int flags,
			  char *idffile, char *parameters, int policy, 
			  int maxconnections, 
			  int (**procedures) (/* ??? */), int length, 
			  short unsigned int *permissions,
			  long unsigned int *random_ids, 
			  unsigned int service, int *service_entry_ptr);

extern  int nc_getcall(protocol *p, long unsigned int timeout);
extern  int nc_getmsg(int service_entry_id, unsigned int proc_number,
		      _btype **rpp);

extern  int server_message(void);
extern  int server_memory(void);
/*
 * Calls used by client and server stubs
 */
extern  _btype *nc_callocate(char **buf, int n);
extern  int nc_cdeallocate(char **p);
extern  int sizeof_string(string *p);
extern  int externalize_string(string *p, register _btype *buf);
extern  int internalize_string(string *p, register _btype *buf);
extern  int internalize_string_noalloc(string *p, register _btype *buf);
extern  int *nc_allocexception(void);
extern  nct_exception *rexcept(void);
extern  int nc_raise(int errtype, int errcode, char *errmsg);
extern  int rnext(int errtype, int errcode, char *errmsg);
extern  float *_fptr();


#endif /* _NO_PROTO */

/* definitions for Nodal interface */
#ifdef NODAL
#define ENCNERRORS      0       /* begin of NCN NODAL interface errors */
#define ENCNILLSAENELEM (ENCNERRORS+1)  /* Illegal number of elements in seq of array element */
#define ENCNILLNATYPE   (ENCNERRORS+2)  /* Illegal numerical array type */
#define ENCNILLNANROW   (ENCNERRORS+3)  /* Illegal number of rows in numerical array */
#define ENCNILLNANCOL   (ENCNERRORS+4)  /* Illegal number of columns in numerical array */
#define ENCNILLSANELEM  (ENCNERRORS+5)  /* Illegal number of elements in sequence of array */
#define ENCNILLCANBYTE  (ENCNERRORS+6)  /* Illegal number of bytes in character array */
#define ENCNILLCASNELEM (ENCNERRORS+7)  /* Illegal number of elements in array of sequence of character */
#define ENCNILLSS       (ENCNERRORS+8)  /* Illegal sequence of sequence */
#endif


/*
 * Predefined types used to manipulate arrays and sequences.
 */
typedef char ncboolean;
#define sizeof_ncboolean(p) sizeof(short)

typedef unsigned short nccardinal;
#define sizeof_nccardinal(p) sizeof(short)
#define externalize_nccardinal(p, buf) move_short(p, buf)
#define internalize_nccardinal(p, buf) move_short(buf, p)

typedef unsigned int ncmedium_cardinal;
#define sizeof_ncmedium_cardinal(p) sizeof(short)

typedef unsigned long nclong_cardinal;
#define sizeof_nclong_cardinal(p) sizeof(short)

typedef char nccharacter;
#define sizeof_nccharacter(p) sizeof(char)
#define externalize_nccharacter(p, buf) move_byte(p, buf)
#define internalize_nccharacter(p, buf) move_byte(buf, p)

typedef float ncfloating;
#define sizeof_ncfloating(p) sizeof(float)

typedef double nclong_floating;
#define sizeof_nclong_floating(p) sizeof(double)
#define externalize_nclong_floating externalize_long_floating
#define internalize_nclong_floating internalize_long_floating

typedef short ncinteger;
#define sizeof_ncinteger(p) sizeof(short)
#define externalize_ncinteger(p, buf) move_short(p, buf)
#define internalize_ncinteger(p, buf) move_short(buf, p)

typedef int ncmedium_integer;
#define sizeof_ncmedium_integer(p) sizeof(short)

typedef long nclong_integer;
#define sizeof_nclong_integer(p) sizeof(short)

typedef char *ncstring;

typedef unsigned short ncunspecified;
#define sizeof_ncunspecified(p) sizeof(short)
#define externalize_ncunspecified(p, buf) move_short(p, buf)
#define internalize_ncunspecified(p, buf) move_short(buf, p)

typedef unsigned int ncmedium_unspecified;
#define sizeof_ncmedium_unspecified(p) sizeof(short)

typedef unsigned long nclong_unspecified;
#define sizeof_nclong_unspecified(p) sizeof(short)

/*
 * Predefined types used to mainpulate simple data types.
 * Simple data types are always represented as double (long float)
 * In the Version 1 of NC/RPC they were represented as float
 */
typedef char nboolean;
#define sizeof_nboolean(p) sizeof(double)
#define _externalize_nboolean _externalize_ncfloating
#define _internalize_nboolean _internalize_ncfloating

typedef unsigned short ncardinal;
#define sizeof_ncardinal(p) sizeof(double)
#define _externalize_ncardinal _externalize_ncfloating
#define _internalize_ncardinal _internalize_ncfloating

typedef unsigned int nmedium_cardinal;
#define sizeof_nmedium_cardinal(p) sizeof(double)
#define _externalize_nmedium_cardinal _externalize_ncfloating
#define _internalize_nmedium_cardinal _internalize_ncfloating

typedef unsigned long nlong_cardinal;
#define sizeof_nlong_cardinal(p) sizeof(double)
#define _externalize_nlong_cardinal _externalize_ncfloating
#define _internalize_nlong_cardinal _internalize_ncfloating

typedef char ncharacter;
#define sizeof_ncharacter(p) sizeof(double)
#define _externalize_ncharacter _externalize_ncfloating
#define _internalize_ncharacter _internalize_ncfloating

typedef float nfloating;
#define sizeof_nfloating(p) sizeof(double)
#define _externalize_nfloating _externalize_ncfloating
#define _internalize_nfloating _internalize_ncfloating

typedef double nlong_floating;
#define sizeof_nlong_floating(p) sizeof(double)
#define _externalize_nlong_floating _externalize_ncfloating
#define _internalize_nlong_floating _internalize_ncfloating
#define externalize_nlong_floating externalize_long_floating
#define internalize_nlong_floating internalize_long_floating

typedef short ninteger;
#define sizeof_ninteger(p) sizeof(double)
#define _externalize_ninteger _externalize_ncfloating
#define _internalize_ninteger _internalize_ncfloating

typedef int nmedium_integer;
#define sizeof_nmedium_integer(p) sizeof(double)
#define _externalize_nmedium_integer _externalize_ncfloating
#define _internalize_nmedium_integer _internalize_ncfloating

typedef long nlong_integer;
#define sizeof_nlong_integer(p) sizeof(double)
#define _externalize_nlong_integer _externalize_ncfloating
#define _internalize_nlong_integer _internalize_ncfloating

typedef char *nstring;
#define sizeof_nstring(p) sizeof_ncstring(p)
#define _externalize_nstring _externalize_ncstring
#define _internalize_nstring _internalize_ncstring
#define externalize_nstring externalize_ncstring
#define internalize_nstring internalize_ncstring

typedef unsigned short nunspecified;
#define sizeof_nunspecified(p) sizeof(double)
#define _externalize_nunspecified _externalize_ncfloating
#define _internalize_nunspecified _internalize_ncfloating

typedef unsigned int nmedium_unspecified;
#define sizeof_nmedium_unspecified(p) sizeof(double)
#define _externalize_nmedium_unspec _externalize_ncfloating
#define _internalize_nmedium_unspec _internalize_ncfloating

typedef unsigned long nlong_unspecified;
#define sizeof_nlong_unspecified(p) sizeof(double)
#define _externalize_nlong_unspecified _externalize_ncfloating
#define _internalize_nlong_unspecified _internalize_ncfloating


#endif /* NCC_H */
