/* rx.h                                                       */
/*
 * User include file to use the rx library
 * CS - 07.09.2000 : general heavy clenaup
 *                   removed all old dead code: RX_KERNEL etc... 
 */

/*
 * Error codes not covered by the Berkeley socket errors
 */
#define ETOSERV 1       /* timeout on the message transmission */
#define ETOREPL 2       /* timeout on reply for acknowledged call */
#define EPENDING 3      /* asynch call pending or reply not read */
#define EBADCALLN 4     /* invalid call number or address in call descriptor */
#define EINITIALIZ 5    /* RX already initialised */
#define ENOPEER 6       /* no RX association at the remote port */
#define ENOADDR 7       /* no RX port at the remote IP address (only remote kernel impl) */
#define ERX 8           /* unexpected behavior of RX */
#define EBADLEN 9       /* bad send/receive length in the call descr */
#define EBADUSE 10      /* bad use of RX protocol */
#define ETORCV 11       /* timeout on receiving (kernel only) */
#define ENAMERES 12     /* host not found or no access to nameserver */
/*
 * protocol and call descriptor flags
 */
#define RX_NOACK        (1<<0)  /* no explicit ack of call/reply */
#define RX_ACK          (1<<1)  /* explicit ack for call/reply */
#define RX_NOSIG        (1<<2)  /* net/alarm signals disabled */
#define RX_SIG          (1<<3)  /* enable net/alarm signals */
#define RX_NORETR       (1<<5)  /* no retransmissions, used with RX_NOACK */
#define RX_MODE         (RX_NOACK | RX_ACK | RX_NOSIG | RX_SIG | RX_NORETR)
#define RX_NOREPL       (1<<4)  /* no reply expected */

/* sysconfig.h or netinet/in.h headers are required */
#if defined(nc_32uint)
#define local_32bit_unsigned_int nc_32uint
#else
#define local_32bit_unsigned_int struct in_addr
#endif

struct rx_call_descr {
	unsigned long call_number;      /* returned in rx_get_call() and asynch rx_call() */
	local_32bit_unsigned_int host;  /* in net order */
	unsigned short port;    /* in net order */
	unsigned short flags;   /* the above flags */
	char *smsg, *rmsg;      /* send and receive buffers */
	int slen, rlen;         /* send and receive lengths */
	long timeout;           /* used in rx_call(), rx_get_call() and rx_wait()*/
};


/* RX library interface. */


#ifdef _NO_PROTO
extern int rx_init();
extern int rx_setaddress();
extern int rx_call();
extern int rx_get_call();
extern int rx_get_reply();
extern int rx_put_reply();
extern int rx_more();
extern int rx_wait();
extern int rx_newsd();
extern int rx_fork();
extern int rx_fork_norm();
extern int rx_getsockport();
extern int rx_abort();
extern int rx_closeup();
extern int rx_garbage_collect();
extern rx_trace();
extern void rx_statistics();
extern void rx_setcalln();
extern void rx_gettmo();
extern int rx_settmo();
#else
extern int rx_init(short unsigned int *port);
extern int rx_setaddress(struct rx_call_descr *cd, char *hostname, int port);
extern int rx_call(struct rx_call_descr *cd);
extern int rx_get_call(struct rx_call_descr *cd);
extern int rx_get_reply(struct rx_call_descr *cd);
extern int rx_put_reply(struct rx_call_descr *cd);
extern int rx_more(struct rx_call_descr *cd);
extern int rx_wait(long unsigned int *call_number, long int timeout);
extern int rx_newsd(short unsigned int *port);
extern int rx_fork(void);
extern int rx_fork_norm(void);
extern int rx_getsockport(int *sd, short unsigned int *port);
extern int rx_abort(struct rx_call_descr *cd);
extern int rx_closeup(void);
extern int rx_garbage_collect(void);
extern rx_trace(char *flag_string);
extern void rx_statistics();
extern void rx_setcalln(long int calln);
extern void rx_gettmo(int *retr, int* tmo);
extern int rx_settmo(int retr, int tmo);
#endif

/*
 * To set trace within the rx runtime. Works only if rx library has been made
 * in trace mode. In addition the trace command can be used to activate the
 * trace mode interactively.
 */
struct trace{
	int tron;

	int interrupts;
	int timers;
	int assoc;
	int packets;
	int allocation;
	int exceptions;
	int calls;
};

/*
rx_trace(flag string)
	char *flag_string;

	possible traces:
		    case 'i':
			trace_interrupts = 1;
		    case 't':
			trace_timers = 1;
		    case 's':
			trace_associations = 1;
		    case 'p':
			trace_packets = 1;
		    case 'a':
			trace_allocation = 1;
		    case 'x':
			trace_exceptions = 1;
		    case 'c':
			trace->calls = 1;

e.g. "xsp" is a commonly used trace for exceptions, associations (connections)
and packets.
*/


/*
 * Statistics captured by rx in rx_stats. Applications using them have to
 * define "extern struct rx_stats rx_stats".
 */
struct rx_stats {
	int d_snd;              /* data packets sent */
	int c_snd;              /* control packets sent */
	int cd_rcv;             /* call data packets received */
	int rd_rcv;             /* return data packets received */
	int ca_rcv;             /* call ack packets received */
	int ra_rcv;             /* return ack packets received */
	int dr_ass;             /* association not found (dropped) */
	int dr_typ;             /* unexpected packet type (dropped) */
	int dr_seq;             /* wrong sequence number (dropped) */
	int dr_callno;          /* invalid call number (dropped) */
	int retr_snd;           /* retransmitted packets */
	int retr_rcv;           /* retransmissions received */
};
extern struct rx_stats rx_stats;

