/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 *
 * xx.05.90 kk  Version 2.0
 * 10.05.90 kk  replace exceptions by returning errors
 * 18.06.90 kk  Defined calloc() etc. instead of including malloc.h
 *              which is not present on some systems.
 * 19.06.92 kk  replaced calloc by malloc
 */

#include "sysconfig.h"
#if !defined(NO_STDLIB_H)
#include <stdlib.h>
#else
extern void       *malloc();
extern void free();
#endif
#include "errtype.h"
#include "ncc.h"
#include "nc.h"

/*
 * Allocate a buffer of n elements of char.
 */

char *
nc_allocate(int n)
{
	register char *p;
/* round up to even number of bytes (T.G) */
	if (n&1) n++;

	if((n == 0) || (n < 0)){
		return (0);
	}
	p = (char *) malloc( n );
	if( p == 0 ) {
		RETURNERRP((char *)0, ENCCALLOC, ETNCRT, "malloc()" )
	}
	TRACE(allocation && fprintf(stderr, "malloc %d at %x\n", n, p));
	return( p );
}


/*
 * Deallocate a buffer of n elements of char.
 */

int
nc_deallocate(char *p)
{
	if( p != 0 ) {
		free((char *) p);
		TRACE(allocation && fprintf(stderr, "dealloc at %x\n", p));
		return(0);
	} else {
		RETURNERR(ENCCALLOC, ETNCRT, "free()" );
	}
}
