/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 *
 * xx.05.90 kk  Version 2.0
 * 10.05.90 kk  replace negative errors by (-1) and errtype
 * 29.05.90 kk  Declared global variables used in error handling. This is
 *              to make it transparent for the user.
 * 06.07.90 kk  Added export of ncr services (status, fork, etc.)
 * 07.12.94 kk  Created nc_closeup()to allow fork while using NC/RPC
 */

#include "sysconfig.h"
#include "errtype.h"
#include "ncc.h"
#include "nc.h"


/*
 * Allocate space for a service in the connection table.
 */

int             nc_sinitiated = 0;
nct_service     nc_service[NC_MAXSERVICES] = {0};

nc_statistics_t nc_statistics = {0};

/*
 * Globals used in error handling. Actually it would be better to have them
 * declared by the application but we prefer not to bother the user.
 */
int             errtype = 0;
char            *errstring = 0;
int             errline = 0;
char            *errfile = 0;
/*
 * Global to mark that the closeup of nc should not try to communicate
 * (remote free, statistics, etc)
 */
int     nc_disable_rpc = 0;

/*
 * Initiate the space for a services.
 */

int
nc_sinit(void)
{
	register int i;
	register struct nct_cservice *cd;
	int swout;      /* to avoid getting 100 deallocate messages on trace */

	if (nctrace.serviceid == 1) {
		swout = 1;
		nctrace.serviceid = 0;
	} else
		swout = 0;
	for( i = 0; i < NC_MAXSERVICES; ++i ) {
		nc_sdealloc( i );
	}
	if (swout)
		nctrace.serviceid = 1; /* set it back */


	/* reserve the second service for ncr */
	nc_service[1].nc_servicetype = NC_RESERVED;

	nc_sinitiated = 1;
}

/*
 * Allocate space for a service in the connection table.
 */

int
nc_salloc(struct nct_cservice *cd, int type)
{
	int i;

	if( !nc_sinitiated ) {
		nc_sinit();
	}

	for( i = 0; i < NC_MAXSERVICES; ++i ) {
		if( nc_service[i].nc_servicetype == NC_FREE ) {
			nc_service[i].nc_servicetype = type;
			nc_service[i].nc_cservice = cd;
			TRACE(serviceid && fprintf(stderr, "nc_salloc() service %d\n", i));
			return( i );
		}
	}
	RETURNERR(ENCMAXSERVICES, ETNCRT, "table full");
}

/*
 * Deallocate space for a service in the connection table.
 */

int
nc_sdealloc(int service_entry_id)
{
	nc_deallocate( (char *)nc_service[service_entry_id].nc_cservice );
	nc_service[ service_entry_id ].nc_servicetype = NC_FREE;
	nc_service[ service_entry_id ].nc_cservice = 0;
	TRACE(serviceid && fprintf(stderr, "nc_sdealloc() service %d\n", service_entry_id));

}

/*
 * Find the first service of a specific type in the table.
 * Return negative error if not found.
 */

int
nc_sfirst(int type)
{
	int i;

	for( i = 0; i < NC_MAXSERVICES; ++i ) {
		if( nc_service[i].nc_servicetype == type ) {
			return( i );
		}
	}
	RETURNERR(ENCNOSERVICE, ETNCRT, "no such service type");
}

/*
 * Translate from service entry identifier to the pointer.
 */

struct nct_cservice *
nc_sgetptr(int service_entry_id)
{
	return( nc_service[ service_entry_id ].nc_cservice );
}

/*
 * Translate from service entry identifier to type of service.
 */

int
nc_sgettype(int service_entry_id)
{
	return( nc_service[ service_entry_id ].nc_servicetype );
}


/*
 * Find the service entry from the service pointer.
 */

int
nc_sfindid(struct nct_cservice *cd)
{
	int i;

	for( i = 0; i < NC_MAXSERVICES; ++i ) {
		if( nc_service[i].nc_cservice == cd ) {
			return( i );
		}
	}
	RETURNERR(ENCNOSERVICE, ETNCRT, "no such service");
}

void
nc_closeup()
{
	rx_garbage_collect();
	rx_closeup();
	nc_disable_rpc = 1;
}


/* disable any rpc communication at closeup */
void
nc_freeall()
{
	int i;

	for( i = 0; i < NC_MAXSERVICES; ++i ) {
		if( nc_service[i].nc_servicetype != NC_FREE ) {
			free_service(i);
		}
	}
	nc_sinitiated = 0;
}
