/* %W% %E% */
/*
 * xx.05.90 kk  Version 2.0
 * 05.06.90 kk  The error reporting function nc_perror() has been added
 *              and it should replace ncperror(). nc_perror() prints the
 *              caller's message followed by the error type message and
 *              error message. In case of UNIX error perror is called.
 *              nc_perrorstr() is similar but returns pointer to the
 *              error string.
 *              ncperror() is retained for backward compatibility.
 * LP - 12.06.2000 - fixed sys_nerr declaration
 * LP - 23.08.2000 - fixed fprintf in line 287
 * CS - 11.09.2000 - clenaup version, obsolete #ifdef removed
 */

#include "sysconfig.h"
#include "ncc.h"
#include "nc.h"
#include "errtype.h"
#include <errno.h>
#include <stdio.h>

#if defined(OS9)
#include <string.h>
#endif

static char nc_errmsg[200] = {0}; /* error string generated by nc_perrorstr() */
/* on some systems not defined in errno.h */

#if defined(OS9)  /* LP - 12.06.2000 */
  int sys_nerr = 100;
#endif
#if defined(linux)  /* CS - 23.03.2004 */
  int sys_nerr = 124;
  extern int errno;
#endif
#if defined(HPPA)
  extern int sys_nerr;
  extern const char *sys_errlist[]; /* CS - 29.12.2000 */
#endif

/*
 * Print network compiler error in standard format. The format is
 * similar to the system procedure perror().
 */

int
ncperror(char *s, int errcode)
{
	int errsave;

	if (errno != errcode) {
		errsave = errno;
		errno = errcode;
		nc_perror( s );
		errno = errsave;
	} else
		nc_perror( s );
	return( 0 );
}


char *
nc_perrorstr()
{
	extern char *ncerror(int errcode);
	extern char *rxerror();
	int etype;
	etype = errtype;
	if (etype == 0) {
		/* probably Unix error code */
		sprintf(nc_errmsg, "Bad use of error function :error type is 0 (not set) errno is %d\n", errno);
		return(nc_errmsg);
	}
	if (etype > ETOFSREM) {
		sprintf(nc_errmsg, " %d level REMOTE ", etype/ETOFSREM);
		etype = etype % ETOFSREM;
	}
	switch (etype) {
	    case ETUNIX:
		sprintf(nc_errmsg, "UNIX " );
		break;
	    case ETOS9:
		sprintf(nc_errmsg, "OS9 " );
		break;
	    case ETVMS:
		sprintf(nc_errmsg, "VMS " );
		break;
	    case ETNCRT:
		sprintf(nc_errmsg, "NC/RPC runtime " );
		break;
	    case ETNCSTB:
		sprintf(nc_errmsg, "NC/RPC stub " );
		break;
	    case ETPROT:
		sprintf(nc_errmsg, "NC Transport (NC version 1 error)" );
		break;
	    case ETRX:
		sprintf(nc_errmsg, "RX Protocol " );
		break;
	    case ETMIL:
		sprintf(nc_errmsg, "MIL1553 " );
		break;
	    case ETNODAL:
		sprintf(nc_errmsg, "NODAL " );
		break;
	    case ETNCNSTB:
		sprintf(nc_errmsg, "NODAL stub " );
		break;
	    case ETAPPL:
		sprintf(nc_errmsg, "Application " );
		break;
	    default:
		sprintf(nc_errmsg, "unknown (assuming UNIX) ");
	}
	sprintf(nc_errmsg, "error: ");

	/* print the detailed error type */
	switch (etype) {
	    case ETUNIX:
	    case ETOS9:
	    case ETVMS:
		if (errno < sys_nerr)

#if defined(OS9) || defined(linux)
      sprintf(nc_errmsg, "%s\n", strerror(errno));
#else
      sprintf(nc_errmsg, "%s\n", sys_errlist[errno]);
#endif
		else
			sprintf(nc_errmsg, "unknown system error %d\n", errno);
		if ((errstring == (char *)0) || (errstring[0] == '\0'))
			return(nc_errmsg);
		break;
	    case ETNCRT:
		sprintf(nc_errmsg, "%s ",ncerror(errno) );
		break;
	    case ETPROT:
		sprintf(nc_errmsg, "errno = %d ", errno );
		break;
	    case ETRX:
		sprintf(nc_errmsg, "%s ",rxerror(errno) );
		break;
	    case ETMIL:
		sprintf(nc_errmsg, "errno = %d ", errno);
		break;
	    case ETAPPL:
		sprintf(nc_errmsg, "errno = %d ", errno);
		break;
	    default:
		if (errno < sys_nerr)

#if defined(OS9)||defined(linux)
	sprintf(nc_errmsg, "%s\n", strerror(errno));
#else
        sprintf(nc_errmsg, "%s\n", sys_errlist[errno]);
#endif
		else
			sprintf(nc_errmsg, "unknown system error %d\n", errno);
		if ((errstring == (char *)0) || (errstring[0] == '\0'))
			return(nc_errmsg);
	}
	if ((errstring != (char *)0) && (errstring[0] != '\0'))
		sprintf(nc_errmsg, "( %s )", errstring);
	return(nc_errmsg);
}


void
nc_perror(char *s)
{
	extern char *ncerror(int errcode);
	extern char *rxerror();
#if !(defined(OS9) )
	extern int sys_nerr;
#else
	int sys_nerr = 100;
#endif
	int etype;
	etype = errtype;
	fprintf(stderr, "%s: ", s);
	if (etype == 0) {
		/* probably Unix error code */
		fprintf(stderr, "Bad use of error function :error type is 0 (not set) errno is %d\n", errno);
		return;
	}

	if (etype > ETOFSREM) {
		fprintf(stderr, " %d level REMOTE ", etype/ETOFSREM);
		etype = etype % ETOFSREM;
	}
	switch (etype) {
	    case ETUNIX:
		fprintf(stderr, "UNIX " );
		break;
	    case ETOS9:
		fprintf(stderr, "OS9 " );
		break;
	    case ETVMS:
		fprintf(stderr, "VMS " );
		break;
	    case ETNCRT:
		fprintf(stderr, "NC/RPC runtime " );
		break;
	    case ETNCSTB:
		fprintf(stderr, "NC/RPC stub " );
		break;
	    case ETPROT:
		fprintf(stderr, "NC Transport (NC version 1 error)" );
		break;
	    case ETRX:
		fprintf(stderr, "RX Protocol " );
		break;
	    case ETMIL:
		fprintf(stderr, "MIL1553 " );
		break;
	    case ETNODAL:
		fprintf(stderr, "NODAL " );
		break;
	    case ETNCNSTB:
		fprintf(stderr, "NODAL stub " );
		break;
	    case ETAPPL:
		fprintf(stderr, "Application " );
		break;
	    default:
		fprintf(stderr, "unknown (assuming UNIX) ");
	}
	fprintf(stderr, "error: ");

	/* print the detailed error type */
	switch (etype) {
	    case ETUNIX:
	    case ETOS9:
	    case ETVMS:
		if (errno < sys_nerr)

#if defined(OS9)||defined(linux)
	sprintf(nc_errmsg, "%s\n", strerror(errno));
#else
        sprintf(nc_errmsg, "%s\n", sys_errlist[errno]);
#endif
		else
			fprintf(stderr, "unknown system error %d\n", errno);
		if ((errstring == (char *)0) || (errstring[0] == '\0'))
			return;
		break;
	    case ETNCRT:
		fprintf(stderr, "%s ",ncerror(errno) );
		break;
	    case ETPROT:
		fprintf(stderr, "errno = %d ", errno );
		break;
	    case ETRX:
		fprintf(stderr, "%s ",rxerror(errno) );
		break;
	    case ETMIL:
		fprintf(stderr, "errno = %d ", errno);
		break;
	    case ETAPPL:
		fprintf(stderr, "errno = %d ", errno);
		break;
	    default:
		if (errno < sys_nerr)

#if defined(OS9)||defined(linux)
	sprintf(nc_errmsg, "%s\n", strerror(errno));
#else
        sprintf(nc_errmsg, "%s\n", sys_errlist[errno]);
#endif
		else
			fprintf(stderr, "unknown system error %d\n", errno);
	}
	if ((errstring != (char *)0) && (errstring[0] != '\0'))
		fprintf(stderr, "( %s )\n", errstring);
	else
		/* fprintf(stderr, "\n", errstring);  LP - 23.08.2000 */
    fprintf(stderr, "\n");
}


char *
ncerror(int errcode)
{
	static char defaultmessage[20];

	/* Switch on the error number to find the corresponding string. */
	switch( errcode ) {

		/* Network library protocol errors.                     */

		case ENCMAXSERVICES:
			return( "Maximum number of services per process" );

		case ENCNOSERVICE:
			return(  "No service of this type" );

		case ENCCBADSR:
			return( "No match for this service" );

		case ENCCBADPR:
			return( "No such procedure in this service" );

		case ENCCBADID:
			return( "Bad procedure ID" );

		case ENCCNOPROT:
			return( "Transport protocol information is not available" );

		case ENCCALLOC:
			return( "Memory alloc/free failed" );

		case ENCCINCSEND:
			return( "RPC buffer was incompletely sent" );

		case ENCCINCRECV:
			return( "RPC buffer was incompletely received" );

		case ENCCTOOLSEQ:
			return( "Sequence is too long" );

		case ENCCEXPORT:
			return( "The service is already exported" );

		case ENCCTYPE:
			return( "Received illegal RPC message type" );

		case ENCCACCES:
			return( "Permission denied" );

		case ENCCNOMASTER:
			return( " No reply from the master server");

		case ENCCNOREPL:
			return("Timeout on reply for asynch call");

		case ENCCPROTECTION:
			return( "Protection violation" );

		case ENCCEXECUTION:
			return( "Execution type violation" );

		/* Default system error.                                */
		default:
			sprintf( defaultmessage, "%s %d", "Error code", errcode );
			return( defaultmessage );

	}
}
