/*
 * Copyright (C) 1986 by Inge B. Kristiansen.
 * All rights reserved.
 */

#include "sysconfig.h"
#include "ncc.h"
#include "nc.h"


/*
 * Global variable controlling the timeouts.
 */

static timeouts tm = {
	NC_TM_CONN,             /* connection timer */
	NC_TM_CONNLIFE,         /* connection lifetime */
	NC_TM_RETRANS,          /* retransmit timer */
	NC_TM_RETRANSRQ,        /* retransmit request timer */
	NC_TM_NRETRANS,         /* number of retransmissions */
	NC_TM_NORESP,           /* no response timer */
	NC_TM_NET               /* network timer */
};


/*
 * Set the part of the timeout structure which can be modified by the
 * user in the interface description file.
 */

timeouts *
nc_tmformat(struct nct_common_service *cd)
{
	/* Set up the timeout structure to be used.                     */
	tm.tm_c_l = cd->nc_connlife;
	tm.tm_nr_t = cd->nc_noresponse;

	/* Return a pointer to the structure.                           */
	return( &tm );
}


/*
 * Set the timeouts for an individual connection.
 */

int
nc_tminit(struct protocol *p, struct nct_common_service *cd)
{
	timeouts *t;
	int status;

	/* Format the timeout structure first.                          */
	t = nc_tmformat( cd );

	/* Then insert the new values at the transport level.           */
	if( (status = (*p->p_iocntl)( cd->nc_connection, NC_IOCSETTMOUT, IO_WRITE,
			(char *)t, sizeof(timeouts))) < 0 ) {
		return( status );
	}

	/* Return Ok status.                                            */
	return( 0 );
}
