/* %W% %E% */
/*
* Creation 1986 by Inge B. Kristiansen.
 *
 * 02/04/89 (raw): changed getuid/getgid call for OS-9
 * 26/01/90 (raw): fill empty hostname with local hostname
 * xx.05.90 kk  Version 2.0
 * 09.05.90 kk  removed obsolete code and commented out code related to name
 *              server and loader (to be reviewed).
 *              added code related to the rx interface.
 *              replaced negative error returns by returning (-1) and setting
 *              errno/errtype
 *              removed check of return of nc_signals() since it's always 0
 * 18.06.90 kk  Temporary disabled nc_signals() call
 * 10.07.90 kk  Added export of ncr service.
 */

#include "sysconfig.h"
#if !defined(NO_STDLIB_H)
#include <stdlib.h>
#endif
#if (defined (_POSIX_SOURCE) || defined(POSIXsignal))
#include <unistd.h>
#include <sys/types.h>
#endif
#include <signal.h>
#if !defined(NC_FORK_REVERSE)
#include <sys/wait.h>
#endif
#include "errtype.h"
#include "ncc.h"
#include "nc.h"

extern unsigned long nc_select_mask;
extern int nc_rx_fd;            /* to allow multiply protocol select */
extern int nc_ex_spawned;

/*
 * Global variables maintained by this module.
 */

struct nc_queue   exported_services = { 0 };      /* Queue of exported services */
int ncms_imported = 0;
#if !defined(_NO_ATEXIT)
int nc_atexit = 0;
#endif
int nc_setpgid = 1;     /* to allow disabling of setpgid for detached servers */
int nc_fork_reverse = 1; /* switch for two types of parentship */
int nc_sigchld_ev = 0;	/* to know if signal handler set */


char    *niname         = NC_E_STRING;          /* Service name */
char    *nihostname     = NC_E_STRING;          /* Host name */
int     niport          = NC_E_INT;             /* Port number to use */
char    *nidirhostname  = NC_E_STRING;          /* Directory host name */
int     nidirport       = NC_E_INT;             /* Directory port number */


/*
 * Export a service to the network. That is make it possible to
 * call it from another process.
 */


int nc_initservice (register struct nct_cservice *cd);

int
export_service(char *name, char *host_name, unsigned int port,
	int protection, int execution, int load, char *file_name,
	int directory, char *dirhostname, unsigned int dirport,
	protocol *p, protocol *loadp, protocol *dirp, long int connlife,
	long int noresponse, unsigned int flags, char *idffile,
	char *parameters, int policy, int maxconnections,
	int (**procedures) (/* ??? */), int length,
	short unsigned int *permissions, long unsigned int *random_ids,
	unsigned int service, int *service_entry_ptr)
{
	register struct nct_cservice *cd, *ncd;
	register int service_entry_id;
	int status, sid, pid;
	char local_host[NCC_MAXSTR];
	unsigned short lport;

	extern int (*_ncr_procedures[])();
	extern unsigned short _ncr_permission[];
	extern unsigned long _ncr_random[];
	extern nct_service nc_service[];

#if (defined (_POSIX_SOURCE) || defined(POSIXsignal))
	struct sigaction sig_action;
#else
#if defined (BSDsignal)
	struct sigvec sig_vec;
#endif
#endif
#if !defined(NC_FORK_REVERSE)
	void nc_detach_ex ();
#endif

#ifdef OS9
	int giduid;
#endif /* OS9 */

	/* Define tracing activation routine */
	settrace("nc", nc_trace);
	settrace("tf", nc_tracefile);

	/* Update the statistics variable */
	nc_statistics.export++;

#if !defined(_NO_ATEXIT)
	if (!nc_atexit) {
		atexit(nc_freeall);
		nc_atexit = 1;
	}
#endif
	/* Check if the service already is present in the export list.  */
	if( nc_exported_services( service ) != (nct_cservice *)0 ) {
		RETURNERR(ENCCEXPORT, ETNCRT, "" )
	}

	/* Allocate a connection descriptor to be used by the system    */
	/* for subsequent transactions.                                 */
	if( (cd = (nct_cservice *)nc_allocate( sizeof(nct_cservice) )) == 0 ) {
		return(-1);
	}

	/* Allocate a service entry identifier to be used by the user   */
	/* transactions.                                                */
	if( (service_entry_id = nc_salloc( cd, NC_EXPORTED )) < 0 ) {
		nc_deallocate( (char *)cd );
		return(-1);
	}

	/* Insert the information given by the user into the connection */
	/* descriptor.                                                  */
	cd->ncc_name = name;
	cd->ncc_flags = flags;
#ifdef OS9
	giduid=getuid();
	cd->ncc_uid = 0xffff & giduid;
	cd->ncc_gid = giduid >> 16;
#else
	cd->ncc_uid = getuid();
	cd->ncc_gid = getgid();
#endif
	tm_now( &cd->ncc_time );
	cd->ncc_hostname = host_name;
	cd->ncc_port = port;
	cd->ncc_protection = protection;
	cd->ncc_execution = execution;
	cd->ncc_load = load;
	cd->ncc_filename = file_name;
	cd->ncc_directory = directory;
	cd->ncc_dirhostname = dirhostname;
	cd->ncc_dirport = dirport;
	cd->ncc_protocol = p;
	cd->ncc_loadprotocol = loadp;
	cd->ncc_dirprotocol = dirp;
	cd->ncc_connlife = connlife;
	if (p == NC_P_RX)
		cd->ncc_noresponse = noresponse*1000;
	else
		cd->ncc_noresponse = noresponse;
	cd->ncc_idffile = idffile;
	cd->ncc_parameters = parameters;
	cd->ncc_policy = policy;
	cd->ncc_maxconn = maxconnections;
	cd->ncc_service = service;
	cd->ncc_localport = port;
	cd->ncc_connection = NCC_NOCONNECT;
	cd->ncc_length = length;
	cd->ncc_procedures = procedures;
	cd->ncc_permissions = permissions;
	cd->ncc_random_ids = random_ids;

	/* Check if the init_service() routine has been called.         */
	nc_initservice( cd );

	/* If the host name is not given, the local host is assumed.    */
	if( !cd->ncc_hostname ) {
#if !defined(XENIX) /*andrebbe abilitata */
		cd->ncc_hostname = "localhost";
#else
		/* Get the name of the local host.                      */
		if( (cd->ncc_hostname = nc_allocate( NCC_MAXSTR )) == 0 ) {
			nc_sdealloc( service_entry_id );
			return( -1 );
		}
		if( (status = gethostname(cd->ncc_hostname, NCC_MAXSTR)) < 0 ) {
			nc_sdealloc( service_entry_id );
			RETURNERR(errno, ETUNIX, "gethostname()" );
		}
#endif
	}

	if( nc_service[1].nc_servicetype == NC_RESERVED ) {
	        /* export the ncr service omitting not useful information */
		if( (ncd = (nct_cservice *)nc_allocate( sizeof(nct_cservice) )) == 0 ) {
			return(-1);
		}

		/* the second service is reserved for the general services */
		nc_service[1].nc_servicetype = NC_EXPORTED;
		nc_service[1].nc_cservice = ncd;

		ncd->ncc_name = "ncr";
		ncd->ncc_flags = 0777;
		ncd->ncc_hostname = cd->ncc_hostname;

		ncd->ncc_service = 650;  /* computed by the compiler hash function */
		ncd->ncc_connection = NCC_NOCONNECT;
		ncd->ncc_protocol = NC_P_RX;
		ncd->ncc_length = 5;    /* it's not quite true, but we do not know the real size */
		ncd->ncc_procedures =_ncr_procedures;
		ncd->ncc_permissions = _ncr_permission;
		ncd->ncc_execution = execution;
		ncd->ncc_protection = protection;
		ncd->ncc_random_ids = _ncr_random;
		q_atail( (struct nc_queue *)&exported_services, (struct queue_item *)ncd );
	}
#if !defined(NC_FORK_REVERSE)
	if (!nc_fork_reverse && (execution == NC_EX_ANY || execution == NC_EX_DETACH)) {
		/* install handler of child death */
#if (defined (_POSIX_SOURCE) || defined(POSIXsignal))
		sig_action.sa_handler = nc_detach_ex;
		sig_action.sa_flags = 0;

		if ((sigaction (SIGCHLD, &sig_action, NULL)) != 0) {
			perror ("sigaction ()");
			exit (1);
		}
#else
#if defined(BSDsignal)
		sig_vec.sv_handler = nc_detach_ex;
		sig_vec.sv_onstack = 0;
		if ((sigvec (SIGCHLD, &sig_vec, NULL)) != 0) {
			perror ("sigvec ()");
			exit (1);
		}
#endif
#if defined (SYSVsignal)
		signal (SIGCLD, nc_detach_ex);
#endif
#endif
	nc_sigchld_ev = 1;
	}
#endif /* NC_FORK_REVERSE */
	if (p == NC_P_RX) {
		/*
		 * initialise the local communication endpoint and
		 * initialise the call descriptor with the remote address
		 */
		if (execution == NC_EX_MASTER)
			lport = 0;
		else
			lport = port;
		if ((status = rx_init(&lport)) < 0 ) {
			nc_sdealloc( service_entry_id );
			return(-1);
		}
		cd->ncc_localport = lport;
		TRACE( sockets && fprintf(stderr, "export_service(): rx port %d fd %d\n", cd->ncc_localport, status) );
		/* set the nc_rx_fd and add rx connection to the mask */
		nc_rx_fd = status;
		nc_select_mask |= (1 << status);
	} else {

	return(-1);
	}

	/* Enqueue the service on the queue of exported services.       */
	q_ahead( &exported_services, (struct queue_item *)cd );

	/* Update the variables to be returned to the server.           */
	*service_entry_ptr = service_entry_id;
	return( service_entry_id );
}


#if !defined(NC_FORK_REVERSE)
void nc_detach_ex ()
{
#if (defined (_POSIX_SOURCE) || defined(POSIXwait))
	pid_t chld_pid;
	int status;
#else
	int chld_pid;
	union wait status;
#endif

#if (defined (_POSIX_SOURCE) || defined(POSIXwait))
  chld_pid = waitpid (-1, &status, WNOHANG);
#else
#if defined (BSDwait)
  chld_pid = wait3 (&status, WNOHANG, 0);
#else
  chld_pid = wait (&status);/* CS - 11.09.2000 - ")" was missing!*/
  signal (SIGCLD, nc_detach_ex);
#endif
#endif
  TRACE( sockets && fprintf(stderr, "nc_detach_ex():  wait*() returned %d%x\n", chld_pid) );
  /* no statistics so far */
/*
fprintf(stderr, "sig child from %d\n", chld_pid);
*/
#if defined (BSDwait)
  chld_pid = wait3 (&status, WNOHANG, 0);
#endif
  return;
}
#endif /* NC_FORK_REVERSE */


/*
 * Set up user defined initialization given by the user in the
 * routine init_service().
 */

int
nc_initservice(register struct nct_cservice *cd)
{
	/* Set the service name.                                        */
	if( niname ) {
		cd->ncc_name = niname;
	}

	/* Set the own address.                                         */
	if( nihostname ) {
		cd->ncc_hostname = nihostname;
	}
	if( niport ) {
		cd->ncc_localport = cd->ncc_port = niport;
	}

	/* Set the directory address.                                   */
	if( nidirhostname ) {
		cd->ncc_dirhostname = nidirhostname;
	}
	if( nidirport ) {
		cd->ncc_dirport = nidirport;
	}
	return( 0 );
}


/*
 * Check if the service already is present in the export list. If it is
 * the address of the connection descriptor is returned, otherwise 0.
 */

nct_cservice *
nc_exported_services(short unsigned int service)
{
	register struct nct_cservice *cd;

	/* Loop through all the exported services in the export list.   */
	for( cd = (nct_cservice *)exported_services.head; cd != 0;
			cd = (nct_cservice *)cd->ncc_link ) {

		/* Check if the service is in the export list.          */
		if( cd->ncc_service == service ) {
			return( cd );
		}
	}
	/* The service is not in the export list. Return 0.             */
	return( (nct_cservice *)0 );
}
