/*
 * Copyright (C) 1986 by CERN.
 *
 * 19.05.90     First written by Kris Kostro (nc version 2)
 * 02.07.90     removed rrecover() calls, moved rrecover.c to obsolete
 * CS - 07.09.2000 removed all conditionale for RTCP/NCMS
 */

#include "sysconfig.h"
#include "errtype.h"
#include "ncc.h"
#include "nc.h"

/*
 * Global values used to remember the last call. Having this we can give
 * the uid, origin etc. of the last call to the application.
 */
struct ncct_call_header nc_lastcallh;
nct_cservice *nc_lastcd;
_btype *nc_last_msg = 0;
int nc_last_proc = 0;

extern unsigned long nc_select_mask;


extern int nc_permission (struct ncct_call_header *ncc_ihead, struct nct_cservice *cd, int proc_number);
extern nct_service nc_service[];

int
nc_getcall(protocol *p, long unsigned int timeout)
{
	nct_cservice *cd;
	register struct ncct_call_header *ncc_call;
	struct ncct_return_header *ncc_return;
	int sid, fd, fd_ret, nbytes;
	int status, rstat, length;
	short localerr = 0;
	_btype *bp, *rp;
	struct rx_call_descr call_d;

	if (p != NC_P_RX) {
                return(-1);
	} else {
		call_d.rmsg =  (char *)&nc_lastcallh;
		call_d.rlen = sizeof(nc_lastcallh);
		call_d.timeout = timeout;

		if ((status = rx_get_call(&call_d)) < 0)
			return(-1);
		if (status == 0)
			return(0);
	}

	/* Internalize the call header for later use.           */
	ncc_call = &nc_lastcallh;
	if (ncc_call->ncc_byteorder != NC_BYTEORDER) {
	    ncc_call->ncc_length = swapal( ncc_call->ncc_length );
	    ncc_call->ncc_random_id = swapal( ncc_call->ncc_random_id );
	    ncc_call->ncc_service_id = swapas( ncc_call->ncc_service_id );
	    ncc_call->ncc_procedure_number = swapas( ncc_call->ncc_procedure_number );
	    ncc_call->ncc_userid = swapas( ncc_call->ncc_userid );
	    ncc_call->ncc_groupid = swapas( ncc_call->ncc_groupid );
	    ncc_call->ncc_type = swapas( ncc_call->ncc_type );
	}

	TRACE(callmsg && fprintf(stderr, "nc_getcall from %s: mt %d %d v%d.%d len %d s_id %d typ %d pr %d\n",
		print_ipaddr(&(call_d.host)),
		ncc_call->ncc_byteorder, ncc_call->ncc_floattype, ncc_call->ncc_version,
		ncc_call->ncc_revision, ncc_call->ncc_length, ncc_call->ncc_service_id,
		ncc_call->ncc_type, ncc_call->ncc_procedure_number) );

	errno = 0;
	errtype = 0;
	length = ncc_call->ncc_length - CALL_HEADER_BYTES;

	/*
	 * Check validity before doing anything else.
	 */
	if( (cd = nc_exported_services(ncc_call->ncc_service_id)) == (nct_cservice *)0 ) {
		localerr = ENCCBADSR;
		cd = nc_service[1].nc_cservice; /* just for sending error */
	} else {
		/*
		 * Remember the cd in a global place so we can use it later
		 * to get service information.
		 */
		nc_lastcd = cd;
		if( ncc_call->ncc_procedure_number >= cd->ncc_length ) {
			localerr = ENCCBADPR;
		} else if(ncc_call->ncc_random_id != cd->ncc_random_ids[ncc_call->ncc_procedure_number] &&
			ncc_call->ncc_random_id != 0 ) {
		    if (ncc_call->ncc_version == 2 && ncc_call->ncc_revision > 9)
			/* in version 2.10 we added allocation/free attribute to parameters */
			localerr = ENCCBADID;
		} else if( (status = nc_permission( ncc_call, cd, ncc_call->ncc_procedure_number )) < 0 ) {
			localerr = errno;
		}
	}

	if( (bp = nc_allocate( length + sizeof(char) )) == 0 )
		return(-1);
	{
		cd->ncc_cd.call_number = call_d.call_number;
		cd->ncc_cd.host = call_d.host;
		cd->ncc_cd.port = call_d.port;
		cd->ncc_cd.rmsg =  bp;
		cd->ncc_cd.rlen = length;
		if (length > 0) {
			if (rx_more(&(cd->ncc_cd)) < 0) {
				nc_deallocate( bp );
				return(-1);
			}
		}
	}
	/* TRACE data */
	if (nctrace.tron && nctrace.data) {
		char *p;
		int i;

		fprintf(stderr, "call data:");
		for (i = 0, p = bp; i<length; ++i, ++p) {
			if (!(i%20)) fprintf(stderr, "\n");
			fprintf(stderr, "%2x ",*p);
		}
		fprintf(stderr, "\n");
	}

	/* Update the statistics variable */
	nc_statistics.sr_receivelen += length;
	switch (ncc_call->ncc_type) {
		case NCC_CALL : nc_statistics.call_type[NCC_CALL]++;
			break;
		case NCC_MESSAGE : nc_statistics.call_type[NCC_MESSAGE]++;
			break;
		case NCC_NOREPLY : nc_statistics.call_type[NCC_NOREPLY]++;
			break;
		case NCC_UNRELIABLE : nc_statistics.call_type[NCC_UNRELIABLE]++;
			break;
		default :
			break;
	      }

	/*
	 * If there was no error we can call the server stub (or pass the call
	 * to the user in case of message).
	 * Otherwise we return an error message.
	 */
	if (! (localerr || errtype)) {
		if ((p == NC_P_RX) && (ncc_call->ncc_type == NCC_NOREPLY ||
		    ncc_call->ncc_type == NCC_MESSAGE )) {
			/* sync with the client */
			call_d.slen = 0;
			if ((status = rx_put_reply(&call_d)) < 0)
				return(-1);
		}
		if (ncc_call->ncc_type == NCC_MESSAGE ) {
			/* receiving of message */
			nc_last_msg = bp;
			nc_last_proc = ncc_call->ncc_procedure_number;
			return(1);
		}
		(*cd->ncc_procedures[ncc_call->ncc_procedure_number]) ( bp, &rp, &length );
		nc_deallocate( bp );
		if ((ncc_call->ncc_type == NCC_NOREPLY) ||
			(ncc_call->ncc_type == NCC_UNRELIABLE)) {
			/* call with no return parameters */
			nc_deallocate(rp);
			return(1);
		}
	}
	/* errtype might have been set by user's proc */
	if (localerr || errtype) {
		if (localerr) {
			errno = localerr;
			if (errtype == 0) {
				errtype = ETNCRT;
				errstring = "";
			}
		}
		if ((p == NC_P_RX) && (ncc_call->ncc_type != NCC_CALL)) {
			/* sync with the client */
			call_d.slen = 0;
			rx_put_reply(&call_d);
			return(-1);
		}
		length = RETURN_HEADER_BYTES + sizeof_string(&errstring);
		/* discard buffer allocated by stub */
		nc_deallocate(rp);
		/* and alloc a new one */
		rp = nc_allocate(RETURN_HEADER_BYTES + sizeof_string(&errstring));
		bp = rp + RETURN_HEADER_BYTES;
		externalize_string(&errstring, bp);
	}
	/* Set the return header.       */
/* kkk later the return header should be made global and preset with
	version, byteorder etc */
	ncc_return = (ncct_return_header *)rp;
	ncc_return->ncc_byteorder         = NC_BYTEORDER;
	ncc_return->ncc_floattype         = NC_FLOATTYPE;
	ncc_return->ncc_version           = nc_version;
	ncc_return->ncc_revision          = nc_revision;
	ncc_return->ncc_length            = length;
	ncc_return->ncc_errtype           = errtype;
	ncc_return->ncc_errcode           = errno;
	ncc_return->ncc_type              = NCC_RETURN;

	TRACE(retmsg && fprintf(stderr, "nc_getcall reply: totlen %d errtyp %d err %d\n",
		ncc_return->ncc_length, ncc_return->ncc_errtype,
		ncc_return->ncc_errcode) );

	/* TRACE data */
	if (nctrace.tron && nctrace.data) {
		char *p;
		int i;

		fprintf(stderr, "return data:");
		for (i = 0, p = rp+RETURN_HEADER_BYTES; i<length-RETURN_HEADER_BYTES; ++i, ++p) {
			if (!(i%20)) fprintf(stderr,"\n");
			fprintf(stderr,"%2x ",*p);
		}
		fprintf(stderr, "\n");
	}

	/* Update the statistics variable */
	nc_statistics.sr_sendlen += (length - RETURN_HEADER_BYTES);
	nc_statistics.call_type[NCC_RETURN]++;

	/* Send the return message to the client        */
	   {
		cd->ncc_cd.smsg =  rp;
		cd->ncc_cd.slen = length;
		cd->ncc_cd.flags = 0;
		cd->ncc_cd.timeout = cd->ncc_noresponse;
		if (rx_put_reply(&(cd->ncc_cd)) < 0) {
			nc_deallocate(rp);
			return(-1);
	   }
	}
	/* Free the send buffer          */
	nc_deallocate(rp);
	return(1);
}

