/*
 * select.h
 *
 * OS-9 select() call missing prototypes and macros
 *
 * LP - 23.05.2000
 */
 
#ifndef _OS9_SELECT
#define _OS9_SELECT

#include <types.h>

/* Structure used to represent timezones for gettimeofday()
   for compatibilty only, will be left empty or unchanged! */
struct timezone {
	int	tz_minuteswest;	/* minutes west of Greenwich */
	int	tz_dsttime;	    /* type of dst correction */
};

/* adapted from HP-UX types.h and OS-9 select.h */
#define FD_SETSIZE 32                  /* see select() */

#define FD_MASKSIZE	32
typedef u_int32 fd_mask;

#define FD_SIZE		(((FD_SETSIZE) + ((FD_MASKSIZE)-1)) / (FD_MASKSIZE))

#define NFDBITS (sizeof(fd_mask) * 8)  /* 8 bits per byte */

#define howmany(x,y)  (((x)+((y)-1))/(y))
    
typedef struct fd_set {
  fd_mask fds_bits[howmany(FD_SETSIZE, NFDBITS)];
} fd_set;

/*	select() bit-field manipulation macros:
**
**	FD_SET(int fd, fd_set* fdset);
**	FD_CLR(int fd, fd_set* fdset);
**	FD_ISSET(int fd, fd_set* fdset);
**	FD_ZERO(fd_set* fdset);
*/
#if 0 /* from BSD/UNIX */
	#define FD_SET(fd,fdset)	((fdset)->fds_bits[(fd)/(FD_MASKSIZE)] |=  \
							(1 << ((fd) % FD_MASKSIZE)))
	#define FD_CLR(fd,fdset)	((fdset)->fds_bits[(fd)/(FD_MASKSIZE)] &=  \
							~(1 << ((fd) % FD_MASKSIZE)))
	#define FD_ISSET(fd,fdset)	((fdset)->fds_bits[(fd)/(FD_MASKSIZE)] &  \
							(1 << ((fd) % FD_MASKSIZE)))
	#define FD_ZERO(fdset)		{ int field;  \
							  for (field=0; field<FD_SIZE; field++)  \
							      ((fdset)->fds_bits[field] = 0); }
#else /* ASSUME: FD_SETSIZE (max open paths) == 32 */
	#define FD_SET(fd,fdset)	(*((fd_mask*)fdset) |= (1 << fd))
	#define FD_CLR(fd,fdset)	(*((fd_mask*)fdset) &= (~(1 << fd)))
	#define FD_ISSET(fd,fdset)	(*((fd_mask*)fdset) & (1 << fd))
	#define FD_ZERO(fdset)		{*((fd_mask*)fdset) = 0;}
#endif

/* LP - conflicts with signal.h!!! */
#ifndef EINTR
  #define EINTR E_SIGNAL
#endif

/* proto for Microware OS-9 unix.l library */
int select(u_int width, u_int32 *readfds, u_int32 *writefds, \
           u_int32 *exceptfds, struct timeval *timeout);

#endif  /* _OS9_SELECT */

/* EOF */
