/*
 * Client only part of rx.
 */

/* C. S: 6 mar 95: gethostbyaddr in OS9 is crippled, cannot dectect an ip address in
 * dot notation as a valid hostname. Workaround exists via inet_addr() which converts
 * directly an ip address in dot notation int a valid socket address.
 *
 * LP - 23.05.2000 - correct netdb.h include
 * CS - 07.09.2000 : general heavy clenaup
 *                   removed all old dead code: RX_KERNEL etc... 
 */

#include "allincl.h"
#ifdef OS9
  #include <INET/netdb.h>
#else
  #include <netdb.h>
#endif


int
rx_setaddress(struct rx_call_descr *cd, char *hostname, int port)
{
	/*
	   The host address resolution is usually a slow function which
	   requires file access. There are three possibilities to speed
	   it up:
		- Domain name server. General solution but not yet available
		  everywhere.
		- Local cache. This only helps for subsequent calls.
		- RPC server. This is the fastest we can get but is non-standard.
		  It should give us less than 10 ms access time on a 16 Mhz 386.
		  The additional overhead in case that the server is not
		  running is low.
	*/
	struct hostent *hp;
	struct addr addr;

	cd->port = htons(port);
	if ((hp = gethostbyname(hostname)) == (struct hostent *)0){
#ifdef OS9
                /* try to turn dot address in real address */
                cd->host = inet_addr(hostname); /* C. S. 6 mar 95 */
                if ( cd->host == 0 || cd->host==0xffffffff)
#endif
		RETURNERR(ENAMERES, ETRX, hostname);
	}

#ifdef OS9
        if(hp)
#endif
	/*cd->host = (long)*((long *)hp->h_addr);*/ /*L.P. C.S. nov 2000 */
	cd->host=(long)*((long *)hp->h_addr_list[0]);
/*
	addr.host = cd->host; addr.port = cd->port;
	TRACE(assoc && fprintf(stderr, "hostname %s resolved to %s\n",
		print_ipaddr(&addr)));
*/
	return(0);
}



int

rx_call(struct rx_call_descr *cd)
{
	/*
	   This call performs a command/response (rpc) to the remote
	   address (set by rx_setaddress()). The data of the length cd->slen
	   at address cd->smsg will be sent. The call normally blocks until
	   the return message has been received. The return message will be
	   put at address cd->rmsg (can be the same as cd->smsg). rlen
	   limits the size of the message to receive and the continuation of
	   the message can be read in later using rx_more().
	   rx_call() returns length of the reply (or part of it) read or (-1)
	   if it failed. In case of failure errno, errtype and errstring will
	   be set.
	   This call can also be used for asynchronous calls. In this case
	   cd->rlen has to be set to zero and the call blocks until the call
	   message has been acknowledged. The return value will be 0 in case
	   of success and (-1) in case of failure. On successful return
	   cd->call_number will be set to the actual call number and can be
	   later used in rx_get_reply() and rx_wait() calls.

	   After an acknowledged call a good RPC protocol waits for the reply
	   periodically sending probe packets. Unfortunately probe packets can
	   not be replied by protocols in user mode.
	   Threfore we base "no reply" timeout on the user specified call timeout.
	   and abort the call after this timeout.
	   If the call timeout is 0 then the default timeout NO_REPLY_TMO
	   will be set.
	   Actually it would be better to have 0 timeout to signify: "if the
	   server is using a kernel implementation, then there is no call
	   timeout, otherwise use default timeout"
	   The protocol timeouts are as follows:
		first retransmission after RETRANSMIT_TMO
		subsequent retransmissions after tmo*2

		what remains is what is left to the total of the user specified
		call timeout.
	   In version 2.13 timeouts has been moved to the global rx structure to 
	   allow for customisation or RX timeouts.
	*/
	long status;
	register struct association *a;
	int i, len;
	struct timeval time_now;
	TRACE(calls && fprintf(stderr, "rx_call(): ") && print_cd(cd));
	/*
	   New call is not allowed if the reply has not been all read in.
	   New asynch replies are not allowed if the number of packets
	   queued exceeds the maximum (kernel only).
	*/

	/* search the association in the active list */
	for (a = (struct association *)RXHEAD; (char *)a != NULL &&
		(a->addr_peer.host != cd->host || a->addr_peer.port != cd->port ||
		    a->type != RX_CLIENT);
		a = (struct association *)a->link)
	;
	if ((char *) a == NULL) {
		/* not found, alloc a new one */
		a = rx_alloc_assoc();
		TRACE(assoc && fprintf(stderr, "allocated new association at %lx for the first call\n",
			(long)a));
		a->type = RX_CLIENT;
		a->ourwindow = rx_glob.window;
		a->hiswindow = RX_INIT_WINDOW;
		a->addr_peer.host = cd->host; a->addr_peer.port = cd->port;
	}
	else {

		TRACE(assoc && fprintf(stderr, "found active association at %lx for addr %s\n",
			(long)a, print_ipaddr(&a->addr_peer)));
		if (a->state == RECEIVED || a->state == IDLE || a->state == SENT || a->state == ERROR_STATE) {
			a->tm = 0;
			a->state = SENDING;
		} else
			RETURNERR(EPENDING, ETRX, "call pending")
		/* cleanup possible reply of previous call (timed out but received)
		   or not read in  */
		while (a->in_pkts.head != (struct queue_item *)0)
			free(q_rhead(&a->in_pkts));
	}
	a->seq = 0;     /* on sending seq is the sequence not yet ack'ed */
	cd->call_number = a->call_number = (rx_glob.call_number ++);

	a->hdr.call_number = ntohl(a->call_number);
	a->hdr.pk_type = RX_CALL_DATA;

	a->rmsg = cd->rmsg;
	a->rlen = cd->rlen;
	/* adjust call timeout to RX_MS_TICK multiply or set default */
	if ((a->user_tmo = cd->timeout) == 0)
		a->user_tmo = NO_REPLY_TMO;
	if (a->user_tmo < RX_MS_TICK)
		a->user_tmo = RX_MS_TICK;
	a->prot_flags = cd->flags;
	tm_now(&a->timestamp);

	/* send packets */
	/*
	 * We send packets up to the window size without waiting for ACK.
	 * For kernel implementation we copy data from the user space
	 * directly to the interface. This requires process to be rescheduled
	 * for long messages and retransmissions. We could as well avoid it
	 * saving packets in kernel space. Since we optimise for short
	 * messages and no packet loss, we prefer to save time and space on
	 * copying.
	 *
	 * For user space implementation we have to give the rx header and
	 * user data to the system call interface. This is best done using
	 * the scatter/gather socket call sendmsg(). On systems which do not
	 * support this interface we have to copy header and data together
	 * before sending them off. Two strategies are possible to deal with
	 * retransmissions: Save the whole packet for retransmission or
	 * recopy in case of retransmission. We use the second method for the
	 * time being, again optimising for short packets and no retransmissions.
	 *
	 * Data is never sent from the interrupt routine (but the control
	 * packets will be). Packets are sent in sets with the PLEASE_ACK
	 * or LAST_PACKET flags in the last packet of the set. Sending waits
	 * after each set to resume which is trigged by the interrupt routine.
	 * Sending will resume after the previous set has been ack'ed or
	 * if retransmission is necessary.
	 */
	a->last_seq = (cd->slen + RX_SIZE - 1)/RX_SIZE - 1;
	/* !!! add later special case a->last_seq == 0 for efficiency */

	/* loop until everything sent or error */
	while (a->state == SENDING) {
		/* multisegment send except of last segment */
		for (i = 0; (a->seq + i < a->last_seq) && (i + 1 < a->hiswindow); i++) {
			a->hdr.seq_number = ntohs(a->seq + i);
			a->hdr.flags = 0x0;
			if ((status = send_segment(
			    a, cd->smsg + ((a->seq + i) * RX_SIZE), RX_SIZE)) < 0) {
				rx_cleanup_ass(a);
				return(status);
			}
		}
		/* last segment of the multisegment send */
		if (a->seq + i == a->last_seq) {
			/* last segment of the call */
			a->hdr.flags = LAST_SEGMENT;
			if (!(cd->flags & RX_NOACK))
				a->hdr.flags |= PLEASE_ACK;
			/* We must mark 'no reply' calls , otherwise servers
			   association cannot time out */
			if (cd->flags & RX_NOREPL)
				a->hdr.flags |= NOREPL_FLG;

			if ((len = cd->slen % RX_SIZE) == 0)
				len = RX_SIZE;
		} else {
			/* last segment of the set */
			a->hdr.flags = PLEASE_ACK;

			TRACE (timers && fprintf(stderr, "call: set modifier to %d ass %lx",
				a->tm_modifier, (long)a) );

			len = RX_SIZE;
		}
		a->hdr.seq_number = ntohs(a->seq + i);
		a->tm = rx.retr_tmo < a->user_tmo ? rx.retr_tmo : a->user_tmo;
		a->tm_modifier = 0;
		a->tm_action = do_retransmit;
		a->w_intr = 1;  /* for AWAIT_INTERRUPT */

		if ((status = send_segment(
		    a, cd->smsg + ((a->seq + i) * RX_SIZE), len)) < 0) {
			rx_cleanup_ass(a);
			return(status);
		}
		/* check whether retransmission necessary */
		if (!(cd->flags & RX_NORETR)) {
		    /* wait for either of: ACK, reply or timeout */
		    AWAIT_INTERRUPT(a);
		    while (a->tm != 0 &&
				a->tm_modifier <= rx.retransmissions &&
				a->state == SENDING) {

			tm_now(&time_now);

			status = a->user_tmo - MS_LEFT(&time_now, &a->timestamp);

			TRACE (timers && fprintf(stderr, "check retr user %ld left %ld\n",
				a->user_tmo, status) );

			if (status <= RX_MS_TICK) {
				rx_cleanup_ass(a);
				RETURNERR(ETOSERV, ETRX, "user defined timeout");
			}
			/* force end of retransmissions if necessary */
			if (status < MIN_RETRANSMIT_TMO)
				a->tm_modifier = rx.retransmissions;
			/* check if the second retr timeout should be shorter */
			if (status < a->tm)
				a->tm = (status/RX_MS_TICK)*RX_MS_TICK;
			if (!a->tm)
				a->tm = RX_MS_TICK;
			a->hdr.flags |= PLEASE_ACK;
			/* check if last segment */
			if (a->seq != a->last_seq) {
				len = RX_SIZE;
				a->hdr.flags &= (~LAST_SEGMENT);
			}
			a->hdr.seq_number = ntohs(a->seq);
			a->w_intr = 1;  /* for AWAIT_INTERRUPT */
			if (status > MIN_RETRANSMIT_TMO) {
			  /* send packets only if there is a chance to reply */
			  if ((status = send_segment(
			    a, cd->smsg + a->seq * RX_SIZE, len)) < 0) {
				rx_cleanup_ass(a);
				return(status);
			  }
			}
			AWAIT_INTERRUPT(a);
		    }
		} else /* no retransmission for unreliable calls */
		    a->state = SENT;
	}
	if (a->state == ERROR_STATE) {
		rx_cleanup_ass(a);
		RETURNERR(a->error, ETRX, "");
	}
	if (cd->flags & RX_NOREPL) {
		/* no reply data expected */
/*
		a->tm = 0;
*/
		a->state = IDLE;
		return(0);
	}
	ASSERT(a->state != SENDING);

	if (a->state != RECEIVED) {
			tm_now(&time_now);

		/* if the timer not zeroed out in the packet handling routine */
		status = a->user_tmo - MS_LEFT(&time_now, &a->timestamp);

		TRACE (timers && fprintf(stderr, "wait for reply user %ld left %ld\n",
			a->user_tmo, status) );
		TRACE (timers && fprintf(stderr, "user_tmo %ld stamp %2ld.%02ld now %2ld.%02ld\n",
		a->user_tmo,
		a->timestamp.tv_sec, a->timestamp.tv_usec/10000,
		time_now.tv_sec, time_now.tv_usec/10000) );

		if (status <= 0) {
			rx_cleanup_ass(a);
			RETURNERR(ETOREPL, ETRX, "user defined timeout");
		} else {
			a->tm = (status/RX_MS_TICK)*RX_MS_TICK;
			if (!a->tm)
				a->tm = RX_MS_TICK;
			a->tm_action = no_reply;
		}
	}
	/* wait for the return message */
	if (cd->rlen == 0) {
		/* this is an asynchronous call */
		return(0);
	}
	/* 'normal' case: blocking call with return message */
	while (a->state == SENT || a->state == RECEIVING) {
		a->w_intr = 1;  /* for AWAIT_INTERRUPT */
		AWAIT_INTERRUPT(a);
	}
	if (a->state == ERROR_STATE) {
		rx_cleanup_ass(a);
		RETURNERR(a->error, ETRX, "");
	}
	ASSERT(a->state == RECEIVED);
	/*
	   in case of user space protocol the reply has been filled
	   at interrupt time. We just have to return the actual length.
	 */
/*
	a->readp = NULL;
*/
	return(a->rlen);
}


int
rx_abort(struct rx_call_descr *cd)
{
	/* abort a running asynchronous call */

	register struct association *a;
	TRACE(calls && fprintf(stderr, "rx_abort(): ") && print_cd(cd));


	/* search the association in the active list */

	for (a = (struct association *)RXHEAD; (char *)a != NULL &&
		(a->addr_peer.host != cd->host || a->addr_peer.port != cd->port ||a->type != RX_CLIENT);
		a = (struct association *)a->link);
	if ((char *) a == NULL)
		/* not found, forget it */
		return(0);
	send_cntr(NIP, a, 0, ASS_ABORT);
	rx_cleanup_ass(a);
}
