/*
 * Fork support for rx.
 *
 * LP - 23.05.2000 - correct cast in bind() & getsockname()
 * CS - 07.09.2000 : general heavy clenaup
 *                   removed all old dead code: RX_KERNEL etc...
 */
#include "allincl.h"

/* when we fork a new server the old socket has to be closed and a new one
   established. The new socket is set in rx.newsd. After the fork each
   process has to close one of the descriptors. The spawned process should
   set rx.sd to rx.newsd and rx.newsd to -1;
*/
int
rx_newsd(short unsigned int *port)
                     
/*
   To get a new descriptor & port for rx. This is necessary when forking
   a new server instance. Returns the new fd and the port will be set.
*/
{
	struct sockaddr_in sin_from;
	int sinlen = sizeof(struct sockaddr_in);
	int status;

	TRACE(calls && fprintf(stderr, "rx_newsd() \n"));
	if ((rx.newsd != -1) || (rx.initialised == 0) ){
		close(rx.newsd);
		rx.newsd = -1;
		if (rx.initialised == 0)
		    RETURNERR(EINITIALIZ, ETRX, "new socket already open");
	}

	if ((rx.newsd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		rx.newsd = -1;
		RETURNERR(errno, ETUNIX, "socket()");
	}
#if !defined(NOMEMLIB)
	memset((char*)&sin_from, '\0', sizeof(struct sockaddr_in));
#else
	/* another brand of memory copy functions from BSD */
	bzero((char*)&sin_from, sizeof(struct sockaddr_in));
#endif
	sin_from.sin_family = AF_INET;
	sin_from.sin_addr.s_addr = htonl(INADDR_ANY);
	sin_from.sin_port = *port;
	if ((status = bind(rx.newsd, (struct sockaddr *)&sin_from, \
                     sizeof(struct sockaddr_in))) < 0) {
		close(rx.newsd);
		rx.newsd = -1;
		RETURNERR(errno, ETUNIX, "bind()");
	}

	/* bind does not get the new port so we have to get it */
	if (getsockname(rx.newsd, (struct sockaddr *)&sin_from, &sinlen) < 0) {
		close(rx.newsd);
		rx.newsd = -1;
		RETURNERR(errno, ETUNIX, "getsockname()")
	}
	*port = ntohs(sin_from.sin_port);
	return(rx.newsd);
}


/*
 * rx_fork() returns values as the fork() system call. The child continues
 * to listen on the old port and the parent (detached) is using a new port
 * which is returned in port.
 */
rx_fork(void)
{
	int pid;

	/*
	 * Until we have sorted out the roles of client and server it is not
	 * quite clear who will get the packets. This should not matter
	 * since we have retransmissions.
	 */
#if !(defined(OS9))
	if ((pid = fork()) < 0) {       /* cannot fork */
		close(rx.newsd);
		rx.newsd = -1;
		RETURNERR(errno, ETUNIX, "fork()")
	} else if (pid == 0) {          /* child */
		/*
		 * In child we continue to process requests on
		 * the well-defined server port.
		 */
		close(rx.newsd);
		rx.newsd = -1;
		return(0);      /* child returns 0 */
	}
	/*
	   Parent closes the current rx port and assigns the new port.
	 */
	if (rx_closeup() < 0)
		return(-1);
	rx.sd = rx.newsd;
	rx.newsd = -1;
	rx.initialised = 1;     /* reset flag cleaned by rx_closeup */
	return(pid);    /* parent returns the child's pid */
#else /* OS9 */
	char *pseudoargv[2];  /* 'cause brain damaged os9 linker */
	extern char *_prgname(), **_environ;
	extern int os9fork();

	/*
	   We close the current rx port and assign the new port.
	 */
	if (rx_closeup() < 0)
		return(-1);
	rx.sd = rx.newsd;
	rx.newsd = -1;
	rx.initialised = 1;     /* reset flag cleaned by rx_closeup */

	pseudoargv[0] = _prgname();
	pseudoargv[1] = 0;
	return(os9exec(os9fork, pseudoargv[0], pseudoargv, _environ, 0, 0, 3));
#endif
}

#if !(defined(OS9))

/*
 * rx_fork_norm() is like rx_fork() but the father continues
 * to listen on the old port and the child is using a new port
 * which is returned in port. This is intended for fathers which want to
 * wait or systems which are capable of nowait.
 */
rx_fork_norm(void)
{
	int pid;

	/*
	 * Until we have sorted out the roles of client and server it is not
	 * quite clear who will get the packets. This should not matter
	 * since we have retransmissions.
	 */
	if ((pid = fork()) < 0) {       /* cannot fork */
		close(rx.newsd);
		rx.newsd = -1;
		RETURNERR(errno, ETUNIX, "fork()")
	} else if (pid != 0) {          /* father */
		/*
		 * We continue to process requests on
		 * the well-defined server port.
		 */
		close(rx.newsd);
		rx.newsd = -1;
		return(pid);
	}
	/*
	   Child closes the current rx port and assigns the new port.
	 */
	if (rx_closeup() < 0)
		return(-1);
	rx.sd = rx.newsd;
	rx.newsd = -1;
	rx.initialised = 1;     /* reset flag cleaned by rx_closeup */
	return(0);
}
#endif /* OS9 */

rx_getsockport(int *sd, short unsigned int *port)
{
	*sd = rx.sd;
	*port = rx.port;
	return(0);
}
