/*
 * Definitions for the RX protocol not in rx.h
 * CS - 07.09.2000 : general heavy clenaup
 *                   removed all old dead code: RX_KERNEL etc... 
 */

/*
 * Generally useful definitions.
 */



/*
 * Things which are not defined in OS9
 */
#if defined(OS9)
#define EINTR 4
#endif


/* Random number generation function. random() is better but not available
   everywhere yet */
#if defined(HPPA)
#define random rand
#define srandom srand
#endif

/*
 * Public, i.e. externally visible.
 */
#define public
#define private static

#ifdef DEBUG
/*
 * Tracing facility.
 */
#define TRACE(x) if (rx_glob.trace.tron) rx_glob.trace.x

#else /* DEBUG */
/*
 * Ignore tracing
 */
#define TRACE(x)
#endif
/*
 * Assertion facility.
 */
#ifdef DEBUG

#define ASSERT(x) if (!(x)) { \
	fprintf(stderr, "File %s, line %d: assertion (x) failed.\n", \
		__FILE__, __LINE__); \
	abort(); } else
#define SHOULDNT_HAPPEN 0
#else   /* NO DEBUG */
#define ASSERT(x)
#endif


/*
 * Inline functions.
 * Arguments may be evaluated more than once.
 */
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#define MIN(x,y) ((x) < (y) ? (x) : (y))


#define send_cntr(NIP, x, y, z) send_cntrl(x, y, z)
#define RX rx
#define RXHEAD rx.ass.head


/*
 * Message/segment types
 */
#define RX_CALL_DATA    1
#define RX_RETURN_DATA  2
#define RX_CALL_ACK     3
#define RX_RETURN_ACK   4
#define RX_CNTRL        5


/*
 * state of the association
 */
#define IDLE            0
#define SENDING         1
#define SENT            2
#define RECEIVING       3
#define RECEIVED        4
#define ERROR_STATE     5


/*
 * The protocol is asymmetric so we need type of association
 */
#define RX_CLIENT       1
#define RX_SERVER       2
/*
 * Control bits
 */
#define ACK (1<<0)
#define PLEASE_ACK (1<<1)
#define NOREPL_FLG (1<<2)
#define LAST_SEGMENT (1<<3)

#define NO_PEER (1<<4)
#define UNREL_FLG (1<<4)
#define BACK_OFF (1<<5)
#define ASS_ABORT (1<<6)
/*
 * Timer intervals
 * Intervalls are multiply of RX ticks. The tick will typically correspond
 * to 50 or 100 miliseconds. Since intervals are counters which are decremented
 * every tick, in some implementations (e.g. kernel) the first decrement will
 * occur after a fraction of a tick.
 *
 * tick in milliseconds, usually less than 1000 ms
 * should correspond to ~ 4 times an empty RPC call
 * depending on host/network performance 10, 20, 50, 100 are recommended values
 */
#if defined(NC_SELECT_MS)
#define RX_MS_TICK NC_SELECT_MS
#else
#define RX_MS_TICK 50  /* 50 ms is default value */
#endif

#define RX_SELECT_TMO 2000
#define RETRANSMIT_TMO 400        /* timeout for first ACK - subseq multiplied by 2  */
#define MIN_RETRANSMIT_TMO RX_MS_TICK*2
#define BACKOFF_RETR_TMO 2000
#define RETRANSMISSIONS 3       /* n of retransmission timeouts before err result */
#define NO_REPLY_TMO 60000         /* initial value for reply timeout */
#define NO_ACTIVITY_TMO 200     /* association lifetime in seconds since last use */
#define NET_WAIT_CNT 10         /* to avoid checking no activity too often */
#define GARBAGE_COLLECT_CNT 500 /* freeing free lists */

/*
 * various limits
 */
#define RX_SIZE         512     /* rx data size (should be not greater than MTU of internet) */
#define RX_MAX_ASS_CNT  200     /* maximum simult. associations */
#define MAXWINDOW       5       /* maximum packets sent before ack */
#define MAXPKTQ         20      /* maximum packets queue per port (kernel) */
#define RX_INIT_WINDOW  3       /* assumed initial window size */

#define RX_MAGIC htons(1234)

struct rx_hdr {
	/* for reasons of performance header length is multiply of 32 bit words */
	nc_32uint call_number;
	unsigned short seq_number;
	unsigned short magic;
	unsigned char pk_type;  /* type of packet */
	unsigned char window;   /* advertised window */
	unsigned char flags;
	unsigned char unused1;

};


struct inout_pkt {
	char * link;
	long len;
	char data[sizeof(struct rx_hdr) + RX_SIZE];
};



struct addr {
	nc_32uint host;     /* in net order */
	unsigned short port;    /* in net order */
};


struct association{
	struct association *link;       /* link to the next association */
	struct addr addr_peer;          /* address of the peer */
	unsigned long call_number;      /* current call number */
	unsigned char type;             /* client or server */
	unsigned char state;            /* state of association */
	unsigned char w_intr;           /* to wait for the right interrupt */
	unsigned short error;           /* error number in case of error */
	struct rx_hdr hdr;              /* rx header */
	/* timers */
	long tm;                        /* action timer to decrement */
	void (*tm_action)();            /* handler for timeout action */
	struct timeval timestamp;       /* for ass lifetime and call tmo */
	unsigned long user_tmo;         /* user timeout on call */
	unsigned tm_modifier;           /* used as counter or last value */
	unsigned short prot_flags;      /* flags which apply to the current call */
	/* sending info */
	unsigned ourwindow;             /* window which goes to every packet */
	unsigned hiswindow;             /* window anounced by peer */
	unsigned seq;                   /* last sequence number sent/received */
	unsigned last_seq;              /* last sequence of the whole message on sending */
	struct nc_queue in_pkts;           /* received packets for kernel implementations
					   and in case the receive buffer not known */
	char *rmsg;                     /* buffer for the received message */
	int rlen;                       /* max message size */
	char *readp;                    /* read pointer for streamed read */
};


/*
 * rx_glob holds variables which are global for kernel and user
 * space implementations. rx holds variables which are 'global'
 * per UDP port used for communication
 */

struct rx_glob {
	unsigned long call_number;
	int window;
	unsigned long first_call, last_call;
	struct nc_queue free_packets;
	struct nc_queue free_assoc;
	int current_length;
	struct inout_pkt *current_packet;
	struct sockaddr_in current_from;
#define current_host current_from.sin_addr.s_addr
#define current_port current_from.sin_port
	long no_activity;       /* timestamp of last check */
	int net_wait_cnt;       /* to avoid checking no activity too often */
	struct inout_pkt *out_packet;
	struct trace trace;
};

struct rx {
	struct nc_queue ass;
	int sleep;
	long call_tmo;          /* timeout for rx calls rx_get_call() and rx_wait */

	int initialised;
	int port;
	int sd;
	int newsd;
	unsigned short mode;            /* to set protocol variants */
	int freelists;                  /* to define if freelists should be used */
	int retransmissions;		/* number of retransmissions */
	int retr_tmo;                   /* retransmission timeout */
	int norepl_tmo;                 /* no reply timeout */
};


/*
 * External and miscellaneus definitions.
 */


extern struct rx_glob rx_glob;
extern struct rx rx;
extern void rx_dbg(void);
extern int send_segment(struct association *a, char *buf, int len); /* C.S. 07-00 */
extern void rx_flush_buffers(void);
extern void port_traced(unsigned short x, unsigned short y);
/*extern void handle_rx_pkt(register struct rx_hdr *h, int len) ; */ /*C.S. sept 2000*/
/*extern void no_activity(long int now); */ /*C.S. sept 2000*/
extern int rx_net_wait(void);
/*extern int rx_net_check(void);*/ /*C.S. sept 2000*/
/*extern void rx_cleanup_ass(struct association *a);*/ /*C.S. sept 2000*/
extern int rx_garbage_collect(void);
/*
 * Error codes
 */

extern int errno;
extern int errtype, errline;
extern char *errstring, *errfile;

extern struct association *rx_alloc_assoc(void);
extern struct inout_pkt *rx_alloc_packet(void);
extern void do_retransmit(struct association *a);
extern void no_reply(struct association *a);
extern send_cntrl( struct association *a, unsigned char type, unsigned char flags);
extern send_err(unsigned long host, unsigned short port, unsigned short lport,\
	        unsigned char type, unsigned char flags);

#ifdef NET_POLL
extern void rx_tick();
#endif

extern void rx_free_assoc(struct association *a);


#ifdef DEBUG
extern char *print_ipaddr(struct addr *paddr);
extern int print_cd(struct rx_call_descr *cd);
#endif


/*
 * Network interrupts, critical regions etc.
 */


#define AWAIT_INTERRUPT(a) { \
	TRACE(interrupts && fprintf(stderr, "... await interrupt \n")); \
	while (a->w_intr == 1) \
		rx_net_wait(); \
}
#define AWAIT_ANY_INTERRUPT { \
	TRACE(interrupts && \
	      fprintf(stderr, "... await any interrupt \n")); \
	rx_net_wait(); \
}

#define MS_LEFT(t1, t2) ( (((t1)->tv_sec - (t2)->tv_sec)*1000) + (((t1)->tv_usec - (t2)->tv_usec)/1000) )
#define SEC(t1)  ((t1)->tv_sec )
