/*
 * Interface for server to reply to a call.
 * CS - 07.09.2000 : general heavy clenaup
 *                   removed all old dead code: RX_KERNEL etc...
 */

#include "allincl.h"

int
rx_put_reply(struct rx_call_descr *cd)
{
	/*
	   Sends reply to the call previously received by rx_get_call().
	   Returns 0 if everything went OK or -1 in case of error.
	   Typically the cd would contain address and call number filled in
	   by rx_get_call().
	 */
	long status;
	register struct association *a;
	int i, len;
	struct timeval time_now;

	TRACE(calls && fprintf(stderr, "rx_put_reply(): ") && print_cd(cd));

	if (cd->slen < 0)
		RETURNERR(EBADLEN, ETRX, "bad length for send")

	/* find the association */

	for (a = (struct association *)RXHEAD; (char *)a != NULL &&
		(a->addr_peer.host != cd->host || a->addr_peer.port != cd->port
		|| cd->call_number != a->call_number);
		    a = (struct association *)a->link);

	if ((char *) a == NULL) {
		/* not found, return an error */
		if (cd->slen == 0) return(0); /* forget it, is just a sync */
		else
		 RETURNERR(EBADCALLN, ETRX, "no match for call no")
	}
	if (cd->slen == 0) {
		/* this is used to sync for calls with no reply */

		TRACE(calls && fprintf(stderr, "preparing sync to addr %s\n",
			print_ipaddr(&a->addr_peer)));

		/* cleanup association if there are any packets  */

		if (a->in_pkts.head != (struct queue_item *)0) {
			rx_cleanup_ass(a);
			return(0);
		} else {
			a->tm = 0;
			a->state = SENT;
			return(0);
		}
	}

	TRACE(calls && fprintf(stderr, "preparing reply to addr %s\n",
		print_ipaddr(&a->addr_peer)));

	a->state = SENDING;
	a->seq = 0;

	a->hdr.pk_type = RX_RETURN_DATA;
	a->hdr.call_number = htonl(a->call_number);

	a->rmsg = cd->rmsg;
	a->rlen = cd->rlen;
	/* adjust call timeout to RX_MS_TICK multiply or set default */
	if ((a->user_tmo = cd->timeout) == 0)
		a->user_tmo = NO_REPLY_TMO;
	if (a->user_tmo < RX_MS_TICK)
		a->user_tmo = RX_MS_TICK;

	/* send packets, same as for call */
	a->last_seq = (cd->slen + RX_SIZE - 1)/RX_SIZE - 1;
	/* kkk add later special case total_segments == 1 for efficiency */

	tm_now(&a->timestamp);

	/* loop until everything sent or error */
	while (a->state == SENDING) {
		/* multisegment send except of last segment */
		for (i = 0; (a->seq + i < a->last_seq) && (i + 1 < a->hiswindow); i++) {
			a->hdr.seq_number = ntohs(a->seq + i);
			a->hdr.flags = 0x0;
			if ((status = send_segment(
			    a, cd->smsg + ((a->seq + i) * RX_SIZE), RX_SIZE)) < 0) {
				rx_cleanup_ass(a);
				return(status);
			}

		}
		/* last segment of the multisegment send */
		if (a->seq + i == a->last_seq) {
			/* by default we are asking to ack reply */
			a->hdr.flags = LAST_SEGMENT | PLEASE_ACK;
			if (!(rx.mode & RX_ACK))
				a->hdr.flags ^= PLEASE_ACK;
			if ((len = cd->slen % RX_SIZE) == 0)
				len = RX_SIZE;
		} else {
			/* last segment of the set */
			a->hdr.flags = PLEASE_ACK;
			len = RX_SIZE;
		}
		a->hdr.seq_number = ntohs(a->seq + i);
		a->tm = rx.retr_tmo < a->user_tmo ? rx.retr_tmo : a->user_tmo;
		a->tm_modifier = 0;
		a->tm_action = do_retransmit;
		a->w_intr = 1;  /* for AWAIT_INTERRUPT */
		if ((status = send_segment(
		    a, cd->smsg + ((a->seq + i) * RX_SIZE), len)) < 0) {
			rx_cleanup_ass(a);
			return(status);
		}
		/* wait for either of: ACK, reply or timeout */
		AWAIT_INTERRUPT(a);
		/* check whether retransmission necessary */
		while (a->seq <= a->last_seq && a->tm != 0 &&
				a->tm_modifier <= rx.retransmissions &&
				a->state == SENDING) {

			tm_now(&time_now);
			status = a->user_tmo - MS_LEFT(&time_now, &a->timestamp);

			TRACE (timers && fprintf(stderr, "check retr user %d left %d\n",
				a->user_tmo, status) );
			if (status <= RX_MS_TICK) {
				a->tm = 0;
				RETURNERR(ETOSERV, ETRX, "user defined timeout");
			}
			if (status < a->tm)
				a->tm = (status/RX_MS_TICK)*RX_MS_TICK;
			if (!a->tm)
				a->tm = RX_MS_TICK;
			a->hdr.flags |= PLEASE_ACK;
			if (a->seq != a->last_seq) {
				len = RX_SIZE;
				a->hdr.flags &= (~LAST_SEGMENT);
			}
			a->hdr.seq_number = ntohs(a->seq);
			a->w_intr = 1;  /* for AWAIT_INTERRUPT */
			if ((status = send_segment(
			    a, cd->smsg + a->seq * RX_SIZE, len)) < 0) {
				rx_cleanup_ass(a);
				return(status);
			}
			AWAIT_INTERRUPT(a);
		}
	}
	if (a->state == ERROR_STATE) {
		rx_cleanup_ass(a);
		RETURNERR(a->error, ETRX, "");
	}
	ASSERT(a->state != SENDING);   /* reply sent, finished */
	return(0);
}
