/* ------------------------------------------------------------------- *
 |
 | OS9Lib:  select()
 |
 |
 |     Copyright (c) 1988 by Wolfgang Ocker, Puchheim,
 |                           Ulli Dessauer, Germering and
 |                           Reimer Mellin, Muenchen
 |                           (W-Germany)
 |
 |  This  programm can  be  copied and  distributed freely  for any
 |  non-commercial  purposes.   It can only  be  incorporated  into
 |  commercial software with the written permission of the authors.
 |
 |  If you should modify this program, the authors would appreciate
 |  a notice about the changes. Please send a (context) diff or the
 |  complete source to:
 * ----------------------------------------------------------------- */
/* ----------------------------------------------------------------- *
   MANY CHANGES- ALMOST REWRITTEN FORM SCRATCH!
   most imortant of all, added test on file descriptor at exit
   after a device signaled data ready. 
   Added significance
   Claudio Scafuri 3 jul 1995
 * ----------------------------------------------------------------- *

   Added gettimeofday() for the software bus project!
   Claudio Scafuri 18 Apr 1996
 * ----------------------------------------------------------------- */

#include <time.h>
#include <signal.h>
#include <types.h>
#include "os9select.h"


#define READY_SIG   299      /* tune this ... */
#define E_UNKSVC    0xd0     /* unknown service code */
#define E_NOTRDY    0xf6     /* device not ready*/


static volatile int sigarrived = 0; /* READY_SIG occured ? */

/* signal handler for select() - PRIVATE function */

static  void
   sighand(int signum)
{
   sigarrived = signum;   
}

/*
* S E L E C T
*  returns # of fd's ready for read/write/exeption, -1 if interrupted or
*  a bad fd was encountered or 0 if none fd is ready for the action.
*/

int
   select(int nfds, int* rdp, int* wrp, int* exp, struct  timeval* timeout)
{
   register int      i;
   register unsigned int shift=0;
   register unsigned int signaled = 0; 
   register int rd=0;
   int tmp = 0;
   int      leave = 0;
   u_int32  tout=0;
   error_code err=0;
   
   /* set pointers to dummy value if unspecified */
   
   if( rdp == 0) rdp = &tmp;
   
   /* check special case: poll or undefinite wait */
   if(timeout==0) tout=0; /* wait indefinitely */
   else
      tout=timeout->tv_sec * CLOCKS_PER_SEC + timeout->tv_usec;
   if (tout==0 && timeout ){ /* means polling */
      
      for( i = 0; i< nfds; i++) {
	 
	 /* if data is available for reading mark it as ready */
	 shift = 1<<i;
	 if( (*rdp & shift) && ( _gs_rdy(i) > 0 || (errno == E_UNKSVC))){
	    rd |= shift;
	    leave++;
	 }
      }
      return leave;	 
      
   } 
   
   /* prepare for waiting .., block signals */
   sigmask(1);
   
   for( i = 0 ; i<nfds; i++) {
      shift = 1<<i;
      if( *rdp & shift) {
	 _ss_ssig(i, READY_SIG);
	 signaled |= shift;
      }      
   }
   if(errno == E_NOTRDY) errno=0; /* ignore error */
   
   sigarrived=0;
   signal(READY_SIG,sighand); /* install signal handling */
   sigmask(-1); /* enable reception of signals */
   err=_os9_sleep(&tout); /* wait tout ticks or signal */
   sigmask(1);
   switch(sigarrived){
      case 0:{	   /* no signal, timeout occurred */
	 signal(READY_SIG,SIG_IGN);
	 return(err);
	 break;
      }
      case READY_SIG:{	    
	 leave=0;
	 for( i =0; i<nfds; i++){
	    shift = 1<<i;
	    if(signaled & shift) _ss_rel(i);
	    
	    /* if data is available for reading mark it as ready */
	    if( (*rdp & shift ) && ( _gs_rdy(i) > 0 || (errno == E_UNKSVC))){
	       rd |= shift;
	       leave++;			  
	    }
	 }
	 
	 /* sigmask(-1);*//*reenable signal handling */	 
	 *rdp = rd;
	 return leave;
	 
      }
	 break;
      default:{
	 errno=EINTR;
	 return -1;
	 break;
      }
   }    
   
}
