/*
 * Adapted from iptst to test rx protocol. The below comments refer to
 * iptst.
 *
 * Server program to test functionality and performance of various
 * protocols.
 * This server listens on the TST_PORT to receive connection
 * requests. Requests are processed by the server itself or by a new
 * instance forked depending on the clients request.
 * An optional mode exists to fork only when the server is already busy.
 *
 * Following test modes exist:
 *      source:         client sends messages of length block length
 *                      repeated n times.
 *      transaction     as source but for every message a message of
 *                      identical length is sent back
 * block length repetition count number of forks etc are part of the
 * first packet sent.
 *
 * Future extensions (existing for the similar RATP test)
 * Any (remote) procedure can be called instead of echoing (see echo trans)
 * New instance of the server is forked only if the master server is busy
 * processing previous transaction.
 *
 * transaction(in_buf, in_len, out_buf, out_len, operation_type)
 *      char *in_buf;           ( address of input buffer )
 *      int in_len;             ( inbut buffer length in bytes )
 *      char *out_buf;          ( addres of the output buffer )
 *      int *out_len;           ( address where the length of the output
 *                                buffer will be set )
 *      char *operation_type;   ( string describing the operation )
 */
#include "sysconfig.h"

#if defined(OS9)
#include <types.h>
#include <in.h>
#include <socket.h>
#else
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif
#include <stdio.h>
#include <errno.h>
#include "rxtst.h"
#include "rx.h"

/*
 * Error handling
 */
extern int errno;
int errtype;
char *errstring, *errfile;
int errline;

#define CONN_TIMEOUT 200
#define MAXBLSIZE 10240
/*
 * Variables which can be set through the program call parameters.
 */
	unsigned short port = TST_PORT;
	int save_fork = 0;
	int protocol = P_RX;
	int prototype = SOCK_STREAM;
extern int network_lossage;

/*
 * Global values for the server status
 */
int master_server = 1;
int active_transaction = 0;
int connected = 0;
/* global values for transaction processing */
int  file_fd, sd, nsd;
struct sockaddr_in sin_from, sin_to;
int sinlen = sizeof(sin_from);
char *usrdir = "/usr/";         /* user directory */
char *fname = NULL;

char *in_buf, *out_buf;         /* buffers for command and results */
char *in_p, *out_p;             /* buffer pointers */
unsigned int in_len, out_len, bytes_free, bytes_read;

/* externals */
extern char *malloc();
extern char *strdup();
extern char *strchr();
extern long atol();


/* THIS IS THE ECHO TRANSACTION */
void
transaction(in_buf, in_len, out_buf, out_len, operation_type)
	char *in_buf;           /* address of input buffer */
	unsigned int in_len;             /* inbut buffer length in bytes */
	char *out_buf;          /* addres of the output buffer */
	unsigned int *out_len;           /* address where the length of the output
				   buffer will be set */
	char *operation_type;   /* string describing the operation */
{
/*
	movem(in_buf, out_buf, in_len);
*/
	*out_len = in_len;
	return;
}


main(argc, argv)
	int argc;
	char **argv;
{
	char *argv0;

	int status;
	int pid;
	struct tst_request *rq, srq;
	long bytes_total, current_length, bytes_read, bytes_written;
	char *buf;
#ifdef RX_PROTO
	struct rx_call_descr cd;
#endif

	char *str, *namestr, *uname, *pname;
	int i;
	int emask;
/*
 * Read call parameters.
 */
	argv0 = argv[0];
	while (--argc > 0 && **++argv == '-') {

		switch ((*argv)[1]) {
		    case 'p':
			if (--argc > 0)
				port = atoi(*++argv);
			break;
		    case 't':
			if (--argc > 0)
				rx_trace(*++argv);
			break;
		    case 'l':
			if (--argc > 0)
				network_lossage = atoi(*++argv);
			break;
		    case 'f':
				save_fork = 1;
			break;
		    case 'P':
			switch ((*argv)[2]) {
				case 'u':
					protocol = P_UDP;
					prototype = SOCK_DGRAM;
					break;
				case 't':
					protocol = P_TCP;
					prototype = SOCK_STREAM;
					break;
				case 'x':
					protocol = P_RX;
					break;
				case 'U':
					protocol = P_UDP_CN;
					prototype = SOCK_DGRAM;
					break;
				default:
					goto usage;
			}
			break;
		    case 'z': /* rx trace */
			if (--argc > 0)
				rx_trace(*++argv);
			break;
		    default:
			goto usage;
		}
	}
	if (argc != 0) {
usage:
		fprintf(stderr, "Usage: %s [-p port number] [-P{u|t|x|U}] [-f save fork)] \n\
\t\n", argv0);
		exit(1);
	}

	buf = malloc(MAXBLSIZE);
	rq = (struct tst_request *)buf;

		rx_init(&port);
		/* set up cd */
		cd.smsg = buf;
		cd.rmsg = buf;
		cd.timeout = 0xffffff;  /* quite a while */

	/* wait for the request and read it */
again:
	for (;;) {
/*
		cd.rlen = MAXBLSIZE;
		if ((cd.rlen = rx_get_call(&cd)) < 0) {
			PRINTERR("srxtst: rx_get_call()");
			break;
		}
		if (cd.rlen == 0)
			continue;
*/
		cd.rmsg = buf;
		cd.rlen = sizeof(struct tst_request);
		if ((cd.rlen = rx_get_call(&cd)) < 0) {
			PRINTERR("srxtst: rx_get_call()");
			break;
		}

		cd.rmsg = buf + sizeof(struct tst_request);
		cd.rlen = MAXBLSIZE - sizeof(struct tst_request);
		if ((cd.rlen = rx_more(&cd)) < 0) {
			PRINTERR("srxtst: rx_more()");
			break;
		}
		if (cd.rlen == 0)
			continue;

		/* read the request params */
		srq.rtype = htons(rq->rtype);
		srq.dsize = htonl(rq->dsize);
		srq.bsize = htons(rq->bsize);
		srq.repetition = htons(rq->repetition);
		srq.nforks = htons(rq->nforks);
#ifdef DEBUG
		fprintf(stderr, "srxtst: got request type %d size %d bsize %d rep %d nforks %d\n",
			srq.rtype, srq.dsize, srq.bsize, srq.repetition, srq.nforks);
#endif
		/* perform transaction and write back results */

		/* optional sleep to put delay on the server */
/*
		{
		extern void rx_net_check();
		int i;
		for (i = 0; i < 10; i ++) {
			sleep(3);
			rx_net_check();
		}
		}
*/
/*
		transaction(in_buf, (int)(in_p - in_buf), out_buf, &out_len, fname);
*/
#ifdef DEBUG
		fprintf(stderr, "srxtst: replying transaction mode\n");
#endif
		if (srq.rtype != RQ_SRC) {
			cd.slen = srq.dsize;
			if (rx_put_reply(&cd) < 0) {
				fprintf(stderr, "srxtst: rx_put_reply() returned errno %d errtype %d\n",
					errno, errtype);
				if (errtype != 10 || !(errno == ETOSERV || errno == ETOREPL))
					break;
				/* we skip timeout errors */
			}
		}
	} /* for */
}
