/* %W% %E% */
/*
 * Origin       Eric Cooper UCB
 *
 * 26.06.90 kk  Three types of signal interface are defined:
 *              BSDsignal (sigvec() etc.) SYSV3signal (sigset() etc.)
 *              and SYSVsignal (only rudimentary signal())
 *              All systems should define one of the three
 */

#include <signal.h>
#include "tm_include.h"

/*
 * Global variables maintained by this module only. This includes
 * the static wakeup alarm timer, and a flag to check if it is
 * time to wake up.
 */

static struct timer alarm_clock = {0, {0L, 0L}, tm_wakeup};
static int time_to_wake_up;


/*
 * Signal that it is time to wake up.
 */

int
tm_wakeup(void)
{
	time_to_wake_up = 1;
}


/*
 * Sleep for a specified amount of time.
 */

void
tm_sleep(long int seconds, long int microseconds)
{
	register int x;
#if defined(_POSIX_SOURCE)
	sigset_t sigmask;

	sigemptyset(&sigmask);
#endif
	alarm_clock.interval.tv_sec = seconds;
	alarm_clock.interval.tv_usec = microseconds;
	time_to_wake_up = 0;
	tm_start(&alarm_clock);
	while (!time_to_wake_up)
#if defined(_POSIX_SOURCE)
		sigsuspend(&sigmask);
#else
#if defined(BSDsignal)
		sigpause(0);
#else
		pause();
#endif
#endif
}
