/* %W% %E% */
/*
 * Origin       Eric Cooper UCB
 *
 * 24/01/89 (vfr) argument of alarm casted to unsigned
 * 29/05/90 (raw) include stdio.h if DEBUG
 * 26.06.90 kk  Three types of signal interface are defined:
 *              BSDsignal (sigvec() etc.) SYSV3signal (sigset() etc.)
 *              and SYSVsignal (only rudimentary signal())
 *              All systems should define one of the three
 */

#include "tm_include.h"

#if defined(DEBUG)
#include <stdio.h>
#endif

/*
 * Global data structure maintained by the timer package, a queue
 * of timers.
 */

struct nc_queue tm_queue = { 0, 0 };              /* Timer queue */
/*int     (*tm_interrupt)() = tm_clockinterrupt; */ /* Interrupt handling routine */
void     (*tm_interrupt)(int) = tm_clockinterrupt;
int     tm_enabled = 0;                         /* Signal handler enabled */


/*
 * Start a timer with a certain deadline.
 */

void
tm_startclock(register struct timeval *now, register struct timeval *deadline)
{
	struct itimerval it;
	register struct itimerval *i;

	ASSERT(tm_compare(now, deadline) < 0);
	/*
	 * Convert absolute deadline into relative interval.
	 */
	i = &it;
	i->it_value = *deadline;
	i->it_interval.tv_sec = i->it_interval.tv_usec = 0L;
#if !defined(SYSVtime)
	if ((i->it_value.tv_usec -= now->tv_usec) < 0) {
		i->it_value.tv_usec += 1000000;
		i->it_value.tv_sec -= 1;
	}
#else
	i->it_value.tv_usec = 0L;
#endif
	i->it_value.tv_sec -= now->tv_sec;
#if defined(BSDtime)
	if (setitimer(ITIMER_REAL, i, 0) != 0) {
		perror("setitimer");
		exit(1);
	}
#else
	alarm((unsigned)i->it_value.tv_sec);
#endif
}
