/* %W% %E% */
/*
 * Origin       Eric Cooper UCB
 *
 * Modified 1986 by Inge B. Kristiansen.
 * 
 * CS - 05.09.2000 adeguato prototipo tm_clockinterrupt per essere conforme
 *                 alle specifiche di signal (2)
 */

/*
 * Definitions for timer package.
 */
#ifndef _TIMER_H_
#define _TIMER_H_
#include "sysconfig.h"

/*
 * Compare two timeval structures.
 * Result is <, ==, or > 0, according to whether t1 is <, ==, or > t2.
 */

#define tm_compare(t1, t2)      (((t1)->tv_sec == (t2)->tv_sec) ? ((t1)->tv_usec - (t2)->tv_usec) : ((int)((t1)->tv_sec) - (int)((t2)->tv_sec)))

/*
 * These definitions should be included from sys/time.h
 * To cope with systems which do not have it we keep them here.
 */
#if defined(_NO_TIMEVAL_DEF_)
/*
 * Structure returned by gettimeofday(2) system call,
 * and used in other calls.
 */
typedef struct timeval {
	long	tv_sec;		/* seconds */
	long	tv_usec;	/* and microseconds */
} timeval;
#endif /* _NO_TIMEVAL_DEF_ */

#if defined(_NO_ITIMERVAL_DEF_)
/*
 * Names of the interval timers, and structure
 * defining a timer setting.
 */
#define	ITIMER_REAL	0
#define	ITIMER_VIRTUAL	1
#define	ITIMER_PROF	2

typedef struct  itimerval {
	struct	timeval it_interval;	/* timer interval */
	struct	timeval it_value;	/* current value */
} itimerval;
#endif /* _NO_ITIMERVAL_DEF_ */

/*
 * Timer structure.
 */
typedef struct timer {
	struct timer *link;		/* used by queue package */
	struct timeval interval;	/* timer interval */
	int (*action)();                /* action on timer expiration */
	char *info;			/* arbitrary client information */
	struct timeval deadline;	/* used by timer package */
} timer;

/*
 * External declarations.
 */

#ifdef _NO_PROTO
extern  void tm_init();
extern  void tm_enable();
extern  void tm_disable();
extern  int tm_deadline();
extern  void tm_initelapsed();
extern  int tm_printelapsed();
extern  int tm_clockinterrupt();
/*extern  int (*tm_interrupt)();*/
extern  void (*tm_interrupt)();
extern  void tm_now();
extern  int tm_printinterrupt();
extern  int tm_printatimer();
extern  int tm_printtimers();
extern  int tm_printtimeval();
extern  void tm_sleep();
extern  int tm_wakeup();
extern  void tm_start();
extern  void tm_startclock();
extern  void tm_stop();
extern  int tm_timertest();

#else /* _NO_PROTO */
/*extern  void tm_init(int (*handler) (), int enabled);*/
extern void tm_init( void(*handler)(int), int enabled); /* CS - 05.09.2000 */
extern void tm_enable(void);
extern  void tm_disable(void);
extern  int tm_deadline(register struct timer *t, register struct timeval **now);
extern  void tm_initelapsed(void);
extern  int tm_printelapsed(void);
/*extern  int tm_clockinterrupt(void);*/
extern  void tm_clockinterrupt(int sig); /* CS - 05.09.2000 */
/*extern  int (*tm_interrupt)();*/
extern  void (*tm_interrupt)(int);       /* CS - 05.09.2000 */
extern  void tm_now(register struct timeval *t);
extern  int tm_printinterrupt(char *name, int entering);
extern  int tm_printatimer(register struct timer *t);
extern  int tm_printtimers(char *message);
extern  int tm_printtimeval(register struct timeval *tv);
extern  void tm_sleep(long int seconds, long int microseconds);
extern  int tm_wakeup(void);
extern  void tm_start(register struct timer *t);
extern  void tm_startclock(register struct timeval *now, register struct timeval *deadline);
extern  void tm_stop(register struct timer *t);
extern  int tm_timertest(register struct timer *a, register struct timer *b, register struct timer *t);
#endif /* _NO_PROTO */

#endif  /* _TIMER_H_ */
